function Collection(keysDef, log) {
	this.keys = keysDef;
	this.objects = {};
	
	this.getKeyForObject = function(object) {
		if(this.keys.length == 0)
			return null;
			
		var key = object[this.keys[0]];
		for(var idx = 1; idx < this.keys.length; idx++)
			key = key + '|' + object[this.keys[idx]];
			
		return key;
	};
	
	this.add = function(object) {
		if(this.objects === undefined)
			this.objects = {};
		this.objects[this.getKeyForObject(object)] = object;
	};
	
	this.get = function(key) {
		if(this.objects === undefined)
			return undefined;
			
		return this.objects[key];
	};
	
	this.getByIndex = function(index) {
		if(this.objects === undefined)
			return undefined;
		var key = Object.keys(this.objects)[index];
		
		return this.get(key);
	}
	
	this.length = function() {
		if(this.objects === undefined)
			return 0;
		return Object.keys(this.objects).length;
	}

	this.remove = function(key) {
		delete this.objects[key];
	}
	
	this.removeAll = function() {
		this.objects = {};
	};
	
	this.toArray = function() {
		/*var array = $.map(this.objects, function(value, index) {
			return [value];
		});
		
		return array;*/
		
		var array = [];
		for(var idx = 0; idx < this.length(); idx++)
			array.push(this.getByIndex(idx));
		return array;
	}
	
	this.update = function(object) {
		this.add(object);
	};	
};

ConnectionState = {
    UNCONNECTED: 	'UNCONNECTED',
    UNCONNECTED: 	'CONNECTING',
	CONNECTED: 		'CONNECTED',
	ERROR:			'ERROR'
};

QueryState = {
    STARTING: 		'STARTING',
    ACTIVE: 		'ACTIVE',
    UNCONNECTED: 	'UNCONNECTED',
    UPDATING: 		'UPDATING',
	ERROR:			'ERROR',
	SUCCESS:		'SUCCESS',
    UNSUBSCRIBED: 	'UNSUBSCRIBED'
};

QueryType = {
    SUBSCRIPTION: 	'SUBSCRIPTION',
    SNAPSHOT: 		'SNAPSHOT'
};

function LVGlobalConnection() {
	// Set the configuration parameters
	var url 		= '/lv/client';

	// Initialize objects
	var init 		= angular.injector(['ng']);
	var log 		= init.get('$log');
	var lvconn		= null;
	var conns 		= new Collection(['controller']);

	// Initialize class variables
	this.state		= ConnectionState.UNCONNECTED;

	// Initialize a local variable to hold current reference
	var self 		= this;
	
	/** Connect to LiveView and prepare the queries */
	this.connect = function(lvControllerConn) {
		// Add the LVControllerConnection to the collection
		conns.add(lvControllerConn);
		
		// If the global connection isn't yet connected, do that now
		if(this.state === ConnectionState.UNCONNECTED) {
			this.state = ConnectionState.CONNECTING;
			
			LiveView.connect({
					url: url,
					onClose: function(event) {
					}
				}
			).then(
				function(connection){
					//once connected, keep the reference to the connection and run queries
					lvconn = connection;
					self.state = ConnectionState.CONNECTED;
					for(var idx = 0; idx < conns.length(); idx++) {
						connectController(conns.getByIndex(idx));
				}
					
				},
				function(error){
					log.error('[Global] ' + 'Connect: ' + error.message);
					self.state = ConnectionState.ERROR;
				}
			);
		}
		// If already connected, add the controller connection
		else if(this.state == ConnectionState.CONNECTED) {
			connectController(lvControllerConn);
		}		
	};
	
	/** Set the LVConn on the controller connection */
	connectController = function(lvControllerConn) {
		lvControllerConn.setConnection(lvconn);
		lvControllerConn.processConfiguredQueries();	
	};
	
	/** Closes this connection */
	this.close = function(controller) {
		// Remove the LVControllerConnection from the collection
		conns.remove(controller);

		// If all the connections are closed, close the global connection
		if(conns.length() == 0) {
			lvconn.close().then(function(){
					self.state = ConnectionState.UNCONNECTED;
				},
				function(error){
					self.state = ConnectionState.UNCONNECTED;
					log.error('[Global] ' + 'Close: ' + (error != undefined ? error.message : 'undefined'));
				}
			);
		}
	};		
};

// Create a LV Global Connection object
var lvGlobalConn = new LVGlobalConnection();

function LVControllerConnection(config) {
	// Debug flag
	var debug = false;
	
	// Set the configuration parameters
	var scope 		= config.scope;
	var log 		= config.log;
	var controller 	= config.controller;
	var configurl 	= config.configurl;
	
	// Initialize objects
	var lvconn 		= null;
	var queries 	= new Collection(['id']);
	
	// Initialize a global variable as well for controller name
	this.controller = controller;
	
	// Initialize a local variable to hold current reference
	var self 		= this;
	
	/** Load the query configuration and then connect */
	this.init = function() {
		// If there is a configurl, load queries
		if(configurl != null) {
			$.ajax({
				type: 'GET',
				url: configurl,
				dataType: 'json',
				success: function(data) {
					// Add each loaded query to the queries collection
					for(var idx = 0; idx < data.queries.length; idx++)
						queries.add(data.queries[idx]);
						
					// Connect using the LVGlobalConnection
					//   This will call setConnection() and processConfiguredQueries() on this object
					//   when successfully connected.
					lvGlobalConn.connect(self);
				}
			});	
		}
		// Otherwise just connect
		else {
			// Connect using the LVGlobalConnection
			//   This will call setConnection() and processConfiguredQueries() on this object
			//   when successfully connected.
			lvGlobalConn.connect(self);
		}
	};
	
	/** Sets the LV connection */
	this.setConnection = function(conn) {
		lvconn = conn;
	}
	
	/** Retrieve a list of tables in the system */
	this.listTables = function(key, includeSystemTables) {
		lvconn.getTables({
			includeSystemTables: includeSystemTables
		}).then(
			function(result){		
				scope.$broadcast('connection:' + key + ':listTables', "success", result);
			},
			function(error){
				log.error('[' + controller + '] ' + 'List Tables: ' + error);
				scope.$broadcast('connection:' + key + ':listTables', "error", error);
			}
		);
	};
	
	/** Sends a tuple */
	this.sendTuple = function(key, serverUri, inputStream, input, outputStream) {		
		var inputTuple = new LiveView.Tuple(null, input);
		if(outputStream != undefined)
			inputTuple.fieldMap.__SendTupleResponseKey = null;
			
		lvconn.sendTuple(inputTuple, inputStream, outputStream, {
			serverUri: serverUri
		}).then(
			function(result){
				var output = undefined;				
				if(outputStream != undefined) {
					output = result.tuple.fieldMap;
					delete output.__SendTupleResponseKey;
				}
				scope.$broadcast('connection:' + key + ':sendTuple', "success", output);
			},
			function(error){
				log.error('[' + controller + '] ' + 'SendTuple: ' + error);
				scope.$broadcast('connection:' + key + ':sendTuple', "error", error);
			}
		);
	};
	
	/** Process all configured queries and run them as subscriptions or snapshots */
	this.processConfiguredQueries = function() {
		for(var idx = 0; idx < queries.length(); idx++) {
			var query = queries.getByIndex(idx);
			if(query.type == QueryType.SNAPSHOT) 
				this.snapshot(query);
			else
				this.subscribe(query);
		}
	};
	
	/** Adds the query and subscribes on the server */
	this.subscribe = function(query) {
		queries.add(query);
		query.status = QueryState.STARTING;
		query.type = QueryType.SUBSCRIPTION;
		query.objects = new Collection(['tid']);
		scope.$broadcast('query:' + query.id + ':status', query);
		
		lvconn.subscribe(
			new LiveView.Query(query.template, query.parameters), {
				onSnapshotStart: function(event){
					query.schema = event.schema;
					query.status = event.subscription.status.string;
					query.snapshot = true;
					scope.$broadcast('query:' + query.id + ':status', query);
					scope.$broadcast('query:' + query.id + ':subscribe', 'onSnapshotStart', event);
					if(debug == true)
						log.debug('[' + controller + '] ' + 'query:' + query.id + ':subscribe:onSnapshotStart');					
				},
				onInsert: function(event){
					var object = query.objects.get(event.tuple.id);
					if(object == null) {
						object = {tid: event.tuple.id};
						query.objects.add(object);
					}
					for(var idx = 0; idx < query.schema.fields.length; idx++) {
						var fieldName = query.schema.fields[idx].name;
						var fieldValue = event.tuple.fieldMap[fieldName];
						if(fieldValue == undefined)
							object[fieldName] = null;
						else
							object[fieldName] = fieldValue;
					}
					scope.$broadcast('query:' + query.id + ':subscribe', 'onInsert', event, object);
					if(debug == true)
						log.debug('[' + controller + '] ' + 'query:' + query.id + ':subscribe:onInsert tid=' + event.tuple.id);
				},
				onUpdate: function(event){					
					var object = query.objects.get(event.tuple.id);
					for(var idx = 0; idx < Object.keys(event.tuple.fieldMap).length; idx++) {
						var fieldName = Object.keys(event.tuple.fieldMap)[idx];
						var fieldValue = event.tuple.fieldMap[fieldName];
						object[fieldName] = fieldValue;
					}
					scope.$broadcast('query:' + query.id + ':subscribe', 'onUpdate', event, object);					
					if(debug == true)
						log.debug('[' + controller + '] ' + 'query:' + query.id + ':subscribe:onUpdate tid=' + event.tuple.id);
				},
				onDelete: function(event){
					var object = query.objects.get(event.tuple.id);
					query.objects.remove(event.tuple.id);
					scope.$broadcast('query:' + query.id + ':subscribe', 'onDelete', event, object);
					if(debug == true)
						log.debug('[' + controller + '] ' + 'query:' + query.id + ':subscribe:onDelete tid=' + event.tuple.id);
				},
				onSnapshotEnd: function(event){
					query.snapshot = undefined;
					scope.$broadcast('query:' + query.id + ':subscribe', 'onSnapshotEnd', event);
					if(debug == true)
						log.debug('[' + controller + '] ' + 'query:' + query.id + ':subscribe:onSnapshotEnd');
				},
				onError: function(error){
					query.error = error.message;
					query.status = QueryState.ERROR;
					scope.$broadcast('query:' + query.id + ':status', query);
					scope.$broadcast('query:' + query.id + ':subscribe', 'onError', error);
					log.error('[' + controller + '] ' + 'There was an error while running query ');
					log.error(event);
					log.error(error);
				}
			}
		).then(
			function(result){
				if(debug == true)
					log.debug('[' + controller + '] ' + 'query:' + query.id + ':subscribe:success');
				query.error = undefined;
				query.subscription = result.subscription;
				scope.$broadcast('query:subscribe', query);
				scope.$broadcast('query:' + query.id + ':status', query);
				scope.$broadcast('query:' + query.id + ':subscribe', 'success', result);
			},
			function(error){
				log.error('[' + controller + '] ' + 'Query: ' + error.message);
				query.error = error.message;
				query.status = QueryState.ERROR;
				scope.$broadcast('query:' + query.id + ':status', query);
				scope.$broadcast('query:' + query.id + ':subscribe', 'error', error);
			}
		);					
	};
	
	/** Unsubscribes the specified query subscription but does not remove it */
	this.unsubscribe = function(query, keep) {
		lvconn.unsubscribe(
			query.subscription
		).then(
			function(result){
				//log.debug('[' + controller + '] ' + 'Unsubscribed from query ' + query.id);
				var id = query.id;
				scope.$broadcast('query:unsubscribe', query);
				scope.$broadcast('query:' + id + ':unsubscribe', "success", result);
				if(keep == false) {
					queries.remove(query.id);	
					scope.$broadcast('query:' + id + ':remove', 'success');
				}
				else {
					delete query.schema;
					query.objects = new Collection(['tid']);
				}
			},
			function(result){
				log.error('[' + controller + '] ' + 'Query: ' + result.error);
				var id = query.id;
				scope.$broadcast('query:' + id + ':unsubscribe', "error", result);
				if(keep == false) {
					queries.remove(query.id);	
					scope.$broadcast('query:' + id + ':remove', 'success');
					this.closeGlobal();
				}
				else {
					delete query.schema;
					query.objects = new Collection(['tid']);
				}
			}
		);					
	};
	
	/** Adds the query unless keep is false. Then it executes query as a snapshot */
	this.snapshot = function(query, keep) {
		// Broadcast a snapshot change
		scope.$broadcast('query:' + query.id + ':snapshot', 'change');
		
		if(keep != false) queries.add(query);
		query.status = QueryState.STARTING;
		query.type = QueryType.SNAPSHOT;
		query.objects = new Collection(['tid']);
		scope.$broadcast('query:' + query.id + ':status', query);
		
		lvconn.execute(
			new LiveView.Query(query.template, query.parameters), {
				onSuccess: function(event){
					// Set schema and status and apply. 
					//   This is required to trigger draw of UI elements.
					query.schema = event.schema;
					query.status = QueryState.SUCCESS;
					scope.$broadcast('query:' + query.id + ':status', query);
					
					// Loop over the event data and create objects
					//   Broadcast an event for each object
					for(var dataIdx = 0; dataIdx < event.data.length; dataIdx++) {
						var tuple = event.data[dataIdx];
						var object = query.objects.get(tuple.id);
						if(object == null) {
							object = {tid: tuple.id};
							query.objects.add(object);
						}
						for(var idx = 0; idx < query.schema.fields.length; idx++) {
							var fieldName = query.schema.fields[idx].name;
							var fieldValue = tuple.fieldMap[fieldName];
							if(fieldValue == undefined)
								object[fieldName] = null;
							else
								object[fieldName] = fieldValue;
						}
						scope.$broadcast('query:' + query.id + ':snapshot', 'onSuccess:object', event, object);
					}
					if(keep != false) scope.$broadcast('query:snapshot', query);
					scope.$broadcast('query:' + query.id + ':snapshot', 'onSuccess', event);
					
					if(debug == true)
						log.debug('[' + controller + '] ' + 'query:' + query.id + ':snapshot:onSuccess');					
				},
				onProgress: function(event){
					scope.$broadcast('query:' + query.id + ':snapshot', 'onProgress', event);
					if(debug == true)
						log.debug('[' + controller + '] ' + 'query:' + query.id + ':snapshot:onProgress');
				},
				onError: function(error){
					query.error = error.message;
					query.status = QueryState.ERROR;
					scope.$broadcast('query:' + query.id + ':status', query);
					scope.$broadcast('query:' + query.id + ':snapshot', 'onError', error);
					log.error('[' + controller + '] ' + 'There was an error while executing query ' + event.subscription.id);
					log.error('[' + controller + '] ' + 'ERROR:' + error);
				}
			}
		).then(
			function(result){
				query.error = undefined;
				scope.$broadcast('query:' + query.id + ':status', query);
				scope.$broadcast('query:' + query.id + ':snapshot', "success", result);
			},
			function(error){
				log.error('[' + controller + '] ' + 'Query: ' + error.message);
				query.error = error.message;
				query.status = QueryState.ERROR;
				scope.$broadcast('query:' + query.id + ':status', query);
				scope.$broadcast('query:' + query.id + ':snapshot', "error", error);
			}
		);				
	};
	
	/** Removes an existing query from the list of registered queries. If the query 
	 *  is a subscription it will be unsubscribed first.
	 */
	this.remove = function(query) {
		if(query.type == QueryType.SUBSCRIPTION)
			this.unsubscribe(query, false);
		else if(query.type == QueryType.SNAPSHOT) {
			var id = query.id;
			queries.remove(query.id);	
			scope.$broadcast('query:' + id + ':remove', 'success');
			this.closeGlobal();
		}
	};

	/** Close the global connection */
	this.closeGlobal = function() {
		// If all queries are removed, close the LVControllerConnection. 
		//   This may close the global connection too if all other LVControllerConnections
		//   have been closed.
		if(queries.length() == 0) {
			lvGlobalConn.close(controller);
		}
	};
	
	/** Closes this connection */
	this.close = function() {
		// Remove all queries
		for(var idx = queries.length() - 1; idx >= 0; idx--) {
			var query = queries.getByIndex(idx);
			this.remove(query);
		}
	};	

	/** Retrieve the query by id */
	this.getQuery = function(queryId) {
		return queries.get(queryId);
	};
	
	/** Returns a collection of queries */
	this.getQueries = function() {
		return queries;
	};
	
	/** Parses the query template and extracts all parameters from the query
	 *  statement. Parameters must start with @ to be recognized.
	 *  This isn't currently used, but it's an interesting regex.
	 */
	var parseParametersFromTemplate = function(template) {
		var parameters = new Collection(['name']);
		var matches = template.match(/\@.\w*/g);
		if(matches != null) {
			for(var idx = 0; idx < matches.length; idx++) {
				parameters.add({
					name: 	matches[idx],
					value:	null
				});
			}		
		}
		return parameters;
	};
	
	/** Applies parameters to a previously registered query.
 	 *  If queryId is undefined then the parameters are applied to all queries.
	 */
	this.applyParameters = function(parameters, queryId) {
		if(queryId == undefined) {
			for(var idx = 0; idx < queries.length(); idx++) {
				var query = queries.getByIndex(idx);
				applyParametersToQuery(query, parameters);
			}
		}
		else {
			var query = queries.get(queryId);
			applyParametersToQuery(query, parameters);
		}
	};
	
	/** Applies parameters to the specified query */
	var applyParametersToQuery = function(query, parameters) {
		if(query == null || query.subscription == null)
			return;

		// Change the query status
		query.status = QueryState.UPDATING;
		scope.$broadcast('query:' + query.id + ':status', query);
		
		// Loop over the keys and apply to the query
		for(var idx = 0; idx < Object.keys(parameters).length; idx++) {
			var paramName = Object.keys(parameters)[idx];
			query.parameters[paramName] = parameters[paramName];
		}
		
		// Broadcast onDelete for all objects
		for(var idx = 0; idx < query.objects.length(); idx++) {
			var obj = query.objects.getByIndex(idx);
			scope.$broadcast('query:' + query.id + ':subscribe', 'onDelete', null, obj);
			if(debug == true)
				log.debug('[' + controller + '] ' + 'query:' + query.id + ':subscribe:onDelete tid=' + obj.tid);
		}
		
		// Remove all objects from the query
		query.objects.removeAll();

		// Apply to the query subscription
		query.subscription.applyParameters(query.parameters);
	};	
}