// This function adds show/hide handlers on div elements
(function ($) {
	$.each(['show', 'hide'], function (i, ev) {
		var el = $.fn[ev];
		$.fn[ev] = function () {
			this.trigger(ev);
			return el.apply(this, arguments);
		};
	});
})(jQuery);
	
angular.element(document).ready(function() {
	// Hides browser default context menus
	$(document).on("contextmenu", function(e) {
		$('.context-menu').hide();
		//return false;
	});				
	
	// Hide any context menu when clicking
	$(document).on("click", function(e){
		$('.context-menu').hide();
	});
});				
app.directive('livechart', function($log, Overlay, EventListener) {
	return {
		templateUrl: './templates/live-chart.html',
		restrict: 'E',
		scope: {
			'query':	'=',
		},
		transclude: true,
		controller: function($scope) {
		},
		link: function(scope, element, attrs, ctrl, transclude) {		
			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});
			
			// Set the query ID to the scope
			scope.queryId = attrs.queryId;

			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;

			// Read the category and series
			//   Category is the X-axis, series is an array
			//   of objects defining the data series
			scope.category = attrs.category;
			scope.series = eval(attrs.series);
			scope.categories = {xaxis:[], yaxis:[]};
			scope.parents = [];
			scope.colors = Highcharts.getOptions().colors;
			
			// These properties are used for microbatching the updates
			var microbatch = {
				rtime:		new Date(1, 1, 2000, 12,00,00),
				timeout:	false,
				delta:		200
			}				
			
			// Read the options into scope
			scope.options = {
				chart: {
					renderTo: 				$(element).find('div').find('div')[0],
											// Mandatory, location to render the chart
													
					animation:				attrs.chartAnimation == undefined ? true : eval(attrs.chartAnimation),
											// Optional. Enable chart animation. Default is true.
									
					backgroundColor:		attrs.chartBackgroundColor,
											// Optional. Background color for the chart. Default is #FFFFFF.
										
					borderColor:			attrs.chartBorderColor,
											// Optional. Border color for the chart. Default is #4572A7.
										
					borderRadius:			attrs.chartBorderRadius == undefined ? null : eval(attrs.chartBorderRadius),
											// Optional. Border radius for corners. Default is 0.
										
					borderWidth:			attrs.chartBorderWidth == undefined ? null : eval(attrs.chartBorderWidth),
											// Optional. Border width for the chart. Default is 0.
										
					inverted:				attrs.chartInverted == undefined ? false : eval(attrs.chartInverted),
					
					margin:					attrs.chartMargin == undefined ? undefined : eval(attrs.chartMargin),
											// Optional. Margin between outer edge of chart and plot area. Array in the form
											//   [top, right, bottom, left]. Default is 0 margin on all dimensions.
										
					marginTop:				attrs.chartMarginTop == undefined ? undefined : eval(attrs.chartMarginTop),

					marginBottom:			attrs.chartMarginBottom == undefined ? undefined : eval(attrs.chartMarginBottom),

					panning:				attrs.chartPanning == undefined ? true : eval(attrs.chartPanning),
											// Optional. Enable chart panning. Default is false.
									
					panKey:					attrs.chartPanKey,
											// Optional. Key to press to switch between zooming and panning. 
									
					plotBackgroundColor:	attrs.chartPlotBackgroundColor,
											// Optional. Plot background color for the chart.
										
					plotBackgroundImage:	attrs.chartPlotBackgroundImage,
											// Optional. Plot background image for the chart.
										
					plotBorderColor:		attrs.chartPlotBorderColor,
											// Optional. Plot border color for the chart.
										
					plotBorderWidth:		attrs.chartPlotBorderWidth,
											// Optional. Plot border width for the chart.
										
					spacing:				attrs.chartSpacing == undefined ? [10,10,15,10] : eval(attrs.chartSpacing), 
											// Optional. Distance between outer edge of chart and content. Array in the form
											//   [top, right, bottom, left]. Default is [10,10,15,10].
											
					style:					attrs.chartStyle == undefined ? {} : eval('(' + attrs.chartStyle + ')'),
											// Optional. CSS styling to apply to chart. 
									
					type: 					attrs.chartType == undefined ? 'line' : attrs.chartType,
											// Optional. Type of chart to display. May be one of area, arearange, areaspline, 
											//   areasplinerange, bar, boxplot, bubble, column, columnrange, errorbar, funnel, 
											//   gauge, heatmap, line, pie, polygon, pyramid, scatter, series, solidgauge, 
											//   spline, treemap, waterfall. Defaults to line if omitted.
											
					zoomType:				attrs.chartZoomType,
											// Optional. Decides in what dimensions the user can zoom by dragging the mouse. 
											//   Can be one of x, y or xy. Default is no zoom.
				},
				colorAxis: {
					min:					attrs.colorAxisMin == undefined ? undefined : eval(attrs.colorAxisMin),
					minColor:				attrs.colorAxisMinColor == undefined ? undefined : attrs.colorAxisMinColor,
					maxColor:				attrs.colorAxisMaxColor == undefined ? undefined : eval(attrs.colorAxisMaxColor),
				},
				credits: {
					enabled:				false,
				},
				legend: {
					align:					attrs.legendAlign == undefined ? "center" : attrs.legendAlign,
											// Optional. Legend alignment. May be left, center, or right. Default is center.

					backgroundColor:		attrs.legendBackgroundColor == undefined ? "rgba(255, 255, 255, 1.00)" : attrs.legendBackgroundColor,
											// Optional. Legend background color for the chart.
					
					borderColor:			attrs.legendBorderColor == undefined ? '#909090' : attrs.legendBorderColor,
											// Optional. Legend border color.
											
					borderRadius:			attrs.legendBorderRadius == undefined ? 0 : eval(attrs.legendBorderRadius),
											// Optional. Legend border radius.

					borderWidth:			attrs.legendBorderWidth == undefined ? 0 : eval(attrs.legendBorderWidth),
											// Optional. Legend border width.

					enabled:				attrs.legendEnabled == undefined ? false : eval(attrs.legendEnabled),
											// Optional. Display the legend. Default is true.
											
					floating:				attrs.legendFloating == undefined ? false : eval(attrs.legendFloating),
											// Optional. Float the legend. Default is false.
											
					itemDistance:			attrs.legendItemDistance == undefined ? 20 : eval(attrs.legendItemDistance),
											// Optional. In a legend with horizontal layout, the itemDistance defines the pixel 
											//   distance between each item. Defaults to 20.
					
					itemMarginBottom:		attrs.legendItemMarginBottom == undefined ? 0 : eval(attrs.legendItemMarginBottom),
											// Optional. The pixel bottom margin for each legend item. Defaults to 0.
											
					itemMarginTop:			attrs.legendItemMarginTop == undefined ? 0 : eval(attrs.legendItemMarginTop),
											// Optional. The pixel top margin for each legend item. Defaults to 0.
					
					labelFormat:			attrs.legendLabelFormat,
											// Optional. A format string for each legend label. Available variables relates to 
											//   properties on the series, or the point in case of pies. Defaults to {name}.
					
					layout:					attrs.legendLayout == undefined ? "horizontal" : attrs.legendLayout,
											// Optional. The layout of the legend items. Can be one of horizontal or vertical. 
											//   Defaults to horizontal.
											
					margin:					attrs.legendMargin == undefined ? 12: eval(attrs.legendMargin),
											// Optional. If the plot area sized is calculated automatically and the legend is not 
											//   floating, the legend margin is the space between the legend and the axis labels 
											//   or plot area. Defaults to 12.
											
					maxHeight:				attrs.legendMaxHeight == undefined ? null : eval(attrs.legendMaxHeight),
											// Optional. Maximum pixel height for the legend. When the maximum height is extended, 
											//   navigation will show.
											
					padding: 				attrs.legendPadding == undefined ? 8 : eval(attrs.legendPadding),
											// Optional. Inner padding of the legend box. Default is 8.

					shadow:					attrs.legendShadow == undefined ? null : eval(attrs.legendShadow),
											// Optional. Whether to apply a drop shadow to the legend. A backgroundColor 
											//   also needs to be applied for this to take effect. Defaults to false.
											
					symbolHeight:			attrs.legendSymbolHeight == undefined ? null : eval(attrs.legendSymbolHeight),
											// Optional. The pixel height of the symbol for series types that use a rectangle in the 
											//   legend. Defaults to the font size of legend items.
											
					symbolPadding:			attrs.legendSymbolPadding == undefined ? 5 : eval(attrs.legendSymbolPadding),
											// Optional. The pixel padding between the legend item symbol and the legend item text. 
											//   Defaults to 5.
											
					symbolRadius:			attrs.legendSymbolRadius == undefined ? 0 : eval(attrs.legendSymbolRadius),
											// Optional. The border radius of the symbol for series types that use a rectangle in 
											//   the legend. Defaults to 0.

					symbolWidth:			attrs.legendSymbolWidth == undefined ? 16 : eval(attrs.legendSymbolWidth),
											// Optional. The pixel width of the legend item symbol. Defaults to 16.
											
					title: {
						style:				attrs.legendTitleStyle == undefined ? {} : eval('(' + attrs.legendTitleStyle + ')'),
											// Optional. Style to apply to legend title.
											
						text:				attrs.legendTitleText,
											// Optional. Legend title text.
					},
					
					useHTML:				attrs.legendUseHtml == undefined ? false : eval(attrs.legendUseHtml),
											// Optional. Use HTML in legend.
											
					verticalAlign:			attrs.legendVerticalAlign == undefined ? "middle" : attrs.legendVerticalAlign,
											// Optional. Vertical alignment. May be top, middle, or bottom.
										
					//width:				attrs.legendWidth == undefined ? undefined : eval(attrs.legendWidth),
											// Optional. Legend width.
											
					x:						attrs.legendX == undefined ? 0 : eval(attrs.legendX),
					
					y:						attrs.legendY == undefined ? 0 : eval(attrs.legendY),
				},
				plotOptions: {
					series: {
						dataLabels:	{
							//format:			attrs.seriesDataLabelsFormat == undefined ? undefined : attrs.seriesDataLabelsFormat,
						},
						//stacking: 			attrs.seriesStacking == undefined ? undefined : eval(attrs.seriesStacking),
					}
				},
				subtitle: {
					align:			attrs.subtitleAlign == undefined ? "center" : attrs.subtitleAlign,
									// Optional. Sub-title alignment. Must be left, center, or right. Default is left if omitted.
									
					floating:		attrs.subtitleFloating == undefined ? undefined : eval(attrs.subtitleFloating),
									// Optional. When floating, the plot area will not make room for the title. Default is false.
									
					style:			attrs.subtitleStyle == undefined ? {} : eval('(' + attrs.subtitleStyle + ')'),
									// Optional. CSS styling to apply to sub-title. 
									
					text: 			attrs.subtitleText,
									// Optional. Sub-title.

					useHTML: 		attrs.subtitleUseHtml,
									// Optional. Use HTML for main title. Default is false.
				},
				title: {
					align:			attrs.titleAlign == undefined ? "center" : attrs.titleAlign,
									// Optional. Main title alignment. Must be left, center, or right. Default is left if omitted.
									
					floating:		attrs.titleFloating == undefined ? undefined : eval(attrs.titleFloating),
									// Optional. When floating, the plot area will not make room for the title. Default is false.
									
					text: 			attrs.titleText,
									// Optional. Main title.

					style:			attrs.titleStyle == undefined ? {} : eval('(' + attrs.titleStyle + ')'),
									// Optional. CSS styling to apply to title. 
									
					useHTML: 		attrs.titleUseHtml,
									// Optional. Use HTML for main title. Default is false.
				},
				tooltip: {
					animation:				attrs.tooltipAnimation == undefined ? true : eval(attrs.tooltipAnimation),
											
					backgroundColor:		attrs.tooltipBackgroundColor == undefined ? "rgba(255, 255, 255, 0.85)" : attrs.tooltipBackgroundColor,
											
					borderColor:			attrs.tooltipBorderColor,
					
					borderRadius:			attrs.tooltipBorderRadius == undefined ? 3 : eval(attrs.tooltipBorderRadius),
					
					borderWidth:			attrs.tooltipBorderWidth == undefined ? 1 : eval(attrs.tooltipBorderWidth),
					
					crosshairs:				attrs.tooltipCrosshairs == undefined ? undefined : eval(attrs.tooltipCrosshairs),
					
					dateTimeLabelFormats:	attrs.tooltipDateTimeFormats == undefined ? {} : eval('(' + attrs.tooltipDateTimeFormats + ')'),
					
					enabled:				attrs.tooltipEnabled == undefined ? true : eval(attrs.tooltipEnabled),
					
					followPointer:			attrs.tooltipFollowPointer == undefined ? false : eval(attrs.tooltipFollowPointer),
					
					followTouchMove:		attrs.tooltipFollowTouchMove = undefined ? false : eval(attrs.tooltipFollowTouchMove),
					
					formatter:				attrs.tooltipFormatter == undefined ? undefined : function() { return eval(attrs.tooltipFormatter); },
											
					//footerFormat:			attrs.tooltipFooterFormat == undefined ? false : attrs.tooltipFooterFormat,
					
					headerFormat:			attrs.tooltipHeaderFormat == undefined ? "<span style='font-size:10px;'>{point.key}</span><br/><span style='padding-left:2px'></span>" : attrs.tooltipHeaderFormat, 
					
					hideDelay:				attrs.tooltipHideDelay == undefined ? 500 : eval(attrs.tooltipHideDelay),
					
					pointFormat:			attrs.tooltipPointFormat == undefined ? "<span style='color:{point.color}'>\u25CF</span> {series.name}: <b>{point.y}</b><br/>" : attrs.tooltipPointFormat,
					
					shadow:					attrs.tooltipShadow == undefined ? true : eval(attrs.tooltipShadow == undefined),
					
					shape:					attrs.tooltipShape == undefined ? "callout" : attrs.tooltipShape,
					
					style:					attrs.tooltipStyle == undefined ? {} : eval('(' + attrs.tooltipStyle + ')'),
					
					useHTML:				true,
					
					valueDecimals:			attrs.tooltipValueDecimals,
					
					valuePrefix:			attrs.tooltipValuePrefix,
					
					valueSuffix:			attrs.tooltipValueSuffix,
		
					xDateFormat:			attrs.tooltipXDateFormat,
				},
				xAxis: {
					allowDecimals:			attrs.xaxisAllowDecimals == undefined ? true : eval(attrs.xAxisAllowDecimals),

					alternateGridColor:		attrs.xaxisAlternateGridColor,
					
					categories:				attrs.xaxisCategories == undefined ? undefined : eval(attrs.xaxisCategories),
					
					ceiling:				attrs.xaxisCeiling == undefined ? undefined : eval(attrs.xaxisCeiling),
					
					dateTimeLabelFormats:	attrs.xaxisDateTimeFormats == undefined ? {} : eval('(' + attrs.xaxisDateTimeFormats + ')'),
					
					endOnTick:				attrs.xaxisEndOnTick == undefined ? false : eval(attrs.xaxisEndOnTick),
					
					floor:					attrs.xaxisFloor == undefined ? undefined : eval(attrs.xaxisFloor),
					
					gridLineColor:			attrs.xaxisGridLineColor == undefined ? '#D8D8D8' : attrs.xaxisGridLineColor,
					
					gridLineDashStyle:		attrs.xaxisGridLineDashStyle,
					
					gridLineWidth:			attrs.xaxisGridLineWidth == undefined ? 0 : eval(attrs.xaxisGridLineWidth),
					
					labels: {
						align:				attrs.xaxisLabelsAlign,
						
						enabled:			attrs.xaxisLabelsEnabled == undefined ? true : eval(attrs.xaxisLabelsEnabled),
						
						format:				attrs.xaxisLabelsFormat,
						
						padding:			attrs.xaxisLabelsPadding == undefined ? undefined : eval(attrs.xaxisLabelsPadding),
						
						staggerLines:		attrs.xaxisLabelsStaggerLines == undefined ? undefined : eval(attrs.xaxisLabelsStaggerLines),
						
						step:				attrs.xaxisLabelsStep == undefined ? undefined : eval(attrs.xaxisLabelsStep),
						
						style:				attrs.xaxisLabelsStyle == undefined ? {} : eval('(' + attrs.xaxisLabelsStyle + ')'),
						
						useHTML:			attrs.xaxisLabelsUseHtml == undefined ? false : eval(attrs.xaxisLabelsUseHtml),
						
					},
					
					lineColor:				attrs.xaxisLineColor == undefined ? "#C0D0E0" : attrs.xaxisLineColor,
					
					lineWidth:				attrs.xaxisLineWidth == undefined ? 1 : eval(attrs.xaxisLineWidth),
					
					max:					attrs.xaxisMax == undefined ? undefined : eval(attrs.xaxisMax),
					
					maxPadding:				attrs.xaxisMaxPadding == undefined ? 0.05 : eval(attrs.xaxisMaxPadding),
					
					min:					attrs.xaxisMin == undefined ? undefined : eval(attrs.xaxisMin),
					
					minPadding:				attrs.xaxisMinPadding == undefined ? 0.05 : eval(attrs.xaxisMinPadding),
					
					minRange:				attrs.xaxisMinRange == undefined ? undefined : eval(attrs.xaxisMinRange),
					
					minTickInterval:		attrs.xaxisMinTickInterval == undefined ? undefined : eval(attrs.xaxisMinTickInterval),
					
					minorGridLineColor:		attrs.xaxisMinorGridLineColor,
					
					minorGridLineDashStyle:	attrs.xaxisMinorGridLineDashStyle,
					
					minorGridLineWidth:		attrs.xaxisMinorGridLineWidth == undefined ? undefined : eval(attrs.xaxisMinorGridLineWidth),
					
					minorTickColor:			attrs.xaxisMinorTickColor,
					
					minorTickInterval:		attrs.xaxisMinorTickInterval,
					
					minorTickLength:		attrs.xaxisMinorTickLength == undefined ? undefined : eval(attrs.xaxisMinorTickLength),
					
					minorTickPosition:		attrs.xaxisMinorTickPosition,
					
					minorTickWidth:			attrs.xaxisMinorTickWidth == undefined ? undefined : eval(attrs.xaxisMinorTickWidth),
					
					offset:					attrs.xaxisOffset == undefined ? 0 : eval(attrs.xaxisOffset),
					
					opposite:				attrs.xaxisOpposite == undefined ? false : eval(attrs.xaxisOpposite),
					
					reversed:				attrs.xaxisReversed == undefined ? undefined : eval(attrs.xaxisReversed),
					
					showEmpty:				attrs.xaxisShowEmpty == undefined ? true : eval(attrs.xaxisShowEmpty),
					
					showFirstLabel:			attrs.xaxisShowFirstLabel == undefined ? true : eval(attrs.xaxisShowFirstLabel),
					
					showLastLabel:			attrs.xaxisShowLastLabel == undefined ? true : eval(attrs.xaxisShowLastLabel),
					
					startOfWeek:			attrs.xaxisStartOfWeek == undefined ? 1 : eval(attrs.xaxisStartOfWeek),
					
					startOnTick:			attrs.xaxisStartOnTick == undefined ? false : eval(attrs.xaxisStartOnTick),
					
					tickAmount:				attrs.xaxisTickAmount == undefined ? undefined : eval(attrs.xaxisTickAmount),
					
					tickColor:				attrs.xaxisTickColor == undefined ? '#C0D0E0' : attrs.xaxisTickColor,
					
					tickInterval:			attrs.xaxisTickInterval == undefined ? undefined : eval(attrs.xaxisTickInterval),
					
					tickLength:				attrs.xaxisTickLength == undefined ? 10 : eval(attrs.xaxisTickLength),
					
					tickPixelInterval:		attrs.xaxisTickPixelInterval == undefined ? 100 : eval(attrs.xaxisTickPixelInterval),
					
					tickPosition:			attrs.xaxisTickPosition,
					
					tickWidth:				attrs.xaxisTickWidth == undefined ? 1 : eval(attrs.xaxisTickWidth),
					
					tickmarkPlacement:		attrs.xaxisTickmarkPlacement == undefined ? "between" : attrs.xaxisTickmarkPlacement,
										
					title: {
						align:				attrs.xaxisTitleAlign,
						margin:				attrs.xaxisTitleMargin == undefined ? null : eval(attrs.xaxisTitleMargin),
						offset:				attrs.xaxisTitleOffset == undefined ? null : eval(attrs.xaxisTitleOffset),
						rotation:			attrs.xaxisTitleRotation == undefined ? 0 : eval(attrs.xaxisTitleRotation),
						style:				attrs.xaxisTitleStyle == undefined ? {} : eval('(' + attrs.xaxisTitleStyle + ')'),
						text: 				attrs.xaxisTitleText,
					},
					
					type:					attrs.xaxisType,
					
					units:					attrs.xaxisUnits == undefined ? undefined : eval(attrs.xaxisUnits),				
				},
				yAxis: {
					allowDecimals:			attrs.yaxisAllowDecimals == undefined ? true : eval(attrs.yaxisAllowDecimals),

					alternateGridColor:		attrs.yaxisAlternateGridColor,
					
					categories:				attrs.yaxisCategories == undefined ? undefined : eval(attrs.yaxisCategories),
					
					ceiling:				attrs.yaxisCeiling == undefined ? undefined : eval(attrs.yaxisCeiling),
					
					dateTimeLabelFormats:	attrs.yaxisDateTimeFormats == undefined ? {} : eval('(' + attrs.yaxisDateTimeFormats + ')'),
					
					endOnTick:				attrs.yaxisEndOnTick == undefined ? false : eval(attrs.yaxisEndOnTick),
					
					floor:					attrs.yaxisFloor == undefined ? undefined : eval(attrs.yaxisFloor),
					
					gridLineColor:			attrs.yaxisGridLineColor == undefined ? '#D8D8D8' : attrs.yaxisGridLineColor,
					
					gridLineDashStyle:		attrs.yaxisGridLineDashStyle,
					
					gridLineWidth:			attrs.yaxisGridLineWidth == undefined ? 1 : eval(attrs.yaxisGridLineWidth),
					
					labels: {
						align:				attrs.yaxisLabelsAlign,
						
						enabled:			attrs.yaxisLabelsEnabled == undefined ? true : eval(attrs.yaxisLabelsEnabled),
						
						format:				attrs.yaxisLabelsFormat,
						
						padding:			attrs.yaxisLabelsPadding == undefined ? undefined : eval(attrs.yaxisLabelsPadding),
						
						staggerLines:		attrs.yaxisLabelsStaggerLines == undefined ? undefined : eval(attrs.yaxisLabelsStaggerLines),
						
						step:				attrs.yaxisLabelsStep == undefined ? undefined : eval(attrs.yaxisLabelsStep),
						
						style:				attrs.yaxisLabelsStyle == undefined ? {} : eval('(' + attrs.yaxisLabelsStyle + ')'),
						
						useHTML:			attrs.yaxisLabelsUseHtml == undefined ? false : eval(attrs.yaxisLabelsUseHtml),
						
					},
					
					lineColor:				attrs.yaxisLineColor == undefined ? "#C0D0E0" : attrs.yaxisLineColor,
					
					lineWidth:				attrs.yaxisLineWidth == undefined ? 1 : eval(attrs.yaxisLineWidth),
					
					max:					attrs.yaxisMax == undefined ? undefined : eval(attrs.yaxisMax),
					
					maxPadding:				attrs.yaxisMaxPadding == undefined ? 0.05 : eval(attrs.yaxisMaxPadding),
					
					min:					attrs.yaxisMin == undefined ? undefined : eval(attrs.yaxisMin),
					
					minPadding:				attrs.yaxisMinPadding == undefined ? 0.05 : eval(attrs.yaxisMinPadding),
					
					minRange:				attrs.yaxisMinRange == undefined ? undefined : eval(attrs.yaxisMinRange),
					
					minTickInterval:		attrs.yaxisMinTickInterval == undefined ? undefined : eval(attrs.yaxisMinTickInterval),
					
					minorGridLineColor:		attrs.yaxisMinorGridLineColor,
					
					minorGridLineDashStyle:	attrs.yaxisMinorGridLineDashStyle,
					
					minorGridLineWidth:		attrs.yaxisMinorGridLineWidth == undefined ? undefined : eval(attrs.yaxisMinorGridLineWidth),
					
					minorTickColor:			attrs.yaxisMinorTickColor,
					
					minorTickInterval:		attrs.yaxisMinorTickInterval,
					
					minorTickLength:		attrs.yaxisMinorTickLength == undefined ? undefined : eval(attrs.yaxisMinorTickLength),
					
					minorTickPosition:		attrs.yaxisMinorTickPosition,
					
					minorTickWidth:			attrs.yaxisMinorTickWidth == undefined ? undefined : eval(attrs.yaxisMinorTickWidth),
					
					offset:					attrs.yaxisOffset == undefined ? 0 : eval(attrs.yaxisOffset),
					
					opposite:				attrs.yaxisOpposite == undefined ? false : eval(attrs.yaxisOpposite),
					
					reversed:				attrs.yaxisReversed == undefined ? undefined : eval(attrs.yaxisReversed),
					
					showEmpty:				attrs.yaxisShowEmpty == undefined ? true : eval(attrs.yaxisShowEmpty),
					
					showFirstLabel:			attrs.yaxisShowFirstLabel == undefined ? true : eval(attrs.yaxisShowFirstLabel),
					
					showLastLabel:			attrs.yaxisShowLastLabel == undefined ? true : eval(attrs.yaxisShowLastLabel),
					
					startOfWeek:			attrs.yaxisStartOfWeek == undefined ? 1 : eval(attrs.yaxisStartOfWeek),
					
					startOnTick:			attrs.yaxisStartOnTick == undefined ? false : eval(attrs.yaxisStartOnTick),
					
					tickAmount:				attrs.yaxisTickAmount == undefined ? undefined : eval(attrs.yaxisTickAmount),
					
					tickColor:				attrs.yaxisTickColor == undefined ? '#C0D0E0' : attrs.yaxisTickColor,
					
					tickInterval:			attrs.yaxisTickInterval == undefined ? undefined : eval(attrs.yaxisTickInterval),
					
					tickLength:				attrs.yaxisTickLength == undefined ? 10 : eval(attrs.yaxisTickLength),
					
					tickPixelInterval:		attrs.yaxisTickPixelInterval == undefined ? 72 : eval(attrs.yaxisTickPixelInterval),
					
					tickPosition:			attrs.yaxisTickPosition,
					
					tickWidth:				attrs.yaxisTickWidth == undefined ? 1 : eval(attrs.yaxisTickWidth),
					
					tickmarkPlacement:		attrs.yaxisTickmarkPlacement == undefined ? "between" : attrs.yaxisTickmarkPlacement,
										
					title: {
						align:				attrs.yaxisTitleAlign,
						margin:				attrs.yaxisTitleMargin == undefined ? null : eval(attrs.yaxisTitleMargin),
						offset:				attrs.yaxisTitleOffset == undefined ? null : eval(attrs.yaxisTitleOffset),
						rotation:			attrs.yaxisTitleRotation == undefined ? 0 : eval(attrs.yaxisTitleRotation),
						style:				attrs.yaxisTitleStyle == undefined ? {} : eval('(' + attrs.yaxisTitleStyle + ')'),
						text: 				attrs.yaxisTitleText,
					},
					
					type:					attrs.yaxisType,
					
					units:					attrs.yaxisUnits == undefined ? undefined : eval(attrs.yaxisUnits),
				},				
			
				"queryViewer": 		attrs.queryViewer == undefined ? 
										false : 
										eval(attrs.queryViewer),
									// Optional. Allow query viewer modal. May be true or false.
									
				"queryEditor": 		attrs.queryEditor == undefined ? 
										false : 
										eval(attrs.queryEditor),
									// Optional. Allow query editor modal. May be true or false.																		
			};

			// Plot options by chart type
			if(attrs.chartType == "area") {
				scope.options.plotOptions.area = {
					dataLabels: {
						enabled: 		attrs.areaDataLabelsEnabled == undefined ? false : eval(attrs.areaDataLabelsEnabled),
						format: 		attrs.areaDataLabelsFormat,
					},
					stacking: 			attrs.areaStacking,
				};
			}
			else if(attrs.chartType == "areaspline") {
				scope.options.plotOptions.areaspline = {
					dataLabels: {
						enabled: 		attrs.areasplineDataLabelsEnabled == undefined ? false : eval(attrs.areasplineDataLabelsEnabled),
						format: 		attrs.areasplineDataLabelsFormat,
					},
					stacking: 			attrs.areasplineStacking,
				};
			}
			else if(attrs.chartType == "arearange") {
				scope.options.plotOptions.arearange = {
					dataLabels: {
						enabled: 		attrs.areaDataLabelsEnabled == undefined ? false : eval(attrs.areaDataLabelsEnabled),
						format: 		attrs.areaDataLabelsFormat,
					},
				};
			}
			else if(attrs.chartType == "bar") {
				scope.options.plotOptions.bar = {
					dataLabels: {
						enabled: 		attrs.barDataLabelsEnabled == undefined ? false : eval(attrs.barDataLabelsEnabled),
						format: 		attrs.barDataLabelsFormat,
						style:			attrs.barDataLabelsStyle == undefined ? {} : eval('(' + attrs.barDataLabelsStyle + ')'), 
					},
					stacking: 			attrs.barStacking,
				};
			}
			else if(attrs.chartType == "bubble") {
				scope.options.plotOptions.bubble = {
					dataLabels: {
						enabled: 		attrs.bubbleDataLabelsEnabled == undefined ? false : eval(attrs.bubbleDataLabelsEnabled),
						format: 		attrs.bubbleDataLabelsFormat,
					},
				};
			}
			else if(attrs.chartType == "column") {
				scope.options.plotOptions.column = {
					dataLabels: {
						enabled: 		attrs.columnDataLabelsEnabled == undefined ? false : eval(attrs.columnDataLabelsEnabled),
						format: 		attrs.columnDataLabelsFormat,
					},
					stacking: 			attrs.columnStacking,
				};
			}
			else if(attrs.chartType == "line") {
				scope.options.plotOptions.line = {
					dataLabels: {
						enabled: 		attrs.lineDataLabelsEnabled == undefined ? false : eval(attrs.lineDataLabelsEnabled),
						format: 		attrs.lineDataLabelsFormat,
					},
					stacking: 			attrs.lineStacking == undefined ? undefined : eval(attrs.lineStacking),
				};
			}
			else if(attrs.chartType == "heatmap") {
				scope.options.xAxis.categories = scope.categories.xaxis;
				scope.options.yAxis.categories = scope.categories.yaxis;
			}
			else if(attrs.chartType == "pie") {
				scope.options.plotOptions.pie = {
					allowPointSelect:	attrs.pieAllowPointSelect == undefined ? undefined : eval(attrs.pieAllowPointSelect),
					cursor:				attrs.pieCursor,
					dataLabels: {
						enabled:		attrs.pieDataLabelsEnabled == undefined ? false : eval(attrs.pieDataLabelsEnabled),
						format: 		attrs.pieDataLabelsFormat,
					},
					showInLegend:		attrs.pieShowInLegend == undefined ? false : eval(attrs.pieShowInLegend),
					startAngle:			attrs.pieStartAngle == undefined ? undefined : eval(attrs.pieStartAngle),
					endAngle:			attrs.pieEndAngle == undefined ? undefined : eval(attrs.pieEndAngle),
				};
			}
			else if(attrs.chartType == "scatter") {
				scope.options.plotOptions.scatter = {
					dataLabels: {
						enabled: 		attrs.scatterDataLabelsEnabled == undefined ? false : eval(attrs.scatterDataLabelsEnabled),
						format: 		attrs.scatterDataLabelsFormat,
					},
					marker: {
						radius:			attrs.scatterMarkerRadius == undefined ? undefined : eval(attrs.scatterMarkerRadius),
					},
				};
			}
			else if(attrs.chartType == "spline") {
				scope.options.plotOptions.spline = {
					dataLabels: {
						enabled: 		attrs.splineDataLabelsEnabled == undefined ? false : eval(attrs.splineDataLabelsEnabled),
						format: 		attrs.splineDataLabelsFormat,
					},
					stacking: 			attrs.splineStacking == undefined ? undefined : eval(attrs.splineStacking),
				};
			}
			else if(attrs.chartType == "treemap") {
				scope.options.plotOptions.treemap = {
					alternateStartingDirection:	attrs.treemapAlternateStartingDir == undefined ? undefined : eval(attrs.treemapAlternateStartingDir),
					layoutAlgorithm:	attrs.treemapLayoutAlgorithm,
				};
			}


			// Color Scale object
			{
				function ColorScale(scale) {
					this.scale = scale;
					this.lastUsed = -1;
					this.colorMap = {};
				};
				
				ColorScale.prototype.color = function(key) {
					if(this.colorMap[key] != undefined) 
						return this.colorMap[key];
					this.lastUsed = this.lastUsed + 1;
					if(this.lastUsed == this.scale.length) 
						this.lastUsed = 0;
					var color = this.scale[this.lastUsed];
					this.colorMap[key] = color;
					return color;
				};
			}

			// Define color scale object
			var	color_category20a = [
									  '#1f77b4', '#aec7e8', '#ff7f0e', '#ffbb78',
									  '#2ca02c', '#98df8a', '#d62728', '#ff9896',
									  '#9467bd', '#c5b0d5', '#8c564b', '#c49c94',
									  '#e377c2', '#f7b6d2', '#7f7f7f', '#c7c7c7',
									  '#bcbd22', '#dbdb8d', '#17becf', '#9edae5'
									];
			
			scope.colorScale = new ColorScale(color_category20a);


			// Process an object
			scope.processInsertObject = function(object) {
				// Loop over the series defined for the chart
				for(var idx = 0; idx < scope.series.length; idx++) {
					var thisSeries = scope.series[idx];
					var series = chart.get(thisSeries.series);
					var datum = {
						id:		thisSeries.series + "_" + object.tid,
						name:	object[scope.category],
						//x:		scope.options.xAxis.type != 'category' ? object[scope.category] : undefined, 
						y:		object[thisSeries.series],
					};
					
					if(scope.options.xAxis.type != 'category')
						datum.x = object[scope.category];
					
					if(scope.options.chart.type == 'arearange') {
						datum.high = object[thisSeries.high];
						datum.low = object[thisSeries.low];
					}
					else if(scope.options.chart.type == 'bar') {
						if(object[thisSeries.series] > 0)
							datum.color = 'green';
						else
							datum.color = 'red';
					}
					else if(scope.options.chart.type == 'bubble') {
						datum.z = object[thisSeries.radius];
					}
					else if(scope.options.chart.type == 'heatmap') {
						if(scope.categories.xaxis.indexOf(object[scope.category]) == -1) {
							scope.categories.xaxis.push(object[scope.category]);
						}
						if(scope.categories.yaxis.indexOf(object[thisSeries.series]) == -1) {
							scope.categories.yaxis.push(object[thisSeries.series]);
						}
						datum.x = scope.categories.xaxis.indexOf(object[scope.category]);
						datum.y = scope.categories.yaxis.indexOf(object[thisSeries.series]);
						datum.value = object[thisSeries.value];
					}
					else if(scope.options.chart.type == 'treemap') {
						if(thisSeries.parent != undefined) {
							for(var pIdx = 0; pIdx < thisSeries.parent.length; pIdx++) {
								var pArr = scope.parents[pIdx];
								if(pArr == undefined) {
									pArr = {};
									scope.parents[pIdx] = pArr;
								}
								if(pArr[object[thisSeries.parent[pIdx].series]] == undefined) {
									pArr[object[thisSeries.parent[pIdx].series]] = true;
									var pDatum = {
										id:		object[thisSeries.parent[pIdx].series],
										name:	object[thisSeries.parent[pIdx].series],
										color:	thisSeries.parent[pIdx].color,
									}
									if(pIdx < thisSeries.parent.length - 1) {
										pDatum.parent = object[thisSeries.parent[pIdx + 1].series];
									}
									else {
										pDatum.color = scope.colors[Object.keys(scope.parents[pIdx]).length - 1];
									}
									series.addPoint(pDatum, false);
								}
							}
						}
						
						datum.name = object[thisSeries.series];
						datum.x = undefined;
						datum.y = undefined;
						datum.value = object[thisSeries.value];
						datum.colorValue = object[thisSeries.value];
						
						String.prototype.hashCode = function() {
						  var hash = 0, i, chr, len;
						  if (this.length == 0) return hash;
						  for (i = 0, len = this.length; i < len; i++) {
						    chr   = this.charCodeAt(i);
						    hash  = ((hash << 5) - hash) + chr;
						    hash |= 0; // Convert to 32bit integer
						  }
						  return hash;
						};
						
						
						datum.color = scope.colorScale.color(datum.name);
						if(thisSeries.parent != undefined)
							datum.parent = object[thisSeries.parent[0].series];
					}
					
					series.addPoint(datum, false);
				}
			};
			
			scope.processUpdateObject = function(object) {
				for(var idx = 0; idx < scope.series.length; idx++) {
					var thisSeries = scope.series[idx];
					var point = chart.get(thisSeries.series + "_" + object.tid);
					if(point != null) {
						var datum = {
							id:		thisSeries.series + "_" + object.tid,
							name:	object[scope.category],
							//x:		scope.options.xAxis.type != 'category' ? object[scope.category] : undefined, 
							y:		object[thisSeries.series],
						};
						
						if(scope.options.xAxis.type != 'category')
							datum.x = object[scope.category];
						
						if(scope.options.chart.type == 'arearange') {
							datum.high = object[thisSeries.high];
							datum.low = object[thisSeries.low];
						}
						else if(scope.options.chart.type == 'bar') {
							if(object[thisSeries.series] > 0)
								datum.color = 'green';
							else
								datum.color = 'red';
						}
						else if(scope.options.chart.type == 'bubble') {
							datum.z = object[thisSeries.radius];
						}
						else if(scope.options.chart.type == 'heatmap') {
							if(scope.categories.xaxis.indexOf(object[scope.category]) == -1) {
								scope.categories.xaxis.push(object[scope.category]);
							}
							if(scope.categories.yaxis.indexOf(object[thisSeries.series]) == -1) {
								scope.categories.yaxis.push(object[thisSeries.series]);
							}							
							
							datum.x = scope.categories.xaxis.indexOf(object[scope.category]);
							datum.y = scope.categories.yaxis.indexOf(object[thisSeries.series]);
							datum.value = object[thisSeries.value];
						}
						else if(scope.options.chart.type == 'treemap') {
							if(thisSeries.parent != undefined) {
								for(var pIdx = 0; pIdx < thisSeries.parent.length; pIdx++) {
									var pArr = scope.parents[pIdx];
									if(pArr == undefined) {
										pArr = {};
										scope.parents[pIdx] = pArr;
									}
									if(pArr[object[thisSeries.parent[pIdx].series]] == undefined) {
										pArr[object[thisSeries.parent[pIdx].series]] = true;
										var pDatum = {
											id:		object[thisSeries.parent[pIdx].series],
											name:	object[thisSeries.parent[pIdx].series],
											color:	thisSeries.parent[pIdx].color,
										}
										if(pIdx < thisSeries.parent.length - 1) {
											pDatum.parent = object[thisSeries.parent[pIdx + 1].series];
										}
										else {
											pDatum.color = scope.colors[Object.keys(scope.parents[pIdx]).length - 1];
										}
										series.addPoint(pDatum, false);
									}
								}
							}

							datum.name = object[thisSeries.series];
							datum.x = undefined;
							datum.y = undefined;
							datum.value = object[thisSeries.value];
							if(thisSeries.parent != undefined)
								datum.parent = object[thisSeries.parent[0].series];
						}
						
						point.update(datum, false);
					}
					else {
						$log.error('point not found for update: ' + thisSeries.series + "_" + object.tid);
					}
				}						
			};
			
			scope.processDeleteObject = function(object) {
				for(var idx = 0; idx < scope.series.length; idx++) {
					var thisSeries = scope.series[idx];
					var point = chart.get(thisSeries.series + "_" + object.tid);
					if(point != null) {
						point.remove(false);
					}
				}
			};
			
			// Function to call when chart is ready
			scope.chartReady = function(chart) {
				// Create the series from the configuration
				for(var idx = 0; idx < scope.series.length; idx++) {
					var thisSeries = scope.series[idx];
					var seriesOptions = {
						id:		thisSeries.series,
						type:	attrs.chartType,
						name: 	thisSeries.name == undefined ? thisSeries.series : thisSeries.name,
					};
					if(attrs.chartType == 'pie') {
						seriesOptions.innerSize = attrs.pieInnerSize;
					}
					else if(attrs.chartType == 'heatmap') {
						seriesOptions.borderWidth = 1;
					}
					else if(attrs.chartType == 'treemap') {
						seriesOptions.levels = [];
						if(thisSeries.parent != undefined) {
							for(var levIdx = 0; levIdx < thisSeries.parent.length; levIdx++) {
								seriesOptions.levels.push({
									borderWidth: 1,
									level: levIdx + 1,
									dataLabels: {
										enabled: true,
										align: levIdx % 2 == 0 ? 'left' : 'right',
										verticalAlign: 'top',
										style: {
											fontSize: (15 - 2 * levIdx) + 'px',
											fontWeight: 'bold'
										}
									}
								});
							}
						}
					}
					var series = chart.addSeries(seriesOptions, false);
				}
				chart.redraw();
				
				// Listen to snapshot events
				scope.$on('query:' + scope.queryId + ':snapshot', function(event, action, serverEvent, object) {
					var redraw = false;
					if(action == 'change') {
						for(var idx = 0; idx < scope.series.length; idx++) {
							var thisSeries = scope.series[idx];
							var series = chart.get(thisSeries.series);
							series.setData([], false);
						}
						chart.redraw();
					}
					else if(action == 'onSuccess:object') {
						scope.processInsertObject(object);
						redraw = true;
					}
					
					if(redraw == true) {
						// Microbatch handler
						microbatch.rtime = new Date();
						if(microbatch.timeout === false) {
							microbatch.timeout = true;
							setTimeout(eventEnd, microbatch.delta);
						}					
						function eventEnd() {
							if (new Date() - microbatch.rtime < microbatch.delta)
								setTimeout(eventEnd, microbatch.delta);
							else {
								microbatch.timeout = false;
								chart.redraw();
							}               
						};
					}					
				});
							
				// Listen to subscribe events
				scope.$on('query:' + scope.queryId + ':subscribe', function(event, action, serverEvent, object) {
					var redraw = false;
					if(action == 'onSnapshotStart')
						scope.snapshot = true;
					else if(action == 'onInsert') {
						scope.processInsertObject(object);
						redraw = true;
					}
					else if(action == 'onUpdate') {
						scope.processUpdateObject(object);
						redraw = true;
					}
					else if(action == 'onDelete') {
						scope.processDeleteObject(object);
						redraw = true;
					}
					else if(action == 'onSnapshotEnd')
						scope.snapshot = undefined;
					
					if(redraw == true) {
						// Microbatch handler
						microbatch.rtime = new Date();
						if(microbatch.timeout === false) {
							microbatch.timeout = true;
							setTimeout(eventEnd, microbatch.delta);
						}					
						function eventEnd() {
							if (new Date() - microbatch.rtime < microbatch.delta)
								setTimeout(eventEnd, microbatch.delta);
							else {
								microbatch.timeout = false;
								chart.redraw(false);
							}               
						};
					}
				});
					
				// Listen to unsubscribe events
				scope.$on('query:' + scope.queryId + ':unsubscribe', function(event, message, serverEvent) {
					for(var idx = 0; idx < scope.series.length; idx++) {
						var thisSeries = scope.series[idx];
						var series = chart.get(thisSeries.series);
						series.setData([], false);
					}
					chart.redraw();
				});			
			};
			
			// Create the chart and set initial chart size
			var chart = new Highcharts.Chart(scope.options, function() {
				scope.chartReady(this);
				this.setSize(element.parent().width(), element.parent().height());
			});	
						
			// Bind listener on window resize to reflow chart
			EventListener.windowResize(function() {
				if(chart == null) return;
				setTimeout(function(){
					chart.setSize(element.parent().width(), element.parent().height());
				});
			});
			
			// Bind listener on component display for tabs, applications, and visibility-toggle divs
			EventListener.componentDisplay(element, function() {
				if(chart == null) return;
				setTimeout(function(){
					chart.setSize(element.parent().width(), element.parent().height());
				});
			});

			// Display the context menu
			$(element).on("contextmenu", function(e) {
				var container = $(element).parent();
				var menu = $(element).find(".context-menu");
				
				var x = e.pageX - container.offset().left;
				var y = e.pageY - container.offset().top;
				var width = menu.width();
				
				if(e.pageX + width > window.innerWidth)
					x = e.pageX - width - container.offset().left;
				
				$('.context-menu').hide();
				menu.css({
					display: "block",
					left: x,
					top: y
				});

				return false;
			});			
		
			// Display the query details modal
			scope.displayQuery = function() {
				var modal = {
					mode: 		"display",
					query:		scope.query,
					allowEdit:	scope.options.queryEditor,
					onSuccess:	function() {
						// If the table is in a tab, select the tab this table is in on successful
						//   complete of the modal actions
						var tabs = $(element).closest('div.tab-content').prev('ul.nav-tabs');
						if(tabs.length > 0) {
							var showId = tabs.find('.active').find('a[data-toggle="tab"]').attr('href');
							setTimeout(function() {
								var showTab = tabs.find('a[href=' + showId + ']');
								showTab.tab('show');
								showTab.trigger("shown.bs.tab");
							}, 10);
						}
					},
				};

				// Display the nearest modal
				$(element).closest('.application').scope().$broadcast("display.query", modal);
			};		
		}		
	};	
});

app.directive('livetable', function($log, $window, $filter, $compile, SimpleDateFormat, NumberFormat, Overlay, EventListener) {
	return {
		templateUrl: './templates/live-table.html',
		replace: false,
		transclude: true,
		restrict: 'E',
		scope: {
			'query':	'=',
			'filter':	'=',
			'selected':	'='
		},
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;

			// Set the query ID to the scope
			scope.queryId = attrs.queryId;                                                                                
			
			// Read the options from attributes
			scope.options = {
				"uid":				"tid",
									// Mandatory. Object property that is unique and passed into the table for comparison.
									
				"filter":			(attrs.filter != undefined && attrs.filter.length > 0),
									// Optional. Custom data filter enabled. May be true or false.
									
				"selectable":		(attrs.selected != undefined && attrs.selected.length > 0),
									// Optional. User can select chart element and object will be assigned
									//   to this scope variable. May be true or false.
					
				"info": 			attrs.info == undefined ? 
										false : 
										eval(attrs.info),
									// Optional. Show or hide information display field. May be true or false.
										
				"ordering": 		attrs.ordering == undefined ? 
										false : 
										eval(attrs.ordering),
									// Optional. Enable column header ordering. May be true or false.
									
				"order":			attrs.order == undefined || attrs.order == "" || attrs.ordering == "false" ? 
										[['tid', 'asc']] : 
										eval(attrs.order),
									// Optional. Initial sort order to apply if ordering is enabled. Array in the format
									// [[property1, 'asc'|'desc'], [property2, 'asc'|'desc'], ...]
									// Default is always ascending on tid
									
				"orderMulti":		attrs.orderMulti == undefined ? 
										false : 
										eval(attrs.orderMulti),
									// Optional. Allow multiple column ordering if ordering is enabled. May be true or false.
									
				"orderClasses":		attrs.orderClasses == undefined ? 
										false : 
										eval(attrs.orderClasses),
									// Optional. Enable highlighting of ordered columns if ordering is enabled. May be true or false.
									
				"paginationSize": 	attrs.paginationSize == undefined || attrs.paginationSize  == "" ? 
										25 : 
										eval(attrs.paginationSize),
									// Optional. Pagination size row count. Must be an integer.
									
				"queryDetailIcon": 	attrs.queryDetailIcon == undefined ? 
										false : 
										eval(attrs.queryDetailIcon),
									// Optional. Display query detail icon in header. May be true or false.
									
				"queryViewer": 		attrs.queryViewer == undefined ? 
										false : 
										eval(attrs.queryViewer),
									// Optional. Allow query viewer modal. May be true or false.
									
				"queryEditor": 		attrs.queryEditor == undefined ? 
										false : 
										eval(attrs.queryEditor),
									// Optional. Allow query editor modal. May be true or false.
																		
				"searching": 		attrs.searching == undefined ? 
										false : 
										eval(attrs.searching),
									// Optional. Enable table searching. May be true or false.
									
				"searchCols":		attrs.searchCols == undefined || attrs.searchCols == "" ? 
										eval("[]") : 
										eval(attrs.searchCols), 
									// Optional. Array of columns to search on if enabled.
									// [null,{'search':'Sea'},null,{'search':'^[0-9]','escapeRegex':false},null]
									
				"searchDelay":		attrs.searchDelay == undefined ? 
										100 : 
										eval(attrs.searchDelay),
									// Optional but recommended. Delay to wait before executing search in msec if searing is enabled. Must be integer.
										
				"stripeClasses":	attrs.stripeClasses == undefined || attrs.stripeClasses == "" ? 
										false : 
										eval(attrs.stripeClasses),
									// Optional. Apply stripe classes to table. May be true or false.
			};

			// Read the column model options from attributes
			scope.options.columnModels = {};			
			$(element).find('column').each(function(index, columnElement) {
				var name = $(columnElement).attr('name');
				var options = {
					"node":					columnElement,
					
					"classes":				$(columnElement).attr('classes'),
											// Optional. List of classes to apply to cells in this column.

					"conditionalClasses":	$(columnElement).attr('conditional-classes') == undefined || $(columnElement).attr('conditional-classes') == "" ?
												undefined :
												$(columnElement).attr('conditional-classes'),
											// Optional. Conditional classes to apply based on cell value.
					
					"content":				$(columnElement).find('column-content').length != 1 ?
												undefined :
												$(columnElement).find('column-content').html(),
											// Optional. Column content.
										
					"format":				$(columnElement).attr('format'),
											// Optional. Formatting to apply to the data in the column.

					"orderable":			$(columnElement).attr('orderable') == undefined || $(columnElement).attr('orderable') == "" ? 
												true : 
												eval($(columnElement).attr('orderable')),
											// Optional. Allow ordering on this column. May be true or false.

					"searching":			$(columnElement).attr('searching') == undefined || $(columnElement).attr('searching') == "" ? 
												true : 
												eval($(columnElement).attr('searching')),
											// Optional. Allow searching on this column. May be true or false.

					"title":				$(columnElement).attr('title'),
											// Optional. Title to display.

					"visible":				$(columnElement).attr('visible') == undefined || $(columnElement).attr('visible') == "" ? 
												true : 
												eval($(columnElement).attr('visible')),
											// Optional. Show or hide the column. May be true or false.

					"width":				$(columnElement).attr('width') == undefined || $(columnElement).attr('width') == "" ? 
												undefined : 
												$(columnElement).attr('width')
											// Optional. Width of the column. May be number in px or percentage.
				};
				scope.options.columnModels[name] = options;
			});
			
			// Declare a table variable
			var table = null;

			// Declare a dataset collection
			scope.dataset = new Collection(['tid']);
			
			// Setup default sort order
			scope.order = [];
			
			// Setup pagination for infinite scrolling
			scope.pagination = {
				size:	scope.options.paginationSize,
				begin: 	0,
				end:	scope.options.paginationSize
			};
			
			// Setup an overlay watching the query status
			Overlay.create(scope, scope.queryId, element);
			
			// Listen to query status messages for schema definition and display the table structure
			scope.$on('query:' + scope.queryId + ':status', function(event, query) {
				if(query.schema == undefined) return;

				// If the table is already defined, just return
				if(table != null) return;

				// Create the table
				table = createTable();
				
				// Setup the default ordering
				scope.order = [];
				
				// Get the configured sort for defaults
				var initialOrder = {};
				if(scope.options.ordering == true) {
					var len = scope.options.orderMulti == false ? 1 : scope.options.order.length;
					for(var idx = 0; idx < len; idx++)
						initialOrder[scope.options.order[idx][0]] = scope.options.order[idx][1];
				}
				
				// Setup the column models
				scope.columnModels = [];
				for(var idx = 0; idx < query.schema.fields.length; idx++) {
					var fieldName = query.schema.fields[idx].name;
					var columnModelOptions = scope.options.columnModels[fieldName];
					var columnModel = new ColumnModel(query.schema.fields[idx], columnModelOptions, initialOrder[query.schema.fields[idx].name]);					
					scope.columnModels.push(columnModel);
				}

				// Apply the default configured ordering
				if(scope.options.ordering == true) {
					var len = scope.options.orderMulti == false ? 1 : scope.options.order.length;
					for(var idx = 0; idx < len; idx++) {
						for(var col = 0; col < scope.columnModels.length; col++) {
							if(scope.columnModels[col].field.name == scope.options.order[idx][0])
								scope.order.push(scope.columnModels[col]);
						}
					}
				}
				
				scope.$apply();
				
				// Draw the table
				table.draw();
			});
			
			// Listen to snapshot events
			scope.$on('query:' + scope.queryId + ':snapshot', function(event, action, serverEvent, object) {
				if(action == 'change') {
					scope.dataset.removeAll();
					if(table == null) return;
					table.update();
					table = null;
				}
				else if(action == 'onSuccess:object') {
					// Add the new row to the dataset
					if(table == null) return ;
					var row = createRow(object);
					scope.dataset.add(row);
					table.update();
				}
			});
				
			// Listen to subscribe events
			scope.$on('query:' + scope.queryId + ':subscribe', function(event, action, serverEvent, object) {
				if(action == 'onSnapshotStart')
					scope.snapshot = true;
				else if(action == 'onInsert') {						
					// Add the new row to the dataset
					if(table == null) return;
					var row = createRow(object);
					scope.dataset.add(row);
					table.update();
				}
				else if(action == 'onUpdate') {
					// Get the row from the dataset and update the row (no need to update the table)
					if(table == null) return ;
					var row = scope.dataset.get(object.tid);
					row.data = object;
					row.update();
				}
				else if(action == 'onDelete') {
					// Remove the row from the dataset
					if(table == null) return ;
					scope.dataset.remove(object.tid);
					table.update();
				}
				if(action == 'onSnapshotEnd')
					scope.snapshot = undefined;
			});
				
			// Listen to unsubscribe events
			scope.$on('query:' + scope.queryId + ':unsubscribe', function(event, message, serverEvent) {
				scope.dataset.removeAll();
				table.update();
				table = null;
			});
					
			// Watch the search box to mark as filtered
			scope.$watch('search', function() {
				if(table != null) table.update();
			});
			
			// Watch the column searches to mark as filtered
			scope.$watch(function() {
				return isFiltered();
			}, function(newVal, oldVal) {
				scope.filtered = newVal;
			});
			
			// Watch the column searches to update the table
			scope.$watch(function() {
				if(scope.columnModels == undefined) return false;
				var s = "";
				for(var idx = 0; idx < scope.columnModels.length; idx++) {			
					var column = scope.columnModels[idx];
					s = s + column.search;
				}
				return s;
			}, function(newVal, oldVal) {
				if(table != null) table.update();
			});
			
			// Watch the external filter to mark as filtered
			scope.$watch('filter', function() {
				if(table != null) table.update();
			});
			
			// Watch the scroll bar to change the pagination
			$(element).find('.livetable-table-body').scroll(function(event) {
				var scrollPercentage = $(this).scrollTop() / ($(element).find('.livetable-table-body').find('.livetable-structure-table').height() - $(this).height());
				if(scrollPercentage >= 0.75 && scope.results.length < scope.dataset.length()) {
					scope.pagination.end = scope.pagination.end + scope.pagination.size;
					table.update();
				}
			});
			
			// Select an element
			//   Data Internal > External
			scope.select = function(row) {
				if(scope.options.selectable == false) return;
				
				if(scope.selectedRow == undefined) {
					scope.selectedRow = row;
					row.selected = true;
					row.update();
				}
				else {
					if(scope.selectedRow == row) {
						scope.selectedRow = undefined;
						row.selected = false;
						row.update();
					}
					else {
						var previous = scope.selectedRow;
						previous.selected = false;
						previous.update();
						scope.selectedRow = row;
						row.selected = true;
						row.update();
					}
				}
				scope.$emit("table:onSelect", scope.selectedRow);				
				scope.$emit("table:" + scope.elementId + ":onSelect", scope.selectedRow);				
			};
			
			// Watch for selection to update row classes
			//   Data External > Internal
			scope.$watch('selected.data', function(newVal, oldVal) {
				if(newVal == undefined) {
					var row = scope.selectedRow;
					scope.selectedRow = undefined;
					if(row != undefined) {
						row.selected = false;
						row.update();
					}
				}
				else {
					var previous = scope.selectedRow;
					var row = scope.dataset.get(newVal.tid);

					if(previous == undefined && row == undefined) {
					}
					else if(previous == undefined && row != undefined) {
						scope.selectedRow = row;
						row.selected = true;
						row.update();
					}
					else if(previous != undefined && row == undefined) {
						scope.selectedRow = undefined;
						previous.selected = false;
						previous.update();
					}
					else if(previous != undefined && row != undefined) {
						scope.selectedRow = undefined;
						previous.selected = false;
						previous.update();

						scope.selectedRow = row;
						row.selected = true;
						row.update();
					}						
				}
			});
			
			// Display the header context menu
			scope.headerMenu = function(column, e) {
				var container = $(element).parent();
				var menu = $(element).find(".header-context-menu");
				//var offsetLeft = container.offset().left;
				//var offsetTop = container.offset().top + 55; 
				var offsetLeft = 0;
				var offsetTop = 0;
				
				var x = e.pageX - offsetLeft;
				var y = e.pageY - offsetTop;
				var width = menu.width();
				
				if(e.pageX + width > window.innerWidth)
					x = e.pageX - width - offsetLeft;
				
				// Timeout required to render the menu properly
				setTimeout(function() {
					$(element).find('.context-menu').css({display: 'none'});
					menu.css({ position: 'fixed', left: x, top: y, display: 'block' });
					scope.context = {column:column};
				});
				
				return false;
			};			
			
			// Clear all ordering
			scope.clearOrdering = function() {
				for(var idx = 0; idx < scope.columnModels.length; idx++) {
					scope.columnModels[idx].ordering = undefined;
				}
				scope.order = [];
				table.update();
			};
			
			// Clear all filters
			scope.clearSearches = function() {
				scope.search = undefined;
				for(var idx = 0; idx < scope.columnModels.length; idx++) {
					scope.columnModels[idx].search = undefined;
				}
				table.update();
			};
			
			// Toggle field searches
			scope.toggleFieldSearches = function() {
				if(scope.searchesVisible == undefined)
					scope.searchesVisible = true;
				else
					scope.searchesVisible = undefined;

				// Timeout to re-draw the table correctly
				setTimeout(function() {
					table.draw();
				},10);
			};
			
			// Display the query details modal
			scope.displayQuery = function() {
				var modal = {
					mode:		"display",
					query:		scope.query,
					allowEdit:	scope.options.queryEditor,
					onSuccess:	function() {
						// If the table is in a tab, select the tab this table is in on successful
						//   complete of the modal actions
						var tabs = $(element).closest('div.tab-content').prev('ul.nav-tabs');
						if(tabs.length > 0) {
							var showId = tabs.find('.active').find('a[data-toggle="tab"]').attr('href');
							setTimeout(function() {
								var showTab = tabs.find('a[href=' + showId + ']');
								showTab.tab('show');
								showTab.trigger("shown.bs.tab");
							}, 10);
						}
					},
				};

				// Display the nearest modal
				$(element).closest('.application').scope().$broadcast("display.query", modal);
			};
			
			// Display the body context menu
			scope.bodyMenu = function(object, column, e) {
				var container = $(element).parent();
				var menu = $(element).find(".body-context-menu");

				var x = e.pageX - container.offset().left;
				var y = e.pageY - container.offset().top;
				var width = menu.width();
				
				if(e.pageX + width > window.innerWidth)
					x = e.pageX - width - container.offset().left;
				
				$(element).find('.context-menu').css({display: 'none'});
				menu.css({ left: x, top: y, display: 'block' });
				scope.context = {object: object, column:column };
				
				return false;
			};
			
			// Bind listener on window resize to adjust table heights for auto-sized tables
			EventListener.windowResize(function() {
				if(table == null) return;
				setTimeout(function(){table.draw();});
			});
			
			// Bind listener on component display for tabs, applications, and visibility-toggle divs
			EventListener.componentDisplay(element, function() {
				if(table == null) return;
				setTimeout(function(){table.draw();});
			});						

			// Create a new table instance
			var createTable = function() {
				var table = new Table();
				scope.tableBody = $(element).find('.livetable-table-body').find('.livetable-structure-table');				
				return table;
			};
			
			// Create a new row instance
			var createRow = function(object) {
				var row = new TableRow(object);
				
				// Loop over the column models and create a table cell instance for each column in this row
				for(var idx = 0; idx < scope.columnModels.length; idx++) {
					var columnModel = scope.columnModels[idx];
					var tableRowCell = new TableRowCell(columnModel, object, row);
					row.cells.push(tableRowCell);					
				}
				return row;				
			};
			
			// Determine if the table is filtered by any of search, column filter, or external filter object
			var isFiltered = function() {
				if(!(scope.search == undefined || scope.search == "") == true) return true;
				if(scope.columnModels != undefined) {
					for(var idx = 0; idx < scope.columnModels.length; idx++) {			
						var column = scope.columnModels[idx];
						if(column.search != undefined && column.search.length > 0) {
							return true;
						}
					}
				}
				if(!(scope.filter == undefined || Object.keys(scope.filter).length == 0) == true) return true;
				return false;
			}

			// DEFINE OBJECTS			
			// Column model object
			{
				function ColumnModel(field, options, ordering) {
					this.field = field;
					this.options = options == undefined ? {} : options;
					this.ordering = ordering;
					this.$compile = undefined;
					
					if(this.options == undefined) this.options = {};
					
					if(this.options.content != undefined) {
						this.$compile = $compile(this.options.content);
					}
				};
				
				ColumnModel.prototype.bodyClasses = function(object, data) {
					var c = "";
					if(scope.options.ordering == true && scope.options.orderClasses == true && this.ordering != undefined)
						c = c + " ordered";
					if(this.options.classes != undefined)
						c = c + ' ' + this.options.classes;
					if(this.options.conditionalClasses != undefined) {											
						var cond = eval('(' + this.options.conditionalClasses + ')');
						for(var idx = 0; idx < Object.keys(cond).length; idx++) {
							var cls = Object.keys(cond)[idx];
							if(cond[cls] == true) c = c + ' ' + cls;
						}
					}
					return c.trim();
				};
				
				ColumnModel.prototype.content = function(context, node, object, data) {
					var self = this;

					if(this.options.content == undefined) {
						var display = this.format(data);
						var nodeContent = display;
						node.html(nodeContent); 
					}
					else {
						var content = context.$compile(context.contentScope, function(clonedElement, scope) {
							// This allows a delay to permit the template to render without flickering in the UI
							scope.$evalAsync(function() {
								// Optimization will only update the cell if the HTML content has changed
								if(node.children().first()[0] == undefined || node.children().first()[0].innerHTML != clonedElement[0].innerHTML) {
									node.empty().html(clonedElement);
								}
							});
						});
					}
				};
			
				ColumnModel.prototype.headerClasses = function() {
					var c = "";
					if(scope.options.ordering == true && this.options.orderable != false) {
						c = c + " orderable";
						if(this.ordering == 'asc') c = c + ' ordered-asc';
						else if(this.ordering == 'desc') c = c + ' ordered-desc';
						else if(this.ordering == undefined) c = c + ' unordered';
					}
					return c.trim();
				};
				
				ColumnModel.prototype.format = function(data) {
					if(this.options.format != undefined) {
						if(this.field.type == "timestamp") return SimpleDateFormat.format(data, this.options.format);
						if(this.field.type == "double") return NumberFormat.format(data, this.options.format);
					}
					return data;
				};
				
				ColumnModel.prototype.order = function() {
					if(scope.options.ordering == false || this.options.orderable == false) return;
					if(this.ordering == 'asc') {
						this.ordering = 'desc';
						if(scope.options.orderMulti == false) {
							if(scope.order[0].field.name != this.field.name)
								scope.order[0].ordering = undefined;
							scope.order[0] = this;				
						}
					}
					else if(this.ordering == 'desc') { 
						this.ordering = undefined;
						var index = scope.order.indexOf(this);
						if(scope.options.orderMulti == false) {
							if(scope.order[0].field.name != this.field.name)
								scope.order[0].ordering = undefined;
							scope.order[0] = this;
						}
						else
							scope.order.splice(index, 1);
					}
					else if(this.ordering == undefined) {
						this.ordering = 'asc';
						if(scope.options.orderMulti == false && scope.order.length > 0) {
							if(scope.order[0].field.name != this.field.name)
								scope.order[0].ordering = undefined;
							scope.order[0] = this;
						}
						else
							scope.order.push(this);
					}
					table.update();
				};
				
				ColumnModel.prototype.orderIndex = function() { return scope.order.indexOf(this); };
				ColumnModel.prototype.style = function() { return this.options.width == undefined ? {} : {'width': this.options.width}; };
				ColumnModel.prototype.title = function() { if(this.options.title == undefined) return this.field.name; else return this.options.title; };
			}
		
			// Table object
			{
				function Table() {
					// These properties are used for microbatching the updates
					this.microbatch = {
						rtime:		new Date(1, 1, 2000, 12,00,00),
						timeout:	false,
						delta:		0
					}				
				};
				
				Table.prototype.draw = function() {
					// Reset the body height
					var bodyHeight = $(element).find('.livetable-container').height() - $(element).find('.livetable-header').height() - $(element).find('.livetable-footer').height() - 5;				
					$(element).find('.livetable-body').height(bodyHeight);
					
					// Reset the table body height
					var tableHeaderHeight = $(element).find('.livetable-table-header').height();
					if(tableHeaderHeight == 0) tableHeaderHeight = 25; // First draw stupidity
					$(element).find('.livetable-table-body').height(bodyHeight - tableHeaderHeight);
				};
				
				Table.prototype.update = function() {
					var self = this;
					
					// Microbatch handler
					self.microbatch.rtime = new Date();
					if(self.microbatch.timeout === false) {
						self.microbatch.timeout = true;
						setTimeout(eventEnd, self.microbatch.delta);
					}					
					function eventEnd() {
						if (new Date() - self.microbatch.rtime < self.microbatch.delta)
							setTimeout(eventEnd, self.microbatch.delta);
						else {
							self.microbatch.timeout = false;
							run();
						}               
					};
					
					// Run the update
					function run() {
						// Get the scroll top
						var scrollTop = $(element).find('.livetable-table-body').scrollTop();
						
						// Apply filters
						scope.$applyAsync(function() {
							scope.filterset = scope.dataset.toArray();
							scope.filterset = $filter('objectOrderBy')(scope.filterset, scope.order, scope.snapshot);
							scope.filterset = $filter('objectFilter')(scope.filterset, scope.search, scope.options.searchCols, scope.filter, scope.columnModels);
							scope.results = $filter('objectLimitTo')(scope.filterset, scope.pagination);
						
							// Detach existing and reattach in correct order
							scope.tableBody.children('.livetable-structure-row-data').detach();						
							for(var idx = 0; idx < scope.results.length; idx++) {
								// Render the nodes. If the node is already rendered, it'll just be updated.  
								//   We do a lazy render only when row is displayed, so always need to render here.
								//   Append node in correct location based on new filtered order.
								scope.results[idx].render(scope.tableBody);
								scope.tableBody.append(scope.results[idx].node);							
							}

							// Reset the scroll top
							$(element).find('.livetable-table-body').scrollTop(scrollTop);
						});
					};
				};
			}
			
			// TableRow object
			{
				function TableRow(object) {
					this.tid = object.tid;
					this.data = object;
					this.node = undefined;
					this.odd = false;
					this.cells = [];
					this.selected = false;
				};
				
				TableRow.prototype.render = function() {
					if(this.node == undefined) {
						// Render the node and cells
						this.node = $(document.createElement('div')).attr('id', this.tid);
						for(var idx = 0; idx < this.cells.length; idx++)
							this.cells[idx].render();
						
						// Register event listeners
						var self = this;
						this.node.on('click', null, this, function(event) {
							scope.select(self);
						});
					}
					this.update();
				};
				
				TableRow.prototype.rowClasses = function() {
					var c = "livetable-structure-row livetable-structure-row-data";
					if(scope.options.stripeClasses == true && this.odd == true)
						c = c + " stripe-odd";
					else if(scope.options.stripeClasses == true && this.even == true)
						c = c + " stripe-even";
					if(scope.options.selectable == true)
						c = c + " selectable";
					if(scope.selectedRow != undefined && this.selected == true)
						c = c + " selected"
					return c.trim();
				};
				
				TableRow.prototype.update = function() {
					if(this.node == undefined) return;
					this.node.attr('class', this.rowClasses());
					for(var idx = 0; idx < this.cells.length; idx++)
						this.cells[idx].update();
				};
			}
			
			// TableRowCell object
			{
				function TableRowCell(columnModel, object, row) {
					this.columnModel = columnModel;
					this.object = object;
					this.row = row;
					this.node = undefined;
					this.context = {};
					this.hasComplexContent = this.columnModel.$compile != undefined;

					if(this.hasComplexContent) {
						this.context.contentScope = scope.$new(true);
						this.context.contentScope.Math = window.Math;
						this.context.contentScope.formatDateTime = function(dateStr, format) {
							return SimpleDateFormat.format(dateStr, format);
						};
						this.context.contentScope.formatNumber = function(numStr, format) {
							return NumberFormat.format(numStr, format);
						};
						this.context.$compile = this.columnModel.$compile;
					}
				};
				
				TableRowCell.prototype.cellClasses = function() {
					return 'body-cell datatype-' + this.columnModel.field.type + ' table-' + scope.elementId + '-' + this.columnModel.field.name + 
								' query-' + scope.queryId + '-' + this.columnModel.field.name + ' ' + this.columnModel.bodyClasses(this.object, this.object[this.columnModel.field.name]);
				};
				
				TableRowCell.prototype.render = function() {
					if(this.columnModel.options.visible == false) return;
					this.node = $(document.createElement('div'));
					this.row.node.append(this.node);
				};
				
				TableRowCell.prototype.update = function() {
					if(this.node == undefined) return;
					this.node.attr('class', this.cellClasses()).css(this.columnModel.style());
				
					// Update the contentScope if it has complex content
					if(this.hasComplexContent) {
						var display = this.columnModel.format(this.object[this.columnModel.field.name]);
						this.context.contentScope.object = this.object;
						this.context.contentScope.data = this.object[this.columnModel.field.name];
						this.context.contentScope.display = display;
					}
				
					this.columnModel.content(this.context, this.node, this.object, this.object[this.columnModel.field.name]);
				};
			}
		}
	};	
});

app.directive('closeQueryIcon', function($log, $window) {
	return {
		template: '<div class="close-icon" ng-click="close();"></div>',
		restrict: 'E',
		scope: {
			'connection':	'=',
			'query':		'='
		},
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			scope.close = function() {
				var tabs = $(element).closest('ul.nav-tabs');
				var showId = null;

				// If there are tabs, determine which is the next tab to display
				if(tabs.length > 0) {
					var closing = $(element).closest('a');
					var active = $(element).closest('ul').children('li.active').children('a');					
					if(closing.attr('href') != active.attr('href'))
						showId = active.attr('href');
					else {
						showId = $(element).closest('li').prev().children('a').attr('href');						
						if(showId == undefined)
							showId = $(element).closest('li').next().children('a').attr('href');
					}
				}
				
				// Remove the query 
				scope.connection.remove(scope.query);

				// If there are tabs, display the correct tab
				//   The timeouts are here because Bootstrap is shit.
				if(tabs.length > 0) {
					setTimeout(function() {
						$(element).closest('ul.nav-tabs').children('li.active').removeClass('active');
						
						setTimeout(function() {						
							var showTab = tabs.find('a[href=' + showId + ']');
							showTab.tab('show');
							showTab.trigger("shown.bs.tab");
						}, 10);
					}, 10);
				}
			};
		}
	};	
});

app.directive('addQueryIcon', function($log, $window) {
	return {
		template: '<div class="add-icon" ng-click="add();"></div>',
		restrict: 'E',
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			// Function to call when clicked
			scope.add = function() {
				// Define a function to call when a query is successfully defined
				//   This will select the tab for the newly created query
				var modal = {
					mode:		"add",
					onSuccess: 	function() {
						var tabs = $(element).closest('ul.nav-tabs');
						if(tabs.length > 0) {
							setTimeout(function() {
								var showId = $(element).closest('ul.nav-tabs').find('a[data-toggle="tab"]').last().attr('href');
								var showTab = tabs.find('a[href=' + showId + ']');
								showTab.tab('show');
								showTab.trigger("shown.bs.tab");
							}, 10);
						}
					},
				};
				
				// Show the modal
				$(element).closest('.application').scope().$broadcast("display.query", modal);
			};
		}
	};	
});

app.directive('queryTypeIcon', function($log, $window) {
	return {
		template: '<div class="query-type-icon" ng-click="showQuery();"></div>',
		restrict: 'E',
		scope: {			
			'query': '=',	// Associated with the icon
			'view':	'=',	// Allow view
			'edit':	'=',	// Allow editing
		},
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			// Display the modal
			scope.showQuery = function() {
				if(scope.view == true || scope.edit == true) {
					var modal = {
						mode: 		"display",
						query: 		scope.query,
						allowEdit:	scope.edit,
						onSuccess:	function() {
							var tabs = $(element).closest('ul.nav-tabs');
							if(tabs.length == 0)
								tabs = $(element).closest('div.tab-content').prev('ul.nav-tabs');

							if(tabs.length > 0) {
								var showId = tabs.find('.active').find('a[data-toggle="tab"]').attr('href');
								setTimeout(function() {
									var showTab = tabs.find('a[href=' + showId + ']');
									showTab.tab('show');
									showTab.trigger("shown.bs.tab");
								}, 10);
							}
						},
					};

					// Display the nearest modal
					$(element).closest('.application').scope().$broadcast("display.query", modal);
				}
			};

			// Watch the query type and set the icon for the correct type
			scope.$watch('query.type', function() {
				if(scope.query == undefined || scope.query.type == undefined)
					return;
				
				if(scope.query.type == QueryType.SUBSCRIPTION)
					$(element).find('div').addClass('query-type-subscription-icon');
				else if(scope.query.type == QueryType.SNAPSHOT)
					$(element).find('div').addClass('query-type-snapshot-icon');

				if(scope.view == true || scope.edit == true) {
					if(scope.query.type == QueryType.SUBSCRIPTION)
						$(element).find('div').addClass('query-type-subscription-icon-viewer');
					else if(scope.query.type == QueryType.SNAPSHOT)
						$(element).find('div').addClass('query-type-snapshot-icon-viewer');
				}			
			});
		}
	};	
});

app.directive('queryModal', function($log, $window) {
	return {
		templateUrl: './templates/query-modal.html',
		restrict: 'E',
		scope: {
			'connection':	'=',
		},
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {			
			// Listen for display events
			scope.$on('display.query', function(event, modal) {
				scope.modal = modal;
				$(element).find('.modal').modal('show');
			});
			
			// On show of the modal
			$(element).find('.modal').on('show.bs.modal', function(event){
				scope.activate();
			});	

			// On hide of the modal, reset the external modal scope parameters
			$(element).find('.modal').on('hidden.bs.modal', function(event){
				scope.reset();
			});

			// Watch for changes in the modal mode
			scope.$watch('modal.mode', function() {
				scope.activate();
			});
			
			// Activate the modal
			scope.activate = function() {
				if(scope.modal == undefined || scope.modal.mode == undefined)
					return;
					
				if(scope.modal.mode == 'add')
					scope.modal.title = "Add Query";
				else if(scope.modal.mode == 'display')
					scope.modal.title = "Query Definition";
				else if(scope.modal.mode == 'edit')
					scope.modal.title = "Edit Query";							
			};	
			
			// Reset the modal
			scope.reset = function() {
				scope.modal.title = undefined;
				scope.modal.mode = undefined;
				scope.modal.query = undefined;
			};
		}
	};	
});

app.directive('queryEditorModal', function($log, $window) {
	return {
		templateUrl: './templates/query-editor.html',
		restrict: 'E',
		scope: false,	// This directive is using parent scope
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			// Initialize list of tables to display
			scope.tables 			= {};
			
			// Initialize query to add
			scope.query = {
				id: 				undefined,
				table: 				undefined,
				name:				undefined,
				select:				undefined,
				predicate:			undefined,
				template:			undefined
			};
			
			// Initialize schema and parameters
			scope.schema 			= undefined;
			scope.parameters 		= new Collection(['name']);
			
			// Has the user tried to submit the modal
			scope.isModalSubmit 	= false;
			
			// Query and fields valid state
			scope.isQueryValid 		= null;
			scope.isFormDataValid 	= null;
			
			// Hide the schema table initially
			$(element).find('.schema-table-group').hide();			

			// On show of the modal, set initial parameters
			$(element).closest('.modal').on('shown.bs.modal', function(event){
				scope.activate();
			});

			// On hide of the modal, reset the scope parameters
			$(element).closest('.modal').on('hidden.bs.modal', function(event){
				scope.reset();
			});

			// Activate the panel
			scope.activate = function() {
				// Reset the validation variables
				scope.isModalSubmit 		= false;
				scope.isFormDataValid 		= null;
				
				// List the tables
				var key = new Date().getTime() + "." + Math.random();
				var watcher = scope.$on('connection:' + key + ':listTables', function(event, status, serverEvent) {
					serverEvent.tables.sort(function(a, b){
						if(a.name > b.name) return 1;
						if(a.name < b.name) return -1;
						return 0;
					});
					for(var i = 0 ; i < serverEvent.tables.length; i++) {
						var table = serverEvent.tables[i];
						scope.tables[table.name] = table;
					}
					scope.$apply();				
					watcher();
				});
				scope.connection.listTables(key, true);
			};
		
			// Reset the panel
			scope.reset = function() {
				// Reset scope variables
				scope.tables = {};
				scope.query = {
					id: 			undefined,
					table: 			undefined,
					name:			undefined,
					select:			undefined,
					predicate:		undefined,
					template:		undefined
				};
				scope.schema 		= undefined;
				scope.parameters 	= new Collection(['name']);
				
				scope.isModalSubmit 	= false;
				
				scope.isQueryValid 		= null;
				scope.isFormDataValid 	= null;

				// Hide the schema table and hide it
				$(element).find('.schema-table-group').hide();

				// Remove all form validations
				scope.removeValidationError();				
			};
					
			// Watch for mode change
			scope.$watch('modal.mode', function() {
				if(scope.modal == undefined || scope.modal.mode == undefined)
					return;
				
				if(scope.modal.mode == 'add' || scope.modal.mode == 'edit')
					scope.activate();
					
				if(scope.modal.mode == 'edit') {
					var elements = scope.modal.query.template.split(new RegExp('FROM', 'i'));
					var select = elements[0].split(new RegExp('SELECT', 'i'))[1].trim();
					var predicateIdx = elements[1].trim().indexOf(' ');
					var tableName = predicateIdx == -1 ? elements[1].trim() : elements[1].substring(0, predicateIdx + 1).trim();
					var predicate = predicateIdx == -1 ? null : elements[1].substring(predicateIdx + 1).trim();

					// Setup the parameters first
					if(scope.modal.query.parameters != undefined) {
						for(var idx = 0; idx < Object.keys(scope.modal.query.parameters).length; idx++) {
							var key = Object.keys(scope.modal.query.parameters)[idx];
							scope.parameters.add({
								name: key,
								value: scope.modal.query.parameters[key]
							});
						}
					}
					
					// Setup the query
					scope.query = {
						id: 			scope.modal.query.id,
						table: 			scope.tables[tableName],
						type:			scope.modal.query.type,
						name:			scope.modal.query.name,
						select:			select,
						predicate:		predicate,
						template:		scope.modal.query.template
					};
					
					var watcher = scope.$watch('tables', function() {
						if(scope.tables == undefined) return;
						scope.query.table = scope.tables[tableName];
						if(scope.query.table != undefined)
							watcher();
					}, true);
				}
			});
			
			// On change of the table
			scope.$watch('query.table', function() {
				if(scope.query.table == null) return;				

				// Show and set the schema for the table
				$(element).find('.schema-table-group').show();
				scope.schema = scope.query.table.schema;

				// Update the template
				scope.updateTemplate();
				
				// Validate the form data
				scope.isFormDataValid = scope.validateFormData();
			});

			// On change of the query id
			scope.$watch('query.id', function() {
				if(scope.query.id == null) return;				
				scope.isFormDataValid = scope.validateFormData();
			});
			
			// On change of the select clause
			scope.$watch('query.select', function() {
				if(scope.query.select == null) return;				
				scope.updateTemplate();
				scope.isQueryValid = null;
				scope.isFormDataValid = scope.validateFormData();
			});
			
			// On change of the predicate clause
			scope.$watch('query.predicate', function() {
				if(scope.query.predicate == null) return;				
				scope.updateTemplate();
				scope.isQueryValid = null;
				scope.isFormDataValid = scope.validateFormData();
			});
			
			// On change of isFormDataValue
			scope.$watch('isFormDataValid', function() {
				if(scope.isQueryValid == true && scope.isFormDataValid == true)
					$(element).find('.btn-submit').removeAttr('disabled');
				else
					$(element).find('.btn-submit').attr('disabled','disabled');
			});
			
			// On change of isQueryValid
			scope.$watch('isQueryValid', function() {
				if(scope.isQueryValid == true && scope.isFormDataValid == true)
					$(element).find('.btn-submit').removeAttr('disabled');
				else
					$(element).find('.btn-submit').attr('disabled','disabled');
			});
			
			// Watch for schema double click event
			scope.$on('table:schema-table:onDblClick', function(event, data) {
				if(scope.query.select == undefined || scope.query.select == null || scope.query.select.length == 0)
					scope.query.select = data.name;
				else
					scope.query.select = scope.query.select + ", " + data.name;
				scope.isFormDataValid = scope.validateFormData();
			});
			
			// Add validation error to element
			scope.addValidationError = function(elementId, text) {
				scope.removeValidationError(elementId);
				var group = $(element).find(elementId);
				var control = group.find('.form-control');
				group.addClass('form-error');
				control.tooltip({
					animation:	false,
					placement:	'right',
					title:		text,
					trigger:	'manual'
				});
				control.tooltip('show');
			};

			// Remove validation error from element
			scope.removeValidationError = function(elementId) {
				// Remove all form errors
				if(elementId == undefined) {
					$(element).find('.form-error').removeClass('form-error');
					$(element).find('.form-control').tooltip('destroy');
				}
				// Remove specific element error
				else {
					var group = $(element).find(elementId);
					var control = group.find('.form-control');
					group.removeClass('form-error');
					control.tooltip('destroy');
				}
			};
			
			// Create new subscription query
			scope.subscribe = function() {
				scope.isModalSubmit = true;
				scope.isFormDataValid = scope.validateFormData();
				if(scope.isFormDataValid == false) return;
				
				// Add the query to the connection and hide the modal
				scope.query.parameters = scope.getParameters();
				scope.connection.subscribe(scope.query);
				$(element).closest('.modal').modal('hide');
				
				// There might be a callback in case of success, if so, call that function
				if(scope.modal.onSuccess != undefined)
					scope.modal.onSuccess();
			};
			
			// Create new snapshot query
			scope.snapshot = function() {
				scope.isModalSubmit = true;
				scope.isFormDataValid = scope.validateFormData();
				if(scope.isFormDataValid == false) return;

				// Add the query to the connection and hide the modal
				scope.query.parameters = scope.getParameters();
				scope.connection.snapshot(scope.query);
				$(element).closest('.modal').modal('hide');
				
				// There might be a callback in case of success, if so, call that function
				if(scope.modal.onSuccess != undefined)
					scope.modal.onSuccess();
			};
			
			// Update the existing query
			scope.update = function() {
				scope.isModalSubmit = true;
				scope.isFormDataValid = scope.validateFormData();
				if(scope.isFormDataValid == false) return;

				// If query is a subscription
				if(scope.query.type == QueryType.SUBSCRIPTION) {				
					// Create a watcher that will fire when the old query is unsubscribed
					var watcher = scope.$on('query:' + scope.query.id + ':unsubscribe', function() {
						// Subscribe the updated query, unregister the watcher, hide the modal
						scope.query.parameters = scope.getParameters();
						scope.connection.subscribe(scope.query);
						watcher();
						$(element).closest('.modal').modal('hide');
						
						// There might be a callback in case of success, if so, call that function
						if(scope.modal.onSuccess != undefined)
							scope.modal.onSuccess();						
					});
					
					// Unsubscribe the existing query
					scope.connection.unsubscribe(scope.modal.query);
				}
				// Otherwise if query is a snapshot
				else if(scope.query.type == QueryType.SNAPSHOT) {
					// Execute the updated query to the connection and hide the modal
					scope.query.parameters = scope.getParameters();
					scope.connection.snapshot(scope.query);
					$(element).closest('.modal').modal('hide');
					
					// There might be a callback in case of success, if so, call that function
					if(scope.modal.onSuccess != undefined)
						scope.modal.onSuccess();				
				}				
			};
			
			// Validate the form data
			scope.validateFormData = function() {
				// Wait until the user tries to submit the modal or validate the fields
				if(scope.isModalSubmit == false) return null;
					
				var valid = true;
				if(scope.query.table == undefined || scope.query.table == null) {
					valid = false;
					scope.addValidationError('#addQueryTableGroup', 'Select a Table');
				}
				else
					scope.removeValidationError('#addQueryTableGroup');
				
				if(scope.query.id == undefined || scope.query.id == null || scope.query.id.length == 0) {
					valid = false;
					scope.addValidationError('#addQueryIdGroup', 'Enter a Query ID');
				}
				else if(scope.modal.mode == 'add' && scope.connection.getQuery(scope.query.id) != undefined) {
					valid = false;
					scope.addValidationError('#addQueryIdGroup', 'Query ID already exists');
				}				
				else
					scope.removeValidationError('#addQueryIdGroup');
					
				if(scope.query.select == undefined || scope.query.select == null || scope.query.select.length == 0) {
					valid = false;									
					scope.addValidationError('#addQuerySelectGroup', 'Enter fields to Select');
				}
				else 
					scope.removeValidationError('#addQuerySelectGroup');
				
				return valid;
			};

			// Validate the query
			scope.validateQuery = function() {
				scope.isModalSubmit = true;
				scope.isFormDataValid = scope.validateFormData();
				if(scope.isFormDataValid == false) return;

				// Execute the snapshot and get the results
				scope.validationQuery = {
					id: scope.query.id + "_validation",
					name: scope.query.name,
					template: scope.query.template,
					parameters: scope.getParameters()
				};
				scope.connection.snapshot(scope.validationQuery, false);		

				// Start a listener for the validation result
				var listener = scope.$on('query:' + scope.validationQuery.id + ':snapshot', function(event, status, serverEvent) {
					scope.isQueryValid = status == "error" ? false : true;
					scope.error = status == "error" ? serverEvent.message : null;
					scope.validationQuery = null;
					scope.$apply();
					
					// Stop the listener
					listener();
				})
			};

			// Create a template from the Select and Predicate boxes
			scope.updateTemplate = function() {
				if(scope.query.select == undefined || scope.query.table == undefined)
					return;
				
				var template = "select " + scope.query.select + " from " + scope.query.table.name;
				if(scope.query.predicate != undefined && scope.query.predicate != null && scope.query.predicate.length > 0)
					template = template + " " + scope.query.predicate;
				scope.query.template = template;
			};	

			// Create a parameters object from the query parameters
			scope.getParameters = function() {
				if(scope.parameters == undefined || scope.parameters == null || scope.parameters.length() == 0)
					return undefined;
					
				var parameters = {};
				for(var idx = 0; idx < scope.parameters.length(); idx++) {
					var p = scope.parameters.getByIndex(idx);
					parameters[p.name] = p.value;
				}				
				return parameters;
			};		

			// Switch to display mode
			scope.cancel = function() {
				scope.modal.mode = 'display';
				scope.reset();
			};
		}
	};	
});

app.directive('queryDisplayModal', function($log, $window) {
	return {
		templateUrl: './templates/query-display.html',
		restrict: 'E',
		scope: false,	// This directive is using parent scope
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			// Initialize schema and parameters
			scope.schema 			= undefined;
			scope.parameters 		= new Collection(['name']);

			// Parameter change state
			scope.isParameterChange	= null;

			// On show of the modal, set initial parameters
			$(element).closest('.modal').on('shown.bs.modal', function(event){
				scope.activate();
				scope.$apply();
			});

			// On hide of the modal, reset the scope parameters
			$(element).closest('.modal').on('hidden.bs.modal', function(event){
				scope.reset();
			});	

			// Activate the panel
			scope.activate = function() {
				if(scope.modal == undefined || scope.modal.query == undefined)
					return;
					
				scope.schema = scope.modal.query.schema;
				if(scope.modal.query.parameters != undefined) {
					for(var idx = 0; idx < Object.keys(scope.modal.query.parameters).length; idx++) {
						var key = Object.keys(scope.modal.query.parameters)[idx];
						scope.parameters.add({
							name: key,
							value: scope.modal.query.parameters[key]
						});
					}
				}
			};
			
			// Reset the modal
			scope.reset = function() {
				// Reset scope variables
				scope.schema 		= undefined;
				scope.parameters 	= new Collection(['name']);
				scope.isParameterChange	= null;
			};
			
			// Watch for mode change
			scope.$watch('modal.mode', function() {
				if(scope.modal == undefined || scope.modal.mode == undefined)
					return;
				
				if(scope.modal.mode == 'display')
					scope.activate();
			});
			
			// On change of parameters.length()
			scope.$watch('parameters.length()', function() {
				if(scope.parameters != undefined && scope.parameters.length() > 0)
					$(element).find('.btn-submit').show();
				else
					$(element).find('.btn-submit').hide();
			});
			
			// On change of isParameterChange
			scope.$watch('isParameterChange', function() {
				if(scope.isParameterChange == true)
					$(element).find('.btn-submit').removeAttr('disabled');
				else
					$(element).find('.btn-submit').attr('disabled','disabled');
			});
			
			// Apply any parameter changes
			scope.apply = function() {
				// Create the parameters object
				var parameters = {};
				for(var idx = 0; idx < scope.parameters.length(); idx++) {
					var param = scope.parameters.getByIndex(idx);
					parameters[param.name] = param.value;
				}

				// Add the query to the connection and hide the modal
				scope.connection.applyParameters(parameters, scope.modal.query.id);
				$(element).closest('.modal').modal('hide');			
			};

			// Switch to edit mode
			scope.edit = function() {
				scope.modal.mode = 'edit';
			};
		}
	};	
});

app.directive('querySchema', function($log, $window) {
	return {
		templateUrl: './templates/query-schema.html',
		restrict: 'E',
		scope: false,	// This directive is using parent scope
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			scope.select = function(data) {       
				scope.$broadcast("table:schema-table:onDblClick", data);
			};
		}
	};
});

app.directive('queryParameters', function($log, $window) {
	return {
		templateUrl: './templates/query-parameters.html',
		restrict: 'E',
		scope: false,	// This directive is using parent scope
		controller: function($scope) {
		},
		link: function(scope, element, attrs) {
			// Watch for changes in the template
			scope.$watch('query.template', function() {
				// If there is a query template, parse it into parameters
				if(scope.query != undefined && scope.query.template != undefined) {
					var matches = scope.query.template.match(/\@.\w*/g);
					if(matches != null) {
						for(var idx = 0; idx < matches.length; idx++) {
							if(scope.parameters.get(matches[idx]) == null) {
								scope.parameters.add({
									name: 			matches[idx],
									value:			null
								});
							}
						}
					}
					var keys = [];
					for(var idx = 0; idx < scope.parameters.length(); idx++) {
						var key = scope.parameters.getByIndex(idx).name;
						if(matches == null || matches.indexOf(key) == -1)
							keys.push(key);
					}
					for(var idx = 0; idx < keys.length; idx++)
						scope.parameters.remove(keys[idx]);	 				
				}				
			});
			
			// Watch for changes in the parameters
			scope.setParameter = function(param, val) {
				param.value = val;
				scope.isQueryValid = null;
				scope.isParameterChange = true;
			};
		}
	};	
});

app.directive('ngEnter', function ($log) {
	return function (scope, element, attrs) {
		element.bind("keydown keypress", function (event) {
			if(event.which === 13) {
				scope.$apply(function (){
					scope.$eval(attrs.ngEnter);
				});
				event.preventDefault();
			}
		});
	};
});

app.directive('ngContextMenu', function($log, $parse) {
    return function(scope, element, attrs) {
        var fn = $parse(attrs.ngContextMenu);
        element.bind('contextmenu', function(event) {
            scope.$apply(function() {
                event.preventDefault();
                fn(scope, {$event:event});
            });
        });
    };
});

app.directive('livemap', function($log, Overlay, EventListener) {
	return {
		templateUrl: './templates/live-map.html',
		restrict: 'E',
		transclude: true,
		controller: function($scope) {
			this.addLayer = function(layer) {
				$scope.layers.push(layer);
				$scope.map.node.addLayer(layer.node);
			};
			this.getLayers = function() {
				return $scope.layers;
			};
			this.getId = function() {
				return $scope.elementId;
			};
			this.getMap = function() {
				return $scope.map.node;
			};
		},
		link: function(scope, element, attrs, ctrl, transclude) {		
			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});
			
			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;

			// Read the options from attributes
			scope.options = {
				"attribution":		attrs.attribution == undefined || attrs.attribution.length == 0 ?
										attrs.tileLayer == undefined || attrs.tileLayer.length == 0 ? 
											'<a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>' :
											undefined :
										attrs.attribution,
									// Optional. Attribution for tile layer.
									// Maporama: <a href="http://www.tibco.com/">Tibco GeoAnalytics</a>
									// OSM: &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>
									// MapQuest Open: Tiles Courtesy of <a href="http://www.mapquest.com/">MapQuest</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>
									// ArcGIS Aerial: Tiles &copy; Esri &mdash; Source: Esri, i-cubed, USDA, USGS, AEX, GeoEye, Getmapping, Aerogrid, IGN, IGP, UPR-EGP, and the GIS User Community
									
				"center":			attrs.center == undefined || attrs.center.length == 0 ?
										undefined : 
										eval(attrs.center),
									// Optional. Initial center for the map. Must be array object [lat,lon]
									//   For example [52.000000,5.300000]

				"filterPanel":		attrs.filterPanel == undefined || attrs.filterPanel.length == 0 ?
										false :
										eval(attrs.filterPanel),
									// Optional. Display the filter panel.
									
				"minZoom":			attrs.minZoom == undefined || attrs.minZoom.length == 0 ?
										undefined : 
										eval(attrs.minZoom),
									// Optional. Minimum zoom level. Must be an integer.

				"maxZoom":			attrs.maxZoom == undefined || attrs.maxZoom.length == 0 ?
										18 : 
										eval(attrs.maxZoom),
									// Optional. Maximum zoom level. Must be an integer.

				"opacity":			attrs.opacity == undefined || attrs.opacity.length == 0 ?
										1.0 : 
										eval(attrs.opacity),
									// Optional. Opacity of tile layer. Must be an integer.
				
				"tileLayer":		attrs.tileLayer == undefined || attrs.tileLayer == 0 ?
										'http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png' :
										attrs.tileLayer,
									// Optional. Tile layer to display.
									/* 
										Maporama: http://tilescf.maporama.com/georasterservices/v101/standard_map/{z}/{x}/{y}.png
										OSM: http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png
										MapQuest Open: http://otile1.mqcdn.com/tiles/1.0.0/map/{z}/{x}/{y}.png
										ArcGIS Aerial: http://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}
										Other providers: http://leaflet-extras.github.io/leaflet-providers/preview/
									*/
									
									
				"zoom":				attrs.zoom == undefined || attrs.zoom.length == 0 ?
										null : 
										eval(attrs.zoom),
									// Optional. Initial zoom level. Must be an integer.
			};

			// Define the center
			var center = scope.options.center == undefined ? null : L.latLng(scope.options.center[0],scope.options.center[1]);

			// Create the map object
			scope.map = {};
			
			// Create the leaflet map
			scope.map.node = L.map($(element).find('div')[0], {
				center: 	center,
				zoom: 		scope.options.zoom,
				minZoom: 	scope.options.minZoom,
				maxZoom: 	scope.options.maxZoom
			});		

			// Add the tile layer
			L.tileLayer( scope.options.tileLayer, {
				maxZoom : scope.options.maxZoom,
				attribution: scope.options.attribution,
				opacity: scope.options.opacity
			}).addTo(scope.map.node);

			// Create an array of layers
			scope.layers = [];
			
			// Send event that map is ready
			scope.$broadcast('map:ready');
			
			// Bind listener on window resize to adjust map
			EventListener.windowResize(function() {
				if(scope.map.node == null) return;
				//table.draw();
			});
			
			// Bind listener on component display for tabs, applications, and visibility-toggle divs
			EventListener.componentDisplay(element, function() {
				if(scope.map.node == null) return;
				setTimeout(function() {
					scope.map.node.invalidateSize();
				});
			});			
		}
	};	
});

app.directive('livemapFilterPanel', function($log, EventListener) {
	return {
		templateUrl: './templates/live-map-filter-panel.html',
		restrict: 'E',
		scope: {},
		transclude: true,
		require: '^livemap',
		link: function(scope, element, attrs, ctrl, transclude) {		

			// Show the map menu overlay when clicking the overlay button
			$(element).find('.livemap-overlay-button').on('click', function(e) {
				$(element).find('.livemap-overlay').show();
				return false;
			});
			
			// Hide the map menu overlay when clicking the overlay close button
			$(element).find('.livemap-overlay-close-icon').on('click', function(e) {
				$(element).find('.livemap-overlay').hide();
				return false;
			});
		
			// Capture double clicks on this panel to avoid zooming the map below
			$(element).on('dblclick', function(e) {
				return false;
			});
			
			// Get the layers on the map
			scope.getLayers = function() {
				return ctrl.getLayers();
			};
			
			// Bind listener on window resize to adjust component sizes
			EventListener.windowResize(function() {
				resizePanels();
			});
			
			// Bind listener on component display for tabs, applications, and visibility-toggle divs
			EventListener.componentDisplay(element, function() {
				resizePanels();
			});

			// Resize the filter panels based on container height
			var resizePanels = function() {
				setTimeout(function() {
					var count = $(element).find('.livemap-overlay-layer').length;
					var heightPerLayer = Math.floor($(element).find('.livemap-overlay-content').height() / count) - 4;
					$(element).find('.livemap-overlay-layer').each(function(index, elem) {
						$(elem).css('height', heightPerLayer + 'px');
					});
				
					$(element).find('.livemap-layer-header').each(function(index, elem) {
						var thisHeight = $(elem).parent().parent().height() - $(elem).height();
						$(elem).parent().find('.livemap-layer-control-scroll-container').css('height', thisHeight + 'px');
					});					
				},10);
			};
		}
	};	
});

app.directive('livemapLayer', function($log, Overlay) {
	return {
		restrict: 'E',
		transclude:	true,
		scope: {
			'query':	'=',
			'selected':	'=',
		},
		require: '^livemap',
		controller: function($scope) {
			this.addComponent = function(component) {
				var componentGroup = $scope.dataset.get(component.tid);
				if(componentGroup == undefined) {
					componentGroup = new Collection(['tidx']);
					componentGroup.tid = component.tid;
					$scope.dataset.add(componentGroup);
				}
				component.data.tidx = componentGroup.length();
				componentGroup.add(component);
				$scope.layer.node.addLayer(component.node);
				$scope.$apply();
			};
			this.getComponent = function(tid) {
				return $scope.dataset.get(tid);
			};
			this.removeComponent = function(component) {
				$scope.layer.node.removeLayer(component.node);
				var componentGroup = $scope.dataset.get(component.tid);
				if(componentGroup != undefined) {
					var component = componentGroup.get(component.tidx);
					if(component != undefined) {
						componentGroup.remove(component.tidx);
					}
				}
				$scope.$apply();
			};
			this.verifyVisibility = function() {
				$scope.layer.verifyVisibility();
			};
			this.select = function(component) {
				component.selectedType = undefined;
				$scope.select(component);
				$scope.$apply();
			};
			this.getColor = function(key) {
				if($scope.options.colorScale == undefined) 
					return undefined;
				return $scope.colorScale.color(key);
			};
			this.getObjectTitleExpression = function() {
				return $scope.options.objectTitle;
			};
			this.getEntity = function() {
				return $scope.options.entity;
			};
			this.getMap = function() {
				return $scope.getMap();
			};
		},
		link: function(scope, element, attrs, ctrl, transclude) {		
			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});

			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;
			
			// Set the query id
			scope.queryId = attrs.queryId;
			
			// Define color categories
			scope.color_categories = {			
				color_category10:	[
									  '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
									  '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'
									],
									
				color_category20a:	[
									  '#1f77b4', '#aec7e8', '#ff7f0e', '#ffbb78',
									  '#2ca02c', '#98df8a', '#d62728', '#ff9896',
									  '#9467bd', '#c5b0d5', '#8c564b', '#c49c94',
									  '#e377c2', '#f7b6d2', '#7f7f7f', '#c7c7c7',
									  '#bcbd22', '#dbdb8d', '#17becf', '#9edae5'
									],

				color_category20b:	[
									  '#393b79', '#5254a3', '#6b6ecf', '#9c9ede',
									  '#637939', '#8ca252', '#b5cf6b', '#cedb9c',
									  '#8c6d31', '#bd9e39', '#e7ba52', '#e7cb94',
									  '#843c39', '#ad494a', '#d6616b', '#e7969c',
									  '#7b4173', '#a55194', '#ce6dbd', '#de9ed6'
									],

				color_category20c: 	[
									  '#3182bd', '#6baed6', '#9ecae1', '#c6dbef',
									  '#e6550d', '#fd8d3c', '#fdae6b', '#fdd0a2',
									  '#31a354', '#74c476', '#a1d99b', '#c7e9c0',
									  '#756bb1', '#9e9ac8', '#bcbddc', '#dadaeb',
									  '#636363', '#969696', '#bdbdbd', '#d9d9d9'
									],
			};

			// Read the options from attributes
			scope.options = {
				"colorScale":		attrs.colorScale == undefined || attrs.colorScale.length == 0 ?
										undefined :
										scope.color_categories[attrs.colorScale],
									// Optional. Use a color scale to define fill and stroke color of path elements.
									//   Use one of the configured color scales: color_category10, color_category20a, 
									//   color_category20b, color_category20c
									//   This color value will override any configured color for stroke or fill on path
									//     elements within the layer. It will not set any color on icon markers.
									
				"entity":			attrs.entity == undefined || attrs.entity.length == 0 ?
										undefined :
										attrs.entity,
									// Optional. Entity on the object to render on the map. May be an array.
									//   This is for special case processing where the object being passed in to 
									//   the layer has a child object on it that is desired to be rendered. This 
									//   will rarely occur in LDM datasets but may in HTTP datasets.

				"hideable":			attrs.hideable == undefined || attrs.hideable.length == 0 ?
										true :
										eval(attrs.hideable),
									// Optional. Allow hiding components on the layer.

				"objectTitle":		attrs.objectTitle == undefined || attrs.objectTitle.length == 0 ?
										undefined :
										attrs.objectTitle,
									// Optional. Object title to display in filter box. This is a javascript expression that
									//   can use the object data.
									//   Example: 'Flight num:' + object.flightNum
									
				"queryDetailIcon":	attrs.queryDetailIcon == undefined ? false : eval(attrs.queryDetailIcon),
									// Optional. Display the query type icon. May be true for false.
									
				"queryViewer":		attrs.queryViewer == undefined ? false : eval(attrs.queryViewer),
									// Optional. Allow display of query view modal. May be true for false. If true,
									//   must also set queryDetailIcon to true also for display.
				
				"queryEditor":		attrs.queryEditor == undefined ? false : eval(attrs.queryEditor),
									// Optional. Allow display of query view and edit modal. May be true for false. If true,
									//   must also set queryDetailIcon to true also for display, but not necessary to set
									//   queryViewer.
				
				"selected":			attrs.selected == undefined || attrs.selected.length == 0 ?
										undefined : 
										attrs.selected,
									// Optional. Selected attribute on the controller

				"title":			attrs.title == undefined || attrs.title.length == 0 ?
										undefined : 
										attrs.title,
									// Optional. Layer title
			};

			// Wait for map ready event
			scope.$on('map:ready', function() {
				scope.layer = createLayer();
				scope.layer.node = new L.LayerGroup();
				ctrl.addLayer(scope.layer);
				scope.$broadcast('map:layer:ready');
			});
				
			// Declare a dataset collection
			scope.dataset = new Collection(['tid']);
			
			// Define a function to select component
			//   Data Internal > External
			scope.select = function(component) {
				if(component == undefined) return;
				
				// If selection is enabled
				if(scope.options.selected != undefined && scope.options.selected.length > 0) {
					if(scope.selectedComponent == undefined) {
						scope.selectedComponent = component;
						component.bringToFront();
						component.selected = true;
						component.selectedType = undefined;
						component.update();
					}
					else {
						if(scope.selectedComponent == component) {
							scope.selectedComponent = undefined;
							component.bringToBack();
							component.selected = false;
							component.selectedType = undefined;
							component.update();
						}
						else {
							var previous = scope.selectedComponent;
							previous.bringToBack();
							previous.selected = false;
							previous.selectedType = undefined;
							previous.update();
							
							scope.selectedComponent = component;
							component.bringToFront();
							component.selected = true;
							component.selectedType = undefined;
							component.update();
						}
					}
					scope.$emit("map:layer:onSelect", scope.selectedComponent);				
					scope.$emit("map:" + ctrl.getId() + ":layer:" + scope.elementId + ":onSelect", scope.selectedComponent);
				}
			};
			
			// Setup an overlay watching the query status
			Overlay.create(scope, scope.queryId, element);

			// Watch the selected object data
			//   Data External > Internal
			scope.$watch('selected.data', function(newVal, oldVal) {
				// Both are undefined, no change
				if(newVal == undefined && scope.selectedComponent == undefined) return;
					
				// New val is undefined, update the previous
				if(newVal == undefined && scope.selectedComponent != undefined) {
					var previous = scope.selectedComponent;
					scope.selectedComponent = undefined;
					if(previous != undefined) {
						previous.bringToBack();
						previous.selected = false;
						previous.update();
					}
					return;
				}
				
				// New val exists, lookup the componentGroup and component
				var componentGroup = scope.dataset.get(newVal.tid);
				var component = undefined;

				if(componentGroup != undefined) {
					var tidx = newVal.tidx;
					if(tidx == undefined) tidx = 0;
					component = componentGroup.get(tidx);
					if(component != undefined) 
						component.selectedType = scope.selected.selectedType;
				}
				
				// If component is not found, unselect the previous if present and return
				var previous = scope.selectedComponent;
				if(component == undefined) {
					if(previous != undefined) {
						previous.bringToBack();
						previous.selected = false;
						previous.update();
					}
					return;
				};						
				
				// If new val exists, but previous doesn't, update the new val
				if(newVal != undefined && previous == undefined) {
					scope.selectedComponent = component;
					component.bringToFront();
					component.selected = true;
					component.update();
					return;
				}
				
				// If both new val and previous exist, and data is different update both
				if(newVal != undefined && previous != undefined && newVal != previous.data) {
					previous.bringToBack();
					previous.selected = false;
					previous.update();

					scope.selectedComponent = component;
					component.bringToFront();
					component.selected = true;
					component.update();
					return;
				}
				
				// If data is the same, take no action
			});

			// Watch the query
			scope.$watch('query', function(newVal, oldVal) {
				scope.layer.query = scope.query;
			});
			
			// Listen to snapshot events
			scope.$on('query:' + scope.queryId + ':snapshot', function(event, action, serverEvent, object) {
				if(action == 'change') {
					for(var idx = 0; idx < scope.dataset.length(); idx++) {
						var component = scope.dataset.getByIndex(idx);
						component.remove();
					}
					scope.dataset.removeAll();
				}
				else if(action == 'onSuccess:object') {
					// Send an event to the children to create a new object
					scope.$broadcast('map:layer:insertObject', object);
				}
			});
				
			// Listen to subscribe events
			scope.$on('query:' + scope.queryId + ':subscribe', function(event, action, serverEvent, object) {
				if(action == 'onSnapshotStart')
					scope.snapshot = true;
				else if(action == 'onInsert') {									
					// Send an event to the children to create a new object
					scope.$broadcast('map:layer:insertObject', object);
				}
				else if(action == 'onUpdate') {
					// Send an event to the children to update an object
					scope.$broadcast('map:layer:updateObject', object);

					// Get the component from the dataset and update the data
					var componentGroup = scope.dataset.get(object.tid);
					if(componentGroup == undefined) return;
					for(var idx = 0; idx < componentGroup.length(); idx++) {
						var component = componentGroup.getByIndex(idx);
						
						// Normal processing
						if(scope.options.entity == undefined)
							component.data = object;
						// Special case for child entity
						else {
							var parent = $.extend(true, {}, object);
							component.data.parent = parent;
						}
						component.update();
					}
				}
				else if(action == 'onDelete') {
					// Send an event to the children to delete an object
					scope.$broadcast('map:layer:deleteObject', object);

					// Remove the layer object from the dataset
					var componentGroup = scope.dataset.get(object.tid);
					if(componentGroup == undefined) return;
					while(componentGroup.length() > 0) {
						var component = componentGroup.getByIndex(0);
						component.remove();
					}
					componentGroup.removeAll();
					scope.dataset.remove(object.tid);
				}
				if(action == 'onSnapshotEnd')
					scope.snapshot = undefined;
			});
				
			// Listen to unsubscribe events
			scope.$on('query:' + scope.queryId + ':unsubscribe', function(event, message, serverEvent) {
				// Send an event to the children to purge objects
				scope.$broadcast('map:layer:purge');
				
				for(var idx = 0; idx < scope.dataset.length(); idx++) {
					var componentGroup = scope.dataset.getByIndex(idx);
					if(componentGroup == undefined) return;
					while(componentGroup.length() > 0) {
						var component = componentGroup.getByIndex(0);
						component.remove();
					}
					componentGroup.removeAll();
				}
				scope.dataset.removeAll();
			});
					
			// Create a new layer instance
			var createLayer = function(object) {
				return new Layer();
			};	

			// Define color scale object
			scope.colorScale = new ColorScale(scope.options.colorScale);
		
			// Return the map object from the parent map directive
			scope.getMap = function() {
				return ctrl.getMap();
			}
			
			// DEFINE OBJECTS
			// Layer object
			{
				function Layer() {
					this.node 				= undefined;
					this.title 				= scope.options.title;
					this.queryId 			= scope.queryId;
					this.query				= scope.query;
					this.options			= scope.options;
					this.visible 			= true;				
				};
				
				Layer.prototype.verifyVisibility = function() {
					for(var idx = 0; idx < scope.dataset.length(); idx++) {
						var component = scope.dataset.getByIndex(idx);
						if(component.visible == false) {
							this.visible = false;
							return;
						}
					}
					this.visible = true;
				};
				
				Layer.prototype.updateVisibility = function() {
					for(var idx = 0; idx < scope.dataset.length(); idx++) {
						var componentGroup = scope.dataset.getByIndex(idx);
						for(var cidx = 0; cidx < componentGroup.length(); cidx++) {
							var component = componentGroup.getByIndex(cidx);
							component.visible = this.visible;
							component.updateVisibility(false);
						}
					}
				};
				
				Layer.prototype.getObjects = function() {
					var objects = [];
					for(var idx = 0; idx < scope.dataset.length(); idx++) {
						objects.push.apply(objects, scope.dataset.getByIndex(idx).toArray());
					}
					return objects;
				};
			}
			
			
			// Color Scale object
			{
				function ColorScale(scale) {
					this.scale = scale;
					this.lastUsed = -1;
					this.colorMap = {};
				};
				
				ColorScale.prototype.color = function(key) {
					if(this.colorMap[key] != undefined) 
						return this.colorMap[key];
					this.lastUsed = this.lastUsed + 1;
					if(this.lastUsed == this.scale.length) 
						this.lastUsed = 0;
					var color = this.scale[this.lastUsed];
					this.colorMap[key] = color;
					return color;
				};
			}
					
		}
	};	
});

app.directive('livemapIconMarker', function($log, $compile, SimpleDateFormat, NumberFormat, Overlay, Map) {
	return {
		restrict: 'E',
		scope: {},
		transclude: true,
		require: '^livemapLayer',
		controller: function($scope) {
		},
		link: function(scope, element, attrs, ctrl, transclude) {
			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});

			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;
			
			// Read the options from attributes
			scope.options = {
				"disabled":			attrs.disabled == undefined || attrs.disabled.length == 0 ?
										false :
										eval(attrs.disabled), 
									// Optional. Disable the layer.
									
				"position":			attrs.position, 
									// Mandatory. Position for the point. Define as an array with object properties for lat and lon.
									//   Example 1: [object.lat, object.lon]
									//     This will plot the object at the position object.lat and object.lon where object is the
									//     data row, and lat and lon are the properties of that object that point to latitude and 
									//     longitude respectively.

				"marker":			{},
									// Mandatory. Automatically generated for this directive type.

				"icon":				attrs.icon == undefined || attrs.icon.length == 0 ?
										undefined :
										{url:	attrs.icon},
									// Optional but you won't see anything if not specified. Icon URL for image to display as standard marker. 
									// May be a literal value:
									//	 Example 1: img/lights/green-light.png
									//     This will always render green-light.png
									// Or a conditional object. Keys in the object define the image URLs. Values define boolean
									//   expressions. The first expression that evaluates to true will get assigned.
									//   Example 2: {'img/lights/green-light.png': object.status == 'Approach', 'img/lights/blue-light.png': object.status == 'Circling', 'img/lights/red-light.png': true}
									//     This will give a green-light.png if object has status Approach, blue-light.png if 
									//     object has status Circling, or red-light.png otherwise.
									
				"iconLabel":		attrs.iconLabel == undefined || attrs.iconLabel.length == 0 ?
										undefined :
										eval(attrs.iconLabel) == true ? {} : undefined,
									// Optional. Apply an icon label. Must be true or false.
									
				"iconLabelContent":	$(element).find('livemap-icon-label').length != 1 ?
										undefined :
										$(element).find('livemap-icon-label').html(),
									// Optional. Icon label HTML content. Defined in a tag structure underneath this element.
									//   Example:
									//     <livemap-icon-label ng-non-bindable>
									//       <span>
									//         <div style="background-color:white;font-size:20px;">
									//			  <img src="img/other/status-ok.png" height="20px" ng-if="object.status == 'Approach'"></img>
									//			  <img src="img/other/status-unknown.png" height="20px" ng-if="object.status == 'Circling'"></img>
									//			  {{::object.flightNumber}}
									//		   </div>
									//	     </span>
									//     </livemap-icon-label>
									//
									//   Notes:
									//     1. <livemap-icon-label> MUST have ng-non-bindable
									//     2. First element within <livemap-icon-label> MUST be a single root element, recommend <span> or <div>
									//     3. Any content you wish between the root element tags
									//     4. Can use any Angular binding on the object, which represents the row
									//     5. Recommend to use bind-once notation (::) {{::object.flightNumber}} to improve performance
									//     6. Can use ng-if, ng-switch, and other structures if desired
									
				"popup":			$(element).find('livemap-popup').length != 1 ?
										undefined :
										$(element).find('livemap-popup').html(),
									// Optional. Popup HTML content. Defined in a tag structure underneath this element.
									//   Example:
									//     <livemap-popup ng-non-bindable>
									//        <span>
									//		    <h3>Flight: {{::object.flightNumber}}</h3>
									//	     </span>
									//     </livemap-icon-label>
									// 
									//   See all notes above as for iconLabelContent. Same rules apply here.
			};

			// Get marker options
			if(scope.options.marker != undefined) {
				scope.options.marker.options = {
					"clickable":		attrs.clickable == undefined || attrs.clickable.length == 0 ?
											true :
											eval(attrs.clickable),
										// Optional. Make the marker clickable. Doesn't seem to work with Leaflet. 
									
					"opacity":			attrs.opacity == undefined || attrs.opacity.length == 0 ?
											1.0 :
											eval(attrs.opacity),
										// Optional. Opacity for the marker.

					"riseOnHover":		attrs.riseOnHover == undefined || attrs.riseOnHover.length == 0 ?
											false :
											eval(attrs.riseOnHover),
										// Optional. Rise over other elements on mouseover. Must be true or false.

					"riseOffset":		attrs.riseOffset == undefined || attrs.riseOffset.length == 0 ?
											250 :
											eval(attrs.riseOffset),
										// Optional. Z-index offset on hover. Ignored if riseOnHover is false.

					"zIndexOffset":		attrs.zIndexOffset == undefined || attrs.zIndexOffset.length == 0 ?
											0 :
											eval(attrs.zIndexOffset),
										// Optional. Z-index offset for normal condition.
				};
			}
			
			// Get icon options
			if(scope.options.icon != undefined) {
				scope.options.icon.options = {
					"iconSize":		attrs.iconSize == undefined || attrs.iconSize.length == 0 ?
										[0, 0] :
										eval(attrs.iconSize),
									// Optional. Size of the icon image in pixels. Object array [x, y] where x and y are numbers.
									
					"iconAnchor":	attrs.iconAnchor == undefined || attrs.iconAnchor.length == 0 ?
										[0, 0] :
										eval(attrs.iconAnchor),
									// Optional. Coordinates of the icon relative to its top left corner. Object array [x, y] where x and y are numbers.

					"popupAnchor":	attrs.popupAnchor == undefined || attrs.popupAnchor.length == 0 ?
										[0, 0] :
										eval(attrs.popupAnchor),
									// Optional. Coordinates of the point where popups open, relative to the icon anchor. Object array [x, y] where x and y are numbers.

					"shadowUrl":	attrs.shadowUrl == undefined || attrs.shadowUrl.length == 0 ?
										undefined :
										attrs.shadowUrl,
									// Optional. Url for shadow image.

					"shadowSize":	attrs.shadowSize == undefined || attrs.shadowSize.length == 0 ?
										undefined :
										eval(attrs.shadowSize),
									// Optional. Shadow size. Object array [x, y] where x and y are numbers.

					"shadowAnchor":	attrs.shadowAnchor == undefined || attrs.shadowAnchor.length == 0 ?
										undefined :
										eval(attrs.shadowAnchor),
									// Optional. Shadow anchor. Object array [x, y] where x and y are numbers.									
				};
				
				// Get icon label options
				if(scope.options.iconLabel != undefined) {
					scope.options.iconLabel.options = {
						"labelAnchor":		attrs.labelAnchor == undefined || attrs.labelAnchor.length == 0 ?
												new L.Point(0, 0) :
												new L.Point(eval(attrs.labelAnchor)[0], eval(attrs.labelAnchor)[1]),
											// Optional. Label anchor position. Object array [x, y] where x and y are numbers.

						"labelClassName":	attrs.labelClasses == undefined || attrs.labelClasses.length == 0 ?
												undefined :
												attrs.labelClasses,
											// Optional. Label class names. These will be applied to the wrapper div in the DOM.

						"wrapperAnchor":	attrs.wrapperAnchor == undefined || attrs.wrapperAnchor.length == 0 ?
												new L.Point(0, 0) :
												new L.Point(eval(attrs.wrapperAnchor)[0], eval(attrs.wrapperAnchor)[1]),
											// Optional. Wrapper anchor. Object array [x, y] where x and y are numbers.									
					};
					
					// Merge attributes
					for(var attrname in scope.options.iconLabel.options)
						scope.options.icon.options[attrname] = scope.options.iconLabel.options[attrname];
				}
			}
			
			// Wait for layer insert object event
			scope.$on('map:layer:insertObject', function(event, object) {
				// Normal processing
				if(ctrl.getEntity() == undefined) {
					var component = createComponentObject(object, 0);
					component.draw();
				}
				// Special case for child entity
				else {
					var parent = $.extend(true, {}, object);
					parent[ctrl.getEntity()] = undefined;
					
					// Timeout gives the service call time to load if its loaded from a different source
					if(object[ctrl.getEntity()] != undefined) {
						for(var idx = 0; idx < object[ctrl.getEntity()].length; idx++) {
							var entity = object[ctrl.getEntity()][idx];	
							entity.tid = object.tid;
							entity.parent = parent;
							var component = createComponentObject(entity, idx);
							component.draw();
						}
					}
				}
			});

			// Wait for layer update object event
			scope.$on('map:layer:updateObject', function(event, object) {
				var position = eval(scope.options.position);
				if(position[0] == null || position[1] == null) {
					return;
				}
					
				// Normal processing
				if(ctrl.getEntity() == undefined) {
					var component = ctrl.getComponent(object.tid);
					if(component == null)  {
						component = createComponentObject(object, 0);
						component.draw();
					}
				}
				// Special case for child entity
				else {
					var component = ctrl.getComponent(object.tid);
					if(component == null)  {
						var parent = $.extend(true, {}, object);
						parent[ctrl.getEntity()] = undefined;

						// Timeout gives the service call time to load if its loaded from a different source
						if(object[ctrl.getEntity()] != undefined) {
							for(var idx = 0; idx < object[ctrl.getEntity()].length; idx++) {
								var entity = object[ctrl.getEntity()][idx];	
								entity.tid = object.tid;
								entity.parent = parent;
								var component = createComponentObject(entity, idx);
								component.draw();
							}
						}
					}
				}
			});

			// Create a new map component for the given object
			var createComponentObject = function(object, idx) {
				var model = new IconMarkerModel(scope.options);
				var component = new IconMarkerComponent(model, object, idx);
				return component;
			};
			
			// DEFINE OBJECTS
			// IconMarkerModel object
			{
				function IconMarkerModel(options) {
					this.options = options;
					this.$compile = undefined;
					
					if(this.options.iconLabelContent != undefined) {
						this.$compile = $compile(this.options.iconLabelContent);
					}				
				};
			}
			
			// IconMarkerComponent object
			{
				function IconMarkerComponent(model, object, idx) {
					this.model		= model;
					this.tid 		= object.tid;
					this.tidx 		= idx;
					this.data 		= object;
					this.node 		= undefined;
					this.selected 	= false;
					this.title 		= undefined;
					this.visible 	= true;
					this.iconOpts	= undefined,
					this.context 	= {};
					this.popup		= undefined;
					
					this.hasComplexContent 	= this.model.options.iconLabelContent != undefined;
					this.hasPopup			= scope.options.popup != undefined;
					
					// Create the icon options object if specified
					if(scope.options.icon != undefined) {
						this.iconOpts = $.extend({}, scope.options.icon.options);
						if(scope.options.icon.url.indexOf('{') == -1)
							this.iconOpts.iconUrl = scope.options.icon.url;
						else {
							var i = eval('(' + scope.options.icon.url + ')');											
							for(var idx = 0; idx < Object.keys(i).length; idx++) {
								var url = Object.keys(i)[idx];
								if(i[url] == true) {
									this.iconOpts.iconUrl = url;
									break;
								}
							}
						}
						
						// Create the icon
						var Icon = L.Icon.Label.extend({
							options: this.iconOpts
						});				
						this.icon = new Icon({});						

						// If the icon has complex content, setup the compile function and scope
						if(this.hasComplexContent) {
							this.context.contentScope = scope.$new(true);
							this.context.contentScope.Math = window.Math;
							this.context.contentScope.formatDateTime = function(dateStr, format) {
								return SimpleDateFormat.format(dateStr, format);
							};
							this.context.contentScope.formatNumber = function(numStr, format) {
								return NumberFormat.format(numStr, format);
							};
							this.context.$compile = model.$compile;
						}
					}
					
					// Create a popup if specified
					if(this.hasPopup) {
						this.popup = Map.createPopup(scope, scope.options.popup);
					}
					
					// These properties are used for microbatching the updates
					this.microbatch = {
						rtime:		new Date(1, 1, 2000, 12,00,00),
						timeout:	false,
						delta:		0
					}					
				};

				IconMarkerComponent.prototype.bringToBack = function() { };
				IconMarkerComponent.prototype.bringToFront = function() { };
				
				IconMarkerComponent.prototype.draw = function() {
					var self = this;
					
					// Create the node
					var opts = $.extend({}, scope.options.marker.options);
					this.node = L.marker([0,0], opts); // Position will be updated later
					
					// Add event handler
					this.node.on('click', function(event) {
						ctrl.select(self);
					});

					// Set the icon to the node if specified
					if(this.icon != undefined)
						this.node.setIcon(this.icon);
					
					// Update the node and add to the layer
					this.update();											
					ctrl.addComponent(this);
				};
				
				IconMarkerComponent.prototype.update = function() {
					var self = this;

					// Microbatch handler
					self.microbatch.rtime = new Date();
					if(self.microbatch.timeout === false) {
						self.microbatch.timeout = true;
						setTimeout(eventEnd, self.microbatch.delta);
					}					
					function eventEnd() {
						if (new Date() - self.microbatch.rtime < self.microbatch.delta)
							setTimeout(eventEnd, self.microbatch.delta);
						else {
							self.microbatch.timeout = false;
							run();
						}               
					};
					
					function run() {
						// Create the icon if specified
						if(scope.options.icon != undefined) {
						
							// Update the contentScope if it has complex content
							if(self.hasComplexContent) {
								self.context.contentScope.object = self.data;
								self.context.contentScope.component = self;
							}

							if(self.hasComplexContent == false) {
								self.node.setIcon(L.icon(self.iconOpts));																
							}
							else {
								var content = self.context.$compile(self.context.contentScope, function(clonedElement, scope) {
									// This allows a delay to permit the template to render without flickering in the UI
									scope.$evalAsync(function() {
										// Optimization will only update the label if it has changed
										if(self.icon.options.labelText != clonedElement[0].innerHTML) {
											//self.icon.options.labelText = clonedElement.html();
											self.icon.options.labelText = '';
											self.node.setIcon(self.icon);

											var n = $(self.node._icon).find('.leaflet-marker-iconlabel');
											n.empty().html(clonedElement);
										}
									});
								});
							}							
						}
						
						// Update the popup
						Map.updatePopup(scope, self);
		
						// Update the marker and popup position
						Map.updatePosition(scope, self);
						
						// Update the title
						Map.updateTitle(scope, ctrl, self);
						
						// Update z-index based on whether the marker is selected
						if(self.selected == true)
							self.node.setZIndexOffset(10000);
						else
							self.node.setZIndexOffset(scope.options.marker.options.zIndexOffset);
						
						// Apply scope
						scope.$apply();
					}
				};
				
				IconMarkerComponent.prototype.remove = function() { ctrl.removeComponent(this); };
				
				IconMarkerComponent.prototype.updateVisibility = function(verify) {
					if(this.visible == true)
						this.node._icon.style.display = '';
					else if(this.visible == false)
						this.node._icon.style.display = 'none';
					if(verify != false)
						ctrl.verifyVisibility();
				};
			}
			
		}
	};	
});

app.directive('livemapCircleMarker', function($log, SimpleDateFormat, NumberFormat, Overlay, Map) {
	return {
		restrict: 'E',
		scope: {},
		transclude: true,
		require: '^livemapLayer',
		controller: function($scope) {
		},
		link: function(scope, element, attrs, ctrl, transclude) {
			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});

			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;
			
			// Read the options from attributes
			scope.options = {
				"disabled":			attrs.disabled == undefined || attrs.disabled.length == 0 ?
										false :
										eval(attrs.disabled), 
									// Optional. Disable the layer.
									
				"position":			attrs.position, 
									// Mandatory. Position for the point. Define as an array with object properties for lat and lon.
									//   Example 1: [object.lat, object.lon]
									//     This will plot the object at the position object.lat and object.lon where object is the
									//     data row, and lat and lon are the properties of that object that point to latitude and 
									//     longitude respectively.

				"circleMarker":		{},
									// Mandatory. Automatically generated for this directive type.
									
				"popup":			$(element).find('livemap-popup').length != 1 ?
										undefined :
										$(element).find('livemap-popup').html(),
									// Optional. Popup HTML content. Defined in a tag structure underneath this element.
									//   Example:
									//     <livemap-popup ng-non-bindable>
									//        <span>
									//		    <h3>Flight: {{::object.flightNumber}}</h3>
									//	     </span>
									//     </livemap-icon-label>
									// 
									//   See all notes above as for iconLabelContent. Same rules apply here.
									
				"radius":			attrs.radius,
									// Mandatory. Radius of the circle marker in pixels.
									// May be a literal value:
									//	 Example 1: 10
									//     Radius will always be 10px
									// Or a conditional object. Keys in the object define the radius. Values define boolean
									//   expressions. The first expression that evaluates to true will be assigned as radius.
									//   Example 2: {'10': object.status == 'Approach', '20': object.status == 'Circling', '30': true}
									//     This will give 10px if object has status Approach, 20px if object has status Circling, 
									//     30px otherwise.									
			};

			// Get path options
			scope.options.path = {
				options:	Map.getPathOptions(attrs)
			}
						
			// Wait for layer create new object event
			scope.$on('map:layer:insertObject', function(event, object) {
				// Normal processing
				if(ctrl.getEntity() == undefined) {
					var component = createComponentObject(object, 0);
					component.draw();
				}
				// Special case for child entity
				else {
					if(object[ctrl.getEntity()] != undefined) {
						var parent = $.extend(true, {}, object);
						parent[ctrl.getEntity()] = undefined;
						
						// Timeout gives the service call time to load if its loaded from a different source
						for(var idx = 0; idx < object[ctrl.getEntity()].length; idx++) {
							var entity = object[ctrl.getEntity()][idx];	
							entity.tid = object.tid;
							entity.parent = parent;
							var component = createComponentObject(entity, idx);
							component.draw();
						}
					}
				}
			});
			
			// Create a new map component for the given object
			var createComponentObject = function(object, idx) {
				var component = new CircleMarkerComponent(object, idx);
				return component;
			};
			
			// DEFINE OBJECTS
			// CircleMarkerComponent object
			{
				function CircleMarkerComponent(object, idx) {
					this.tid 		= object.tid;
					this.tidx 		= idx;
					this.data 		= object;
					this.node 		= undefined;
					this.selected 	= false;
					this.title 		= undefined;
					this.visible 	= true;
					this.popup		= undefined;
					
					this.hasPopup			= scope.options.popup != undefined;
					
					// Create a popup if specified
					if(this.hasPopup) {
						this.popup = Map.createPopup(scope, scope.options.popup);
					}
										
					// These properties are used for microbatching the updates
					this.microbatch = {
						rtime:		new Date(1, 1, 2000, 12,00,00),
						timeout:	false,
						delta:		0
					}					
				};

				CircleMarkerComponent.prototype.bringToBack = function() { this.node.bringToBack();	};
				CircleMarkerComponent.prototype.bringToFront = function() { this.node.bringToFront(); };
				
				CircleMarkerComponent.prototype.draw = function() {
					var self = this;
					
					// Create the node
					var opts = $.extend({}, scope.options.path.options);
					this.node = L.circleMarker(null, opts);			
					
					// Add event handler
					this.node.on('click', function(event) {
						ctrl.select(self);
					});

					// Update position
					Map.updatePosition(scope, self);
					
					// Update and add component to the layer
					this.update();
					ctrl.addComponent(this);
				};
				
				CircleMarkerComponent.prototype.update = function() {
					var self = this;

					// Microbatch handler
					self.microbatch.rtime = new Date();
					if(self.microbatch.timeout === false) {
						self.microbatch.timeout = true;
						setTimeout(eventEnd, self.microbatch.delta);
					}					
					function eventEnd() {
						if (new Date() - self.microbatch.rtime < self.microbatch.delta)
							setTimeout(eventEnd, self.microbatch.delta);
						else {
							self.microbatch.timeout = false;
							run();
						}               
					};
					
					function run() {
						var object = self.data;
						
						// Update the radius 
						if(scope.options.radius.indexOf('{') == -1)
							self.node.setRadius(scope.options.radius);
						else {
							var radius = eval('(' + scope.options.radius + ')');
							var radius = eval('(' + scope.options.radius + ')');											
							for(var idx = 0; idx < Object.keys(radius).length; idx++) {
								var rad = Object.keys(radius)[idx];
								if(radius[rad] == true) {
									self.node.setRadius(rad);
									break;
								}
							}											
						}									
											
						// Update the styles if a path marker
						if(scope.options.path != undefined) {
							var opts = $.extend({}, scope.options.path.options);
							var color = ctrl.getColor(self.tid);
							if(color != undefined) {
								opts.color = color;
								opts.fillColor = color;
							}
							else {
								opts.color = Map.getColor(opts, 'color', object);
								opts.fillColor = Map.getColor(opts, 'fillColor', object);
							}
							self.node.setStyle(opts);
						}
											
						// Update the popup
						Map.updatePopup(scope, self);
							
						// Update the marker and popup position
						Map.updatePosition(scope, self);
											
						// Update the title
						Map.updateTitle(scope, ctrl, self);
					}
				};
				
				CircleMarkerComponent.prototype.remove = function() { ctrl.removeComponent(this); };
				
				CircleMarkerComponent.prototype.updateVisibility = function(verify) {
					if(this.visible == true)
						this.node._path.style.display = '';
					else if(this.visible == false)
						this.node._path.style.display = 'none';
					if(verify != false)
						ctrl.verifyVisibility();
				};
			}
						
		}
	};	
});

app.directive('livemapPolyline', function($log, SimpleDateFormat, NumberFormat, Overlay, Map) {
	return {
		restrict: 'E',
		scope: {},
		transclude: true,
		require: '^livemapLayer',
		controller: function($scope) {
		},
		link: function(scope, element, attrs, ctrl, transclude) {
			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});

			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;
			
			// Read the options from attributes
			scope.options = {
				"disabled":			attrs.disabled == undefined || attrs.disabled.length == 0 ?
										false :
										eval(attrs.disabled), 
									// Optional. Disable the layer.
				
				"groupBy":			attrs.groupBy == undefined || attrs.groupBy.length == 0 ?
										[] :
										eval(attrs.groupBy),
									// Optional. Group by attribute of the object being passed. Array in the format
									//   [property1, property2, ...]
									//   Default is no grouping
									
				"groupProperties":	attrs.groupProperties == undefined || attrs.groupProperties.length == 0 ?
										[] :
										eval(attrs.groupProperties),
									// Optional. Properties from the object to apply to the group. Array in the format
									//   [property1, property2, ...]
									//   Default is only group by properties are copied. If different values for members occur
									//   then only the final value will be reported.
									
				"orderBy":			attrs.orderBy == undefined || attrs.orderBy.length == 0 ?
										[] :
										eval(attrs.orderBy),
									// Optional. Sort order to apply to records. Array in the format
									// [[property1, 'asc'|'desc'], [property2, 'asc'|'desc'], ...]
									// Default is no ordering

				"position":			attrs.position, 
									// Mandatory. Position for the point. Define as an array with object properties for lat and lon.
									//   Example 1: [object.lat, object.lon]
									//     This appllies to polylineMode of grouped. 
									//     Sets the vertex at the position object.lat and object.lon where object is the
									//     data row, and lat and lon are the properties of that object that point to latitude and 
									//     longitude respectively.
									//   Example 2: [[object.startLat, object.startLon],[object.endLat,object.endLon]]
									//     This applies only to polylineMode of single where each object defines a full polyline
									//     Sets the starting vertex at the position object.startLat and object.startLon and the 
									//     ending vertex at the position object.endLat and object.endLon, where object is the
									//     data row, and startLat, startLon, endLat, and endLon are all properties of that object 
									//     that point to starting latitude and longitude and ending latitude and longitude respectively.

				"polyline":			{},
									// Mandatory. Automatically generated for this directive type.
									
				"polylineMode":		attrs.polylineMode == undefined ? 'grouped' : attrs.polylineMode,
									// Optional. Polyline vertex mode. Valid values are: single or grouped. If omitted
									//   then grouped polyline will be used
									//   single 	- one object record is a single polyline, defining both start and end points
									//   grouped 	- one object record is a single vertex on the polyline, 
									
				"popup":			$(element).find('livemap-popup').length != 1 ?
										undefined :
										$(element).find('livemap-popup').html(),
									// Optional. Popup HTML content. Defined in a tag structure underneath this element.
									//   Example:
									//     <livemap-popup ng-non-bindable>
									//        <span>
									//		    <h3>Flight: {{::object.flightNumber}}</h3>
									//	     </span>
									//     </livemap-popup>
									// 
									//   See all notes above as for iconLabelContent. Same rules apply here.									
			};

			// Get path options
			scope.options.path = {
				options:	Map.getPathOptions(attrs)
			};

			// Add polyline options
			scope.options.polyline.options = {
				"smoothFactor":		attrs.smoothFactor == undefined || attrs.smoothFactor.length == 0 ?
										1.0 :
										eval(attrs.smoothFactor),
									// Optional. How much to simplify the polyline at each zoom level. Greater means better performance
									//   less means more accurate. Must be a number.
				
				"noClip":			attrs.noClip == undefined || attrs.noClip.length == 0 ?
										false :
										eval(attrs.noClip),
									// Optional. Disabled polyline clipping. Must be true or false.
			};
			
			// Create local data groups			
			scope.vertexGroupSet = {};
			scope.vertexGroupComponents = {};
			scope.vertexSet = {};
			scope.entitySet = {};
			
			// Process single actions
			var processSingleInsert = function(event, object) {
				var component = createComponentObject(object, 0);
				component.draw();
			};

			// Process group actions
			var processGroupedInsert = function(event, object) {
				var vertexGroup = Map.updateVertexGroup(Map.getVertexKey(object, scope.options), object, scope.options, scope.vertexGroupSet, scope.vertexSet);
				scope.vertexGroupSet[vertexGroup.tid] = vertexGroup;
				var component = scope.vertexGroupComponents[vertexGroup.tid];
				if(component == undefined) {
					component = createComponentObject(vertexGroup, 0);
					scope.vertexGroupComponents[vertexGroup.tid] = component;
					component.draw();
				}
				else
					component.update();
			};
			
			var processGroupedUpdate = function(event, object) {
				// Retrieve the vertex object from the vertexSet and update
				//   the parent vertexGroup component
				var vertex = scope.vertexSet[object.tid];
				if(vertex != undefined) {
					vertex.data = object;
					var vertexGroup = scope.vertexGroupSet[vertex.parent];
					var component = scope.vertexGroupComponents[vertexGroup.tid];

					// Iterate over all the group by properties and copy them all from the object
					//   to the vertexGroup
					for(var idx = 0; idx < scope.options.groupBy.length; idx++)
						vertexGroup[scope.options.groupBy[idx]] = object[scope.options.groupBy[idx]];	

					// Iterate over all the group properties and copy them all from the object to
					//   the vertexGroup
					for(var idx = 0; idx < scope.options.groupProperties.length; idx++)
						vertexGroup[scope.options.groupProperties[idx]] = object[scope.options.groupProperties[idx]];
					
					component.update();
				}
			};

			var processGroupedDelete = function(event, object) {
				// Retrieve the vertex from the vertexSet
				var vertex = scope.vertexSet[object.tid];
				if(vertex != undefined) {
					var vertexGroup = scope.vertexGroupSet[vertex.parent];
					
					// Remove the vertex from the vertexGroup and vertexSet
					vertexGroup.vertices.splice(vertexGroup.vertices.indexOf(vertex),1);
					delete scope.vertexSet[object.tid];
					
					// If the vertexGroup still has vertices, update, otherwise remove
					var component = scope.vertexGroupComponents[vertexGroup.tid];
					if(vertexGroup.vertices.length > 0)
						component.update();	
					else {
						component.remove();
						scope.vertexGroupComponents[vertexGroup.tid] = undefined;
					}
				}
			};
			
			// Process entity actions
			var processEntityInsert = function(evnet, object) {
				var parent = $.extend(true, {}, object);
				parent[ctrl.getEntity()] = undefined;
				
				// Timeout gives the service call time to load if its loaded from a different source
				if(object[ctrl.getEntity()] != undefined) {
					for(var idx = 0; idx < object[ctrl.getEntity()].length; idx++) {
						var entity = object[ctrl.getEntity()][idx];	
						entity.tid = object.tid;
						var vertexGroup = Map.updateVertexGroup(object.tid, entity, scope.options, scope.vertexGroupSet, scope.vertexSet);
						vertexGroup.parent = parent;
						scope.vertexGroupSet[vertexGroup.tid] = vertexGroup;
						scope.entitySet[object.tid] = vertexGroup;
						var component = scope.vertexGroupComponents[vertexGroup.tid];
						if(component == undefined) {
							component = createComponentObject(vertexGroup, idx);
							scope.vertexGroupComponents[vertexGroup.tid] = component;
							component.draw();
						}
						else {
							component.update();
						}
					}
				}
			};
			
			var processEntityUpdate = function(event, object) {
				var parent = $.extend(true, {}, object);
				parent[ctrl.getEntity()] = undefined;

				// Timeout gives the service call time to load if its loaded from a different source
				if(object[ctrl.getEntity()] != undefined) {
					for(var idx = 0; idx < object[ctrl.getEntity()].length; idx++) {
						var entity = object[ctrl.getEntity()][idx];	
						entity.tid = object.tid;
						var vertexGroup = Map.updateVertexGroup(object.tid, entity, scope.options, scope.vertexGroupSet, scope.vertexSet);
						vertexGroup.parent = parent;
						scope.vertexGroupSet[vertexGroup.tid] = vertexGroup;
						scope.entitySet[object.tid] = vertexGroup;
						var component = scope.vertexGroupComponents[vertexGroup.tid];
						if(component == undefined) {
							component = createComponentObject(vertexGroup, idx);
							scope.vertexGroupComponents[vertexGroup.tid] = component;
							component.draw();
						}
						else {
							component.update();
						}
					}
				}
			};
			
			var processEntityDelete = function(event, object) {
				var vertexGroup = scope.entitySet[object.tid];
				if(vertexGroup != undefined) {
					var component = scope.vertexGroupComponents[vertexGroup.tid];
					if(component != undefined) {
						component.remove();
						scope.vertexGroupComponents[vertexGroup.tid] = undefined;
					}
				}
			};
			
			// Listen for layer create new object event
			scope.$on('map:layer:insertObject', function(event, object) {
				if(ctrl.getEntity() == undefined)
					if(scope.options.polylineMode == 'single')
						processSingleInsert(event, object);
					else if(scope.options.polylineMode == 'grouped')
						processGroupedInsert(event, object);
				else
					processEntityInsert(event, object);
			});

			// Listen for layer update object event
			scope.$on('map:layer:updateObject', function(event, object) {
				if(ctrl.getEntity() == undefined) {
					if(scope.options.polylineMode == 'grouped')
						processGroupedUpdate(event, object);
				}
				else
					processEntityUpdate(event, object);
			});

			// Listen for layer delete object event
			scope.$on('map:layer:deleteObject', function(event, object) {
				if(ctrl.getEntity() == undefined) {
					if(scope.options.polylineMode == 'grouped')
						processGroupedDelete(event, object);
				}
				else 
					processEntityDelete(event, object);
			});

			// Listen for layer purge
			scope.$on('map:layer:purge', function(event) {
				scope.vertexGroupSet = {};
				scope.vertexGroupComponents = {};
				scope.vertexSet = {};
				scope.entitySet = {};
			});
			
			// Create a new map component for the given object
			var createComponentObject = function(object, idx) {
				var component = new PolylineComponent(object, idx);
				return component;
			};
			
			// Generate an array of lats and lons
			var generateSingleLatLons = function(object) {
				var latLng = [];
				var position = eval(scope.options.position);
				for(var idx = 0; idx < position.length; idx++)
					latLng.push(L.latLng(position[idx]));
				return latLng;
			};
			
			// Generate an array of lats and lons
			var generateGroupedLatLons = function(vertices) {	
				// Create a local array
				var localVertices = [];				
				angular.forEach(vertices, function(vertex) {
					localVertices.push(vertex);
				});

				// If ordering is enabled, sort by orderBy
				if(scope.options.orderBy.length > 0) {
					localVertices.sort(function (a, b) {
						var compare = function(a, b, dir, field) {
							if(!isNaN(parseInt(a.data[field])) && !isNaN(parseInt(b.data[field])))
								return dir * (parseInt(a.data[field]) == parseInt(b.data[field]) ? 0 : parseInt(a.data[field]) > parseInt(b.data[field]) ? 1 : -1);
							else if(!isNaN(parseFloat(a.data[field])) && !isNaN(parseFloat(b.data[field])))
								return dir * (parseFloat(a.data[field]) == parseFloat(b.data[field]) ? 0 : parseFloat(a.data[field]) > parseFloat(b.data[field]) ? 1 : -1);
							else
								return dir * (a.data[field] == b.data[field] ? 0 : a.data[field] > b.data[field] ? 1 : -1);
						}
					
						for(var idx = 0; idx < scope.options.orderBy.length; idx++) {
							var field = scope.options.orderBy[idx][0];
							var dir = scope.options.orderBy[idx][1] == 'desc' ? -1 : 1;
							var v = compare(a, b, dir, field);
							if(v != 0) return v;
						}
						return 0;
					});
				}
				
				// Create an array of lat/lon, loop over the ordered array
				var latLng = [];
				for(var idx = 0; idx < localVertices.length; idx++) {
					var object = localVertices[idx].data;
					var position = eval(scope.options.position);
					if(position[0] != null && position[1] != null)
						latLng.push(L.latLng(position));
				}
				return latLng;
			};
			
			// DEFINE OBJECTS
			// PolylineComponent object
			{
				function PolylineComponent(object, idx) {
					this.tid 		= object.tid;
					this.tidx 		= idx;
					this.data 		= object;
					this.node 		= undefined;
					this.selected 	= false;
					this.title 		= undefined;
					this.visible 	= true;
					this.popup		= undefined;
					
					this.hasPopup			= scope.options.popup != undefined;
					
					// Create a popup if specified
					if(this.hasPopup) {
						this.popup = Map.createPopup(scope, scope.options.popup);
					}
					
					// These properties are used for microbatching the updates
					this.microbatch = {
						rtime:		new Date(1, 1, 2000, 12,00,00),
						timeout:	false,
						delta:		0
					}					
				};

				PolylineComponent.prototype.bringToBack = function() { this.node.bringToBack();	};
				PolylineComponent.prototype.bringToFront = function() { this.node.bringToFront(); };
				
				PolylineComponent.prototype.draw = function() {
					var self = this;

					// Create the node
					var opts = $.extend({}, scope.options.polyline.options, scope.options.path.options);
					this.node = L.polyline([], opts);			
					
					// Add event handler
					this.node.on('click', function(event) {
						ctrl.select(self);
					});

					// Add component to the layer and update the node
					ctrl.addComponent(this);
					this.update();
				};
				
				PolylineComponent.prototype.update = function() {
					var self = this;

					// Microbatch handler
					self.microbatch.rtime = new Date();
					if(self.microbatch.timeout === false) {
						self.microbatch.timeout = true;
						setTimeout(eventEnd, self.microbatch.delta);
					}					
					function eventEnd() {
						if (new Date() - self.microbatch.rtime < self.microbatch.delta)
							setTimeout(eventEnd, self.microbatch.delta);
						else {
							self.microbatch.timeout = false;
							run();
						}               
					};
					
					function run() {
						// Update the styles if a path marker
						if(scope.options.path != undefined) {
							var opts = $.extend({}, scope.options.path.options);
							var color = ctrl.getColor(self.tid);
							if(color != undefined) {
								opts.color = color;
								opts.fillColor = color;
							}
							else {
								opts.color = Map.getColor(opts, 'color', self.data, self);
								opts.fillColor = Map.getColor(opts, 'fillColor', self.data, self);
								opts.weight = Map.getWeight(opts, 'weight', self.data, self);
							}
							self.node.setStyle(opts);
						}
						
						// Update the popup
						Map.updatePopup(scope, self);
		
						// Update the lats, lons and popup position
						if(scope.options.polylineMode == 'single')
							self.node.setLatLngs(generateSingleLatLons(self.data));
						else if(scope.options.polylineMode == 'grouped')
							self.node.setLatLngs(generateGroupedLatLons(self.data.vertices));
						
						// Update the title
						Map.updateTitle(scope, ctrl, self);
					}
				};
				
				PolylineComponent.prototype.remove = function() { ctrl.removeComponent(this); };
				
				PolylineComponent.prototype.updateVisibility = function(verify) {
					if(this.visible == true)
						this.node._path.style.display = '';
					else if(this.visible == false)
						this.node._path.style.display = 'none';
					if(verify != false)
						ctrl.verifyVisibility();
				};
			}
			
		}
	};	
});


// For SVG element attributes to avoid late binding error in Javascript console
angular.forEach(['x', 'y', 'r', 'width', 'height', 'transform', 'stroke', 'fill'], function(name) {
  var ngName = 'ng' + name[0].toUpperCase() + name.slice(1);
  app.directive(ngName, function() {
    return function(scope, element, attrs) {
      attrs.$observe(ngName, function(value) {
        attrs.$set(name, value); 
      })
    };
  });
});