package com.streambase.fx.adminclient;

import java.util.List;
import java.util.UUID;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;

public class TCPClientRequest {
    public static String DEFAULT_OUTPUT_FILE = "output.log";
    public static final String JSON_METHOD = "method";
    public static final String JSON_ARGS = "args";
    public static final String JSON_REQUESTID = "requestID";

    private boolean printToFile = false;
    private String fileName = DEFAULT_OUTPUT_FILE;
    private JSONObject jsonRequest = new JSONObject();
    private JSONObject jsonResponse;
    private String sendToTCPClientKey = "";
    private int sendToTCPClientIndex = -1;
    private String requestId = UUID.randomUUID().toString();
    private long startRequestTimeMillis = System.currentTimeMillis();

    public TCPClientRequest() {
        jsonRequest.put(JSON_REQUESTID, requestId);
    }

    public JSONObject getJSONObject() {
        return jsonRequest;
    }

    public void setPrintToFile(boolean printToFile) {
        this.printToFile = printToFile;
    }

    public boolean isPrintToFile() {
        return printToFile;
    }

    public void setFileName(String fileName) {
        this.fileName = fileName;
    }

    public String getFileName() {
        return fileName;
    }

    public void setSendToTCPClientKey(String host, String port) {
        this.sendToTCPClientKey = host + ":" + port;
    }

    public String getSendToTCPClientKey() {
        return sendToTCPClientKey;
    }

    public void setSendToTCPClientIndex(String sendToTCPClientIndex) {
        this.sendToTCPClientIndex = Integer.valueOf(sendToTCPClientIndex);
    }

    public int getSendToTCPClientIndex() {
        return sendToTCPClientIndex;
    }

    public String getRequestId() {
        return requestId;
    }

    public long getRequestLengthMillis() {
        return System.currentTimeMillis() - startRequestTimeMillis;
    }

    public void setMethod(UICommandHelper command) {
        jsonRequest.put(JSON_METHOD, command.getCommandName());
    }

    public void setArguments(int argumentsIndex, List<String> args) {
        JSONArray jsonArray = new JSONArray();
        if (argumentsIndex != -1 && argumentsIndex < args.size() - 1) {
            for (String arg : args.subList(argumentsIndex + 1, args.size())) {
                jsonArray.add(arg);
            }
        }
        jsonRequest.put(JSON_ARGS, jsonArray);
    }

    public JSONObject getJSONResponse() {
        return jsonResponse;
    }

    public void setJSONResponse(JSONObject jsonResponse) {
        this.jsonResponse = jsonResponse;
    }

    private JSONObject getMethodSuccess() {
        return jsonResponse.containsKey("methodSuccess") ? (JSONObject) jsonResponse.get("methodSuccess") : null;
    }

    public boolean isRequestSuccessful() {
        JSONObject methodSuccess = getMethodSuccess();
        if (methodSuccess != null && methodSuccess.containsKey("success")) {
            return methodSuccess.getString("success").equals("true");
        }
        return false;
    }

    public String getSuccessMessage() {
        JSONObject methodSuccess = getMethodSuccess();
        if (methodSuccess != null && methodSuccess.containsKey("message")) {
            return methodSuccess.getString("message");
        }
        return "";
    }

    public boolean hasSubscriptions() {
        return getSubscriptions() != null && !getSubscriptions().isEmpty();
    }

    public JSONArray getSubscriptions() {
        return jsonResponse.containsKey("subscriptions") ? (JSONArray) jsonResponse.get("subscriptions") : null;
    }

    public String getSubscriptionID(JSONObject subscription) {
        return subscription != null && subscription.containsKey("SubscriptionId") ? subscription.getString("SubscriptionId") : "";
    }

    public String getSubscriptionMDReqID(JSONObject subscription) {
        return subscription != null && subscription.containsKey("MDReqID") ? subscription.getString("MDReqID") : "";
    }

    public String getSubscriptionSide(JSONObject subscription) {
        return subscription != null && subscription.containsKey("Side") ? subscription.getString("Side") : "";
    }

    public String getSubscriptionMarketDepth(JSONObject subscription) {
        return subscription != null && subscription.containsKey("MarketDepth") ? subscription.getString("MarketDepth") : "";
    }

    public String getSubscriptionQuantity(JSONObject subscription) {
        return subscription != null && subscription.containsKey("Quantity") ? subscription.getString("Quantity") : "";
    }

    public String getSubscriptionQuantityFar(JSONObject subscription) {
        return subscription != null && subscription.containsKey("QuantityFar") ? subscription.getString("QuantityFar") : "";
    }

    public String getSubscriptionCurrency(JSONObject subscription) {
        return subscription != null && subscription.containsKey("Currency") ? subscription.getString("Currency") : "";
    }

    public String getSubscriptionDuration(JSONObject subscription) {
        return subscription != null && subscription.containsKey("Duration") ? subscription.getString("Duration") : "";
    }

    public String getSubscriptionExpiration(JSONObject subscription) {
        return subscription != null && subscription.containsKey("Expiration") ? subscription.getString("Expiration") : "";
    }

    public String getSubscriptionPendingUnsubscribeTimeout(JSONObject subscription) {
        return subscription != null && subscription.containsKey("PendingUnsubscribeTimeout") ? subscription.getString("PendingUnsubscribeTimeout") : "";
    }

    public boolean hasQuotes() {
        return getQuotes() != null && !getQuotes().isEmpty();
    }

    public JSONArray getQuotes() {
        return jsonResponse.containsKey("quotes") ? (JSONArray) jsonResponse.get("quotes") : null;
    }

    public String getQuoteID(JSONObject quote) {
        return quote != null && quote.containsKey("QuoteId") ? quote.getString("QuoteId") : "";
    }

    public String getQuoteSubscriptionId(JSONObject quote) {
        return quote != null && quote.containsKey("SubscriptionId") ? quote.getString("SubscriptionId") : "";
    }

    public String getQuoteMsgSeqNum(JSONObject quote) {
        return quote != null && quote.containsKey("MsgSeqNum") ? quote.getString("MsgSeqNum") : "";
    }

    public String getQuoteVenueName(JSONObject quote) {
        return quote != null && quote.containsKey("VenueName") ? quote.getString("VenueName") : "";
    }

    public String getQuoteSide(JSONObject quote) {
        return quote != null && quote.containsKey("Side") ? quote.getString("Side") : "";
    }

    public String getQuoteType(JSONObject quote) {
        return quote != null && quote.containsKey("QuoteType") ? quote.getString("QuoteType") : "";
    }

    public String getQuoteCurrency(JSONObject quote) {
        return quote != null && quote.containsKey("Currency") ? quote.getString("Currency") : "";
    }

    public String getQuotePrice(JSONObject quote) {
        return quote != null && quote.containsKey("Price") ? quote.getString("Price") : "";
    }

    public String getQuoteQuantity(JSONObject quote) {
        return quote != null && quote.containsKey("Quantity") ? quote.getString("Quantity") : "";
    }

    public String getQuoteQuantityFar(JSONObject quote) {
        return quote != null && quote.containsKey("QuantityFar") ? quote.getString("QuantityFar") : "";
    }

    public String getQuoteSpotRate(JSONObject quote) {
        return quote != null && quote.containsKey("SpotRate") ? quote.getString("SpotRate") : "";
    }

    public String getQuoteForwardPoints(JSONObject quote) {
        return quote != null && quote.containsKey("ForwardPoints") ? quote.getString("ForwardPoints") : "";
    }

    public String getQuoteForwardPointsFar(JSONObject quote) {
        return quote != null && quote.containsKey("ForwardPointsFar") ? quote.getString("ForwardPointsFar") : "";
    }

    public String getQuoteSettlDate(JSONObject quote) {
        return quote != null && quote.containsKey("SettlDate") ? quote.getString("SettlDate") : "";
    }

    public String getQuoteSettlDateFar(JSONObject quote) {
        return quote != null && quote.containsKey("SettlDateFar") ? quote.getString("SettlDateFar") : "";
    }

    public String getQuoteVenueTime(JSONObject quote) {
        return quote != null && quote.containsKey("VenueTime") ? quote.getString("VenueTime") : "";
    }

    public String getQuoteArrivalTime(JSONObject quote) {
        return quote != null && quote.containsKey("ArrivalTime") ? quote.getString("ArrivalTime") : "";
    }

    public String getQuoteAdapterNanos(JSONObject quote) {
        return quote != null && quote.containsKey("AdapterNanos") ? quote.getString("AdapterNanos") : "";
    }

    public String getQuoteHandlerNanos(JSONObject quote) {
        return quote != null && quote.containsKey("HandlerNanos") ? quote.getString("HandlerNanos") : "";
    }

    public String getQuoteExtendedField1(JSONObject quote) {
        return quote != null && quote.containsKey("ExtendedField1") ? quote.getString("ExtendedField1") : "";
    }

    public JSONObject getInstrument(JSONObject instrumentParent) {
        return instrumentParent.containsKey("Instrument") ? (JSONObject) instrumentParent.get("Instrument") : null;
    }

    public String getInstrumentType(JSONObject instrument) {
        return instrument != null && instrument.containsKey("InstrumentType") ? instrument.getString("InstrumentType") : "";
    }

    public String getInstrumentCurrency1(JSONObject instrument) {
        return instrument != null && instrument.containsKey("Currency1") ? instrument.getString("Currency1") : "";
    }

    public String getInstrumentCurrency2(JSONObject instrument) {
        return instrument != null && instrument.containsKey("Currency2") ? instrument.getString("Currency2") : "";
    }

    public String getInstrumentSettlementDate(JSONObject instrument) {
        return instrument != null && instrument.containsKey("SettlementDate") ? instrument.getString("SettlementDate") : "";
    }

    public String getInstrumentSettlementDateFar(JSONObject instrument) {
        return instrument != null && instrument.containsKey("SettlementDateFar") ? instrument.getString("SettlementDateFar") : "";
    }

    public String getInstrumentTenor(JSONObject instrument) {
        return instrument != null && instrument.containsKey("Tenor") ? instrument.getString("Tenor") : "";
    }

    public String getInstrumentTenorFar(JSONObject instrument) {
        return instrument != null && instrument.containsKey("TenorFar") ? instrument.getString("TenorFar") : "";
    }

    public JSONArray getConfig() {
        return jsonResponse.containsKey("config") ? (JSONArray) jsonResponse.get("config") : null;
    }

    public String getConfigName(JSONObject config) {
        return config != null && config.containsKey("Name") ? config.getString("Name") : "";
    }

    public String getConfigValue(JSONObject config) {
        return config != null && config.containsKey("Value") ? config.getString("Value") : "";
    }

    public String getLogFilePath() {
        return jsonResponse.containsKey("logFilePath") ? jsonResponse.getString("logFilePath") : "";
    }

    public JSONObject getHandler() {
        return jsonResponse.containsKey("handler") ? (JSONObject) jsonResponse.get("handler") : null;
    }

    public String getHandlerVenue() {
        return getHandler() != null && getHandler().containsKey("venue") ? getHandler().getString("venue") : "";
    }

    public String getHandlerSession() {
        return getHandler() != null && getHandler().containsKey("session") ? getHandler().getString("session") : "";
    }

    public String getHandlerStatus() {
        return getHandler() != null && getHandler().containsKey("status") ? getHandler().getString("status") : "";
    }

    public String getHandlerVersion() {
        return getHandler() != null && getHandler().containsKey("version") ? getHandler().getString("version") : "";
    }
}
