package com.streambase.fx.adminclient;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;

/*
 * A class used for formating and printing
 * output strings, both to files and to STDOUT
 * Notably this extracts the JAXB generated class fields
 * into a human readable string
 */
public class UIUtil {

    public static final String HOST_COMMAND = "-h";
    public static final String PORT_COMMAND = "-p";
    public static final String REFERENCE_COMMAND = "-r";
    public static final String DEBUG_COMMAND = "-debug";
    public static final String ARGUMENTS_COMMAND = "-a";
    public static final String FILE_COMMAND = "-f";
    private static final String DEST = "dest";
    private static final Object ARGS = "args";
    static final String WHITESPACE = "\\s+";
    static final String LIST_CONNECTIONS = "listconnections";
    static final String HELP = "help";
    static final String SHORT_HELP = "h";
    static final String SHORT_LIST_CONNECTIONS = "lc";
    static final String QUIT = "quit";
    static final String CONTINUE = "continue";
    static final String PING = "ping";
    static final String SHORT_PING = "p";
    static final String SHORT_QUIT = "q";
    static final UICommandHelper[] LOCAL_COMMANDS = new UICommandHelper[] { new UICommandHelper(HELP, SHORT_HELP), new UICommandHelper(QUIT, SHORT_QUIT),
            new UICommandHelper(LIST_CONNECTIONS, SHORT_LIST_CONNECTIONS), new UICommandHelper(PING, SHORT_PING) };

    static final String RETRIEVE_HANDLERS = "handlers";
    static final String SHORT_HANDLERS = "hs";
    static final String RETRIEVE_SUBSCRIPTIONS = "subscriptions";
    static final String SHORT_SUBSCRIPTIONS = "subs";
    static final String RETRIEVE_SNAPSHOT = "snapshot";
    static final String SHORT_SNAPSHOT = "snap";
    static final String RETRIEVE_LOGS = "logs";
    static final String SHORT_LOGS = "l";
    static final String RETRIEVE_CONFIG = "config";
    static final String SHORT_CONFIG = "cfg";
    static final String SUBSCRIBE = "subscribe";
    static final String SHORT_SUBSCRIBE = "sub";
    static final String UNSUBSCRIBE = "unsubscribe";
    static final String SHORT_UNSUBSCRIBE = "unsub";
    static final String SESSION_BOUNCE = "bounce";
    static final String SHORT_BOUNCE = "b";
    static final UICommandHelper[] COMMANDS = new UICommandHelper[] { new UICommandHelper(RETRIEVE_HANDLERS, SHORT_HANDLERS),
            new UICommandHelper(RETRIEVE_SUBSCRIPTIONS, SHORT_SUBSCRIPTIONS), new UICommandHelper(RETRIEVE_SNAPSHOT, SHORT_SNAPSHOT),
            new UICommandHelper(RETRIEVE_LOGS, SHORT_LOGS), new UICommandHelper(RETRIEVE_CONFIG, SHORT_CONFIG),
            new UICommandHelper(SUBSCRIBE, SHORT_SUBSCRIBE), new UICommandHelper(UNSUBSCRIBE, SHORT_UNSUBSCRIBE),
            new UICommandHelper(SESSION_BOUNCE, SHORT_BOUNCE) };

    static final String INPUT_FORM = "cmd [" + HOST_COMMAND + " [host] " + PORT_COMMAND + " [port] || " + REFERENCE_COMMAND + " [ref#]] [" + ARGUMENTS_COMMAND
            + " args...] [" + FILE_COMMAND + " [filename]]";
    static final String INVALID_USAGE = "Improper usage, requires input of the form:\n" + INPUT_FORM + "\nType 'help' for more information";

    static final String VALID_COMMAND_LIST = "The following is a list of valid commands:\n" + RETRIEVE_HANDLERS + " (" + SHORT_HANDLERS + ")\n"
            + RETRIEVE_SUBSCRIPTIONS + " (" + SHORT_SUBSCRIPTIONS + ")\n" + RETRIEVE_SNAPSHOT + " (" + SHORT_SNAPSHOT + ")\n" + RETRIEVE_LOGS + " ("
            + SHORT_LOGS + ")\n" + RETRIEVE_CONFIG + " (" + SHORT_CONFIG + ")\n" + SUBSCRIBE + " (" + SHORT_SUBSCRIBE + ")\n" + UNSUBSCRIBE + " ("
            + SHORT_UNSUBSCRIBE + ")\n" + SESSION_BOUNCE + " (" + SHORT_BOUNCE + ")\n";

    static final String REFERENCE = "%";

    // ERROR MESSAGES
    private static final String MISSING_DEST_ARGS = "The Destination Information supplied was incomplete.\n" + "It must match one of the following formats:\n"
            + "[no args]\n" + "" + REFERENCE_COMMAND + " [reference#] where reference# is the reference number of an active connection." + "" + HOST_COMMAND
            + " [host] " + PORT_COMMAND + " [port] where host and port are the host and port values of an active connection (both must be supplied).";

    private static final String MISSING_ARGS_FORMAT = "The required number of arguments for %s is %d.\nYou only supplied %d.";

    public static boolean printStrToFile(String obj, String filepath) {
        try {
            File f = new File(filepath);
            if (f.exists()) {
                if (!f.canWrite()) {
                    return false;
                }
            } else {
                f.createNewFile();
            }
            FileWriter fw = new FileWriter(f.getAbsoluteFile(), true);
            BufferedWriter bw = new BufferedWriter(fw);
            bw.append(obj);
            bw.close();
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    public static void printTCPClientResponse(TCPClient tcpClient, TCPClientRequest tcpClientRequest) {
        printTCPClientResponse(tcpClient, tcpClientRequest, 0);
    }

    public static void printTCPClientResponse(TCPClient tcpClient, TCPClientRequest tcpClientRequest, int tabs) {
        try {
            String tabString = makeTabs(tabs);
            String output = tabString;
            output += "#Response from: " + tcpClient.getDisplayName() + "\n";
            output += UIUtil.printSuccessMessageToString(tcpClientRequest, tabs + 1) + "\n";
            output += UIUtil.printHandlerToString(tcpClientRequest, tabs + 1);
            output += UIUtil.printSubriptionsToString(tcpClientRequest, tabs + 1);
            output += UIUtil.printConfigToString(tcpClientRequest, tabs + 1);
            output += UIUtil.printLogFileListToString(tcpClientRequest, tabs + 1);
            output += UIUtil.printQuotesToString(tcpClientRequest, tabs + 1);
            if (tcpClientRequest.isPrintToFile()) {
                UIUtil.printStrToFile(output, tcpClient.getFileKey() + "_" + tcpClientRequest.getFileName());
            } else {
                System.out.println(output);
            }
        } catch (Exception e) {
            System.err.println(String.format("Failed to print request information for %s: %s", tcpClient.getDisplayName(), e));
        }
    }

    private static String printHandlerToString(TCPClientRequest tcpClientRequest, int tabs) {
        if (tcpClientRequest.getHandler() == null) {
            return "";
        }
        String tabString = makeTabs(tabs);
        String output = tabString;
        output += "Venue:    %s\n" + tabString;
        output += "Session:     %s\n" + tabString;
        output += "Status:   %s\n" + tabString;
        output += "Version:  %s";
        return String.format(output, tcpClientRequest.getHandlerVenue(), tcpClientRequest.getHandlerSession(), tcpClientRequest.getHandlerStatus(),
                             tcpClientRequest.getHandlerVersion());
    }

    private static String printSuccessMessageToString(TCPClientRequest tcpClientRequest, int tabs) {
        String tabString = "#" + makeTabs(tabs);
        String output = tabString;
        output += "Operation " + (tcpClientRequest.isRequestSuccessful() ? "SUCCEEDED\n" : "FAILED\n") + tabString;
        output += "Message:\n" + tabString + "\t";
        output += tcpClientRequest.getSuccessMessage();
        return output;
    }

    private static String printSubriptionsToString(TCPClientRequest tcpClientRequest, int tabs) {
        if (!tcpClientRequest.hasSubscriptions()) {
            return "";
        }
        String output = "";
        String tabString = makeTabs(tabs);
        int i = 0;
        JSONArray subscriptions = tcpClientRequest.getSubscriptions();
        for (Object subscription : subscriptions) {
            if (subscription instanceof JSONObject) {
                output += tabString;
                output += "[" + i + "]:\n";
                String s = printSubscriptionToString(tcpClientRequest, (JSONObject) subscription, tabs + 1);
                output += s + ((i < subscriptions.size() - 1) ? "\n" : "");
                i += 1;
            }
        }
        return output;
    }

    private static String printSubscriptionToString(TCPClientRequest tcpClientRequest, JSONObject subscription, int tabs) {
        String tabString = makeTabs(tabs);
        String output = tabString;
        output += "Subscription ID:              %s\n" + tabString;
        output += "MD Req ID:                    %s\n" + tabString;
        output += "Instrument:\n%s\n" + tabString;
        output += "Side:                         %s\n" + tabString;
        output += "Market Depth:                 %s\n" + tabString;
        output += "Quantity:                     %s\n" + tabString;
        output += "Quantity Far:                 %s\n" + tabString;
        output += "Currency:                     %s\n" + tabString;
        output += "Duration:                     %s\n" + tabString;
        output += "Expiration:                   %s\n" + tabString;
        output += "Pending Unsubscribe Timeout:  %s";
        return String.format(output, tcpClientRequest.getSubscriptionID(subscription), tcpClientRequest.getSubscriptionMDReqID(subscription),
                             printInstrumentToStr(tcpClientRequest, subscription, tabs + 1), tcpClientRequest.getSubscriptionSide(subscription),
                             tcpClientRequest.getSubscriptionMarketDepth(subscription), tcpClientRequest.getSubscriptionQuantity(subscription),
                             tcpClientRequest.getSubscriptionQuantityFar(subscription), tcpClientRequest.getSubscriptionCurrency(subscription),
                             tcpClientRequest.getSubscriptionDuration(subscription), tcpClientRequest.getSubscriptionExpiration(subscription),
                             tcpClientRequest.getSubscriptionPendingUnsubscribeTimeout(subscription));
    }

    private static String printInstrumentToStr(TCPClientRequest tcpClientRequest, JSONObject subscription, int tabs) {
        JSONObject instrument = tcpClientRequest.getInstrument(subscription);
        if (instrument == null) {
            return "";
        }
        String output = "";
        String tabString = makeTabs(tabs);
        output += tabString;
        output += "Instrument Type:      %s\n" + tabString;
        output += "Currency 1:           %s\n" + tabString;
        output += "Currency 2:           %s\n" + tabString;
        output += "Settlement Date:      %s\n" + tabString;
        output += "Settlement Date Far:  %s\n" + tabString;
        output += "Tenor:                %s\n" + tabString;
        output += "Tenor Far:            %s";

        return String.format(output, tcpClientRequest.getInstrumentType(instrument), tcpClientRequest.getInstrumentCurrency1(instrument),
                             tcpClientRequest.getInstrumentCurrency2(instrument), tcpClientRequest.getInstrumentSettlementDate(instrument),
                             tcpClientRequest.getInstrumentSettlementDateFar(instrument), tcpClientRequest.getInstrumentTenor(instrument),
                             tcpClientRequest.getInstrumentTenorFar(instrument));
    }

    private static String printConfigToString(TCPClientRequest tcpClientRequest, int tabs) {
        if (tcpClientRequest.getConfig() == null || tcpClientRequest.getConfig().size() <= 0) {
            return "";
        }
        String output = "";
        String tabString = makeTabs(tabs);
        output += tabString;
        output += "#Configuration:\n";
        JSONArray configSettings = tcpClientRequest.getConfig();
        if (configSettings != null) {
            String sunTabString = makeTabs(tabs + 1);
            for (Object config : configSettings) {
                output += sunTabString;
                output += tcpClientRequest.getConfigName((JSONObject) config) + "=" + tcpClientRequest.getConfigValue((JSONObject) config) + "\n";
            }
        }
        return output;
    }

    private static String printLogFileListToString(TCPClientRequest tcpClientRequest, int tabs) {
        if (tcpClientRequest.getLogFilePath().length() > 0) {
            String t = makeTabs(tabs);
            return t + "#Log Location:\n" + t + tcpClientRequest.getLogFilePath() + "\n";
        }
        return "";
    }

    public static String printQuotesToString(TCPClientRequest tcpClientRequest, int tabs) {
        if (!tcpClientRequest.hasQuotes()) {
            return "";
        }
        String output = "";
        String tabString = makeTabs(tabs);
        int i = 0;
        JSONArray quotes = tcpClientRequest.getQuotes();
        for (Object quote : quotes) {
            if (quote instanceof JSONObject) {
                output += tabString;
                output += "[" + i + "]:\n";
                String s = printQuoteToString(tcpClientRequest, (JSONObject) quote, tabs + 1);
                output += s + ((i < quotes.size() - 1) ? "\n" : "");
                i += 1;
            }
        }
        return output;
    }

    public static String printQuoteToString(TCPClientRequest tcpClientRequest, JSONObject quote, int tabs) {
        String tabString = makeTabs(tabs);
        String output = tabString;
        output += "Subscription ID:      %s\n" + tabString;
        output += "MsgSeqNum:            %s\n" + tabString;
        output += "Instrument:\n%s\n" + tabString;
        output += "Venue Name:           %s\n" + tabString;
        output += "Quote ID:             %s\n" + tabString;
        output += "Side:                 %s\n" + tabString;
        output += "Type:                 %s\n" + tabString;
        output += "Currency:             %s\n" + tabString;
        output += "Price:                %s\n" + tabString;
        output += "Quantity:             %s\n" + tabString;
        output += "Quantity Far:         %s\n" + tabString;
        output += "Spot Rate:            %s\n" + tabString;
        output += "Forward Points:       %s\n" + tabString;
        output += "Forward Points Far:   %s\n" + tabString;
        output += "Settlement Date:      %s\n" + tabString;
        output += "Settlement Date Far:  %s\n" + tabString;
        output += "Venue Time:           %s\n" + tabString;
        output += "Arrival Time:         %s\n" + tabString;
        output += "Adapter Nanos:        %s\n" + tabString;
        output += "Handler Nanos:        %s\n" + tabString;
        output += "Extended Field 1:     %s" + tabString;

        return String.format(output, tcpClientRequest.getQuoteSubscriptionId(quote), tcpClientRequest.getQuoteMsgSeqNum(quote),
                             printInstrumentToStr(tcpClientRequest, quote, tabs + 1), tcpClientRequest.getQuoteVenueName(quote),
                             tcpClientRequest.getQuoteID(quote), tcpClientRequest.getQuoteSide(quote), tcpClientRequest.getQuoteType(quote),
                             tcpClientRequest.getQuoteCurrency(quote), tcpClientRequest.getQuotePrice(quote), tcpClientRequest.getQuoteQuantity(quote),
                             tcpClientRequest.getQuoteQuantityFar(quote), tcpClientRequest.getQuoteSpotRate(quote),
                             tcpClientRequest.getQuoteForwardPoints(quote), tcpClientRequest.getQuoteForwardPointsFar(quote),
                             tcpClientRequest.getQuoteSettlDate(quote), tcpClientRequest.getQuoteSettlDateFar(quote),
                             tcpClientRequest.getQuoteVenueTime(quote), tcpClientRequest.getQuoteArrivalTime(quote),
                             tcpClientRequest.getQuoteAdapterNanos(quote), tcpClientRequest.getQuoteHandlerNanos(quote),
                             tcpClientRequest.getQuoteExtendedField1(quote));
    }

    private static String makeTabs(int tabs) {
        String t = "";
        for (int i = 0; i < tabs; i++) {
            t += "\t";
        }
        return t;
    }

    /*
     * Prints a subset of the help info, which relates to:
     * 
     * @Param topic
     */
    public static void printHelpInfo(String topic) {
        // TODO Auto-generated method stub
        if (topic.equals(DEST)) {
            System.out.println("\n" + destArgExplanation(1) + "\n");
        } else if (topic.equals(QUIT) || topic.equals(SHORT_QUIT)) {
            System.out.println("\n\tPrompts user for confirmation; if given, it then terminates the program.\n");
        } else if (topic.equals(PING) || topic.equals(SHORT_PING)) {
            System.out.println("\n" + pingExplanation(1) + "\n");
        } else if (topic.equals(LIST_CONNECTIONS) || topic.equals(SHORT_LIST_CONNECTIONS)) {
            System.out.println("\n\tPrints a list of all active connections to SB FX Servers, including:\n\t"
                    + "Local Reference Number, IP Address and Ports\n");
        } else if (topic.equals(ARGS)) {
            System.out.println("\n" + inputFormatExplanation(1) + "\n");
        } else if (topic.equals(SUBSCRIBE) || topic.equals(SHORT_SUBSCRIBE)) {
            System.out.println("\n" + subscribeInfo(1) + "\n");
        } else if (topic.equals(UNSUBSCRIBE) || topic.equals(SHORT_UNSUBSCRIBE)) {
            System.out.println("\n" + unsubscribeInfo(1) + "\n");
        } else if (topic.equals(RETRIEVE_HANDLERS) || topic.equals(SHORT_HANDLERS)) {
            System.out.println("\n" + handlerInfo(1) + "\n");
        } else if (topic.equals(RETRIEVE_CONFIG) || topic.equals(SHORT_CONFIG)) {
            System.out.println("\n" + configInfo(1) + "\n");
        } else if (topic.equals(RETRIEVE_LOGS) || topic.equals(SHORT_LOGS)) {
            System.out.println("\n" + logInfo(1) + "\n");
        } else if (topic.equals(RETRIEVE_SUBSCRIPTIONS) || topic.equals(SHORT_SUBSCRIPTIONS)) {
            System.out.println("\n" + subsInfo(1) + "\n");
        } else if (topic.equals(RETRIEVE_SNAPSHOT) || topic.equals(SHORT_SNAPSHOT)) {
            System.out.println("\n" + snapInfo(1) + "\n");
        } else if (topic.equals(SESSION_BOUNCE) || topic.equals(SHORT_BOUNCE)) {
            System.out.println("\n" + bounceInfo(1) + "\n");
        } else {
            System.out.println(topic + " is not a known Help Topic.\nThis is the list of valid help topics:\n\n" + printHelpTopics(1) + "\n");
        }
    }

    private static String printHelpTopics(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        output += t + "General Argument Info:\n";
        output += t + "//////////////////////\n";
        output += t + DEST + "\n" + t + ARGS + "\n";
        output += t + "//////////////////////\n" + t + "Local Command Info\n";
        output += t + "//////////////////////\n";
        output += t + QUIT + "\n" + t + PING + "\n" + t + LIST_CONNECTIONS + "\n";
        output += t + "//////////////////////\n" + t + "Admin Command Info\n";
        output += t + "//////////////////////\n";
        output += t + RETRIEVE_HANDLERS + "\n";
        output += t + RETRIEVE_SUBSCRIPTIONS + "\n";
        output += t + RETRIEVE_SNAPSHOT + "\n";
        output += t + RETRIEVE_CONFIG + "\n";
        output += t + RETRIEVE_LOGS + "\n";
        output += t + SUBSCRIBE + "\n";
        output += t + UNSUBSCRIBE + "\n";
        output += t + SESSION_BOUNCE + "\n";
        return output;
    }

    private static String bounceInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        final String space5 = "         ";
        output += t;
        output += SESSION_BOUNCE + " : ";
        output += "This command requires either 0 or 3 arguments. If no arguments are provided and the FX adapter has multiple FX sessions,\n" + t + space5
                + "this command will fail. The 3 arguements used to identify a session are BeginString, SenderCompID and TargetCompID.\n" + t + space5
                + "Bouncing a FX session will not reset sequence numbers.\n\n";
        output += t + "E.g. bounce -ref 0\n";
        output += t + t + "Will attempt to bounce the only FX session on the handler at connection 0.\n\n";
        output += t + "     bounce -ref 0 " + ARGUMENTS_COMMAND + " FIX.4.4\n";
        output += t + t + "Is an invalid command as it only provides one of the FX session identifiers, but not all three.\n\n";
        output += t + "     bounce -ref 0 " + ARGUMENTS_COMMAND + " FIX.4.4 SENDER_ID TARGET_ID\n";
        output += t + t + "Will attempt to bounce the FX session specified by these parameters on the handler at connection 0";
        return output;
    }

    private static String snapInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        output += t;
        output += RETRIEVE_SNAPSHOT + " : ";
        output += "The snapshot command requires exactly 1 argument, which is the Subscription ID from which to take a snapshot.\n\n";
        output += t + "E.g. snapshot -ref 0 " + ARGUMENTS_COMMAND + " R-1-1372341719901\n";
        output += t + t + "Retrieves a snapshot of the subscription whose subscriptionID is R-1-1372341719901\n\n";
        output += t + "     snapshot -ref 0 " + ARGUMENTS_COMMAND + "\n";
        output += t + t + "Is an invalid command, since it doesn't provide the requisite subscriptionID.";
        return output;
    }

    private static String subsInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        final String space = "                ";
        output += t;
        output += RETRIEVE_SUBSCRIPTIONS + " : ";
        output += "This command can have 0 to 2 arguments. With no arguments it will return all active subscriptions.\n" + t + space
                + "With 1 it will filter the subscriptions by their Currency value.\n" + t + space
                + "With 2 it will filter them by their Instrument's currency pair denoted by those two arguments.\n\n";
        output += t + "E.g. subs\n";
        output += t + t + "Sends a request to list all active subscriptions\n\n";
        output += t + "     subs " + ARGUMENTS_COMMAND + " USA\n";
        output += t + t + "Sends a request to list all subscriptions whose Currency value is 'USA'\n\n";
        output += t + "     subs " + ARGUMENTS_COMMAND + " USA EUR\n";
        output += t + t + "Sends a request to list all subscriptions\n" + t + t
                + "whose Instrument's Currency1 and Currency2 values are 'USA' and 'EUR' (order is irrelevant)";
        return output;
    }

    private static String logInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        output += t;
        output += RETRIEVE_LOGS + " : ";
        output += "This expects no arguments. It returns a list of paths to log files for the FX Adapter.\n\n";
        output += t + "E.g. logs -host 10.1.255.255 -port 11111\n";
        output += t + t + "This requests logs from the handler which is located at IP address 10.1.255.255 and port 11111.";
        return output;
    }

    private static String configInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        final String space2 = "         ";
        output += t;
        output += RETRIEVE_CONFIG + " : ";
        output += "This expects no arguments. Configuration key, value pairs will be printed in the same format as a config file.\n" + t + space2
                + "If the print to file option is enabled, the destination file will be a usable configuaration file.\n\n";
        output += t + "E.g. config -host 10.1.255.255 -post 11111 " + FILE_COMMAND + " logs.log\n\n";
        output += t + t + "Sends a command to retrieve config information from the handler which is located at IP address 10.1.255.255,\n" + t + t
                + "port 11111, and writes the response to the file '10.1.255.255:11111_logs.log'";
        return output;
    }

    private static String handlerInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        output += t;
        output += RETRIEVE_HANDLERS + " : ";
        output += "This command expects no arguemnts. It returns the handler status, version info and the venue name.\n\n";
        output += t + "E.g. : " + RETRIEVE_HANDLERS;
        return output;
    }

    private static String unsubscribeInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        final String space4 = "              ";
        output += t;
        output += UNSUBSCRIBE + " : ";
        output += "This command requires exactly 1 argument which is the subscription ID of the subscription to unsubscribe from.\n" + t + space4
                + "This command will return a status and message, which reports only on the success or failure of the submission of the command.\n" + t
                + space4 + "If the handler rejects the request, it will not be reported by this response.\n\n";
        output += t;
        output += "E.g. : unsubscribe -ref 0 " + ARGUMENTS_COMMAND + " R-1-1372341719901\n" + t + t
                + "Sends an unsubscribe command for the subscription whose subscriptionID is R-1-1372341719901.\n\n";
        output += t + "       unsubscribe -ref 0 " + ARGUMENTS_COMMAND + "\n" + t + t
                + "Is an invalid command, since it does not provide the requisite subscriptionID.";
        return output;
    }

    private static String subscribeInfo(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        final String space3 = "            ";
        output += t;
        output += SUBSCRIBE + " : ";
        output += "This command requires the same fields as the initial subscription file.\n" + t + space3
                + "The expected order of arguments is as follows:\n" + t + t + space3
                + "InstrumentType,Currency1,Currency2,SettlementDate,SettlementDateFar,Tenor,TenorFar,Side,MarketDepth,Quantity,QuantityFar,Duration,Currency\n" + t + t + space3
                + "Any values which you do not wish to specify should be listed as 'null'.\n\n" + t + space3
                + "This command will return a status and message, which reports only on the success or failure of the submission of the command.\n" + t
                + space3 + "If the handler rejects the request, it will not be reported by this response.\n\n";
        output += t;
        output += "E.g. : subscribe -ref 0 " + ARGUMENTS_COMMAND + " SPOT EUR USD null null null null null 0 null null null null\n\n";
        output += t;
        output += "Sends the command to create a subscription with no values specified except of type SPOT for the EUR/USD currency pair.";
        return output;
    }

    private static String pingExplanation(int i) {
        String output = "";
        String t = makeTabs(i);
        output += t;
        output += "The " + PING + " command causes the client to 'ping' the multicast group,\n" + t
                + "which will update the connection list with any new FX servers and will remove dead connections.\n\n";
        output += t;
        output += "The " + PING + " command is automatically called by the client every " + UDPListenerThread.DEFAULT_SOCKET_TIMEOUT / 1000 + " seconds \n";
        output += t;
        output += "In addition, the user is allowed to issue it manually (this will not affect periodic pings in anyway).\n";
        return output;
    }

    public static void printHelpInfo() {
        System.out.println("\nFor more info on any of these commands,\n" + "Type 'help [command]'\nType 'help dest' for information on the [" + HOST_COMMAND
                + " " + PORT_COMMAND + " || " + REFERENCE_COMMAND + "] arguments\n" + "Type 'help args' for information on the [" + ARGUMENTS_COMMAND
                + " args...] [" + FILE_COMMAND + " [filename]]");
        System.out.println();
        System.out.println("Client Usage: " + INPUT_FORM);
        System.out.println();
        System.out.println("The list of local commands is:");
        System.out.println(printCMDListToStr(1, LOCAL_COMMANDS));
        System.out.println();
        System.out.println("The list of administration commands is:");
        System.out.println(printCMDListToStr(1, COMMANDS));
    }

    private static String destArgExplanation(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        output += t;
        output += "The destination arguments: [" + HOST_COMMAND + " " + PORT_COMMAND + " || " + REFERENCE_COMMAND + "] are optional.\n";
        output += t;
        output += "If none are specified, the command will be sent to every connected FX Handler.\n\n";
        output += t;
        output += "E.g. : unsubscribe " + ARGUMENTS_COMMAND + " R-1-1372341719901\n";
        output += t + "sends an unsubscribe command for subscription R-1-1372341719901 to every connected handler.\n\n";
        output += t;
        output += "If the first argument includes the '" + REFERENCE_COMMAND + "', then the destination\n" + t
                + "is assumed to be a reference to an existing connection.\n\n" + t
                + "The connection numbers can be found with a call to 'listconnections' (lc).\n\n";
        output += t + "E.g. : unsubscribe " + REFERENCE_COMMAND + " 3 " + ARGUMENTS_COMMAND + " R-1-1372341719901\n";
        output += t + "sends the same unsubscribe command, but only to the handler connected through connection 3.\n\n";
        output += t + "Otherwise, the " + HOST_COMMAND + " and " + PORT_COMMAND + " commands can be specified for the remote machine,\n" + t
                + "through which reading and writing to said machine occur.\n" + t + "This information can also be found through the use of (lc).\n\n";
        output += t + "E.g. : unsubsribe " + HOST_COMMAND + " 10.1.255.255 " + PORT_COMMAND + " 11111 " + ARGUMENTS_COMMAND + " R-1-1372341719901\n" + t
                + "N.B. this must specify a connection in the Connection List to succeed\n\n";
        output += t + "Should a connection be dropped, the updated connection list\n" + t + "will be printed automatically for convenience\n\n" + t
                + "N.B. (some reference numbers may change if and only if at least 1 connection is dropped).";
        return output;
    }

    private static String inputFormatExplanation(int tabs) {
        String output = "";
        String t = makeTabs(tabs);
        output += t;
        output += "Usage: " + INPUT_FORM + "\n";
        output += t;
        output += "Any admin command must follow format of the Usage line above.\n\n";
        output += t;
        output += "The primary arguments to the command are considered to be every value between '" + ARGUMENTS_COMMAND + "' and '" + FILE_COMMAND + "', if '"
                + FILE_COMMAND + "' is included, otherwise every value after '" + ARGUMENTS_COMMAND + "'.\n";
        output += t;
        output += "Therefore any command which requires arguements also requires the symbol '" + ARGUMENTS_COMMAND + "' before they are listed.\n\n";
        output += t;
        output += "The '" + FILE_COMMAND
                + "' option is always optional; including this signals that the output should be written to a file, instead of StdOut.\n";
        output += t;
        output += "The next argument following '" + FILE_COMMAND + "' denotes the file name. If no filename is provided a default filename of "
                + TCPClientRequest.DEFAULT_OUTPUT_FILE + " is used.\n";
        output += t;
        output += "The actual filename will be a concatenation of the host_port_filename from which the response arrived.\n";
        output += t;
        output += "If this file does not exist, it will be created; if it does exist, then the output from this command will be appended to that file.\n\n";
        output += t;
        output += "E.g.\n" + t;
        output += "The command: unsubscribe " + ARGUMENTS_COMMAND + " R-1-1372341719901 , is valid and prints to StdOut\n";
        output += t + "The command: unsubscribe R-1-1372341719901 , is invalid since unsubscribe requries arguments and '" + ARGUMENTS_COMMAND
                + "' isn't included\n";
        output += t + "The command: unsubscribe " + ARGUMENTS_COMMAND + " R-1-1372341719901 " + FILE_COMMAND
                + " result.log , is valid and prints to file(s) with the name 'host_port_result.log'\n";
        output += t + "The command: unsubscribe " + FILE_COMMAND + " " + ARGUMENTS_COMMAND
                + " R-1-1372341719901 , is valid and prints to file(s) with the name 'host_port_" + TCPClientRequest.DEFAULT_OUTPUT_FILE + "'\n";
        output += t + "The command: unsubscribe " + ARGUMENTS_COMMAND + " R-1-1372341719901 " + FILE_COMMAND
                + " , is valid and prints to file(s) with the name 'host_port_" + TCPClientRequest.DEFAULT_OUTPUT_FILE + "'";
        return output;
    }

    public static String printCMDListToStr(int tabs, UICommandHelper[] cmdList) {
        String output = "";
        String tabString = makeTabs(tabs);
        // This is adjustable must be greater than longest command name
        int align = 40;
        for (UICommandHelper command : cmdList) {
            String cmd = command.getCommandName();
            output += tabString;
            output += cmd;
            int len = align - cmd.length();
            len = ((len <= 0) ? 1 : len);
            output += new String(new char[len]).replace("\0", " ");
            output += "(" + command.getCommandShortName() + ")";
            output += "\n";
        }
        return output;

    }

    // VALIDATION CODE FOR USER INPUT handling for invalid responses (thrown exceptions or returned booleans) are
    // handled by the caller
    public static boolean isValidPortNum(int port) {
        return (0 <= port && port < 65536);
    }

    /*
     * Will determine whether or not the input is a valid command
     * 
     * @returns String the full-length command name, if valid
     * 
     * @throws Exception with an appropriate error message, if invalid
     */
    public static UICommandHelper validateInputCommand(List<String> args) throws Exception {
        if (args.size() < 1) {
            throw new Exception(INVALID_USAGE);
        }
        String findCommand = args.get(0).toLowerCase();
        // work on assumption our COMMAND List has full command names at even indices and shortcuts at odd indices
        for (UICommandHelper command : COMMANDS) {
            if (command.equals(findCommand)) {
                return command;
            }
        }
        throw new Exception(VALID_COMMAND_LIST);
    }

    // This is the primary entry method to the Parser class which will handle essentially all input processing assumes
    // you have called Validate Command
    public static TCPClientRequest generateCommandData(String userInput) throws Exception {
        TCPClientRequest tcpClientRequest = new TCPClientRequest();
        List<String> args = new ArrayList<String>(Arrays.asList(userInput.split(UIUtil.WHITESPACE)));
        // Determine if user wants to print to a file
        int idxfile = args.indexOf(UIUtil.FILE_COMMAND);
        if (idxfile != -1) {
            tcpClientRequest.setPrintToFile(true);
            args.remove(idxfile); // remove the command option
            if (idxfile < args.size() && !args.get(idxfile).contains("-")) {
                tcpClientRequest.setFileName(args.get(idxfile));
                args.remove(idxfile); // now remove the command details as it will be in the spot the option just was
            }
            System.out.println("Printing responses to file: " + tcpClientRequest.getFileName());
        }

        // will throw exception with helpful message if fail
        UICommandHelper command = validateInputCommand(args);
        args.remove(0); // cut out the command from our input array
        args = getDestArgs(tcpClientRequest, args);
        validateNumArgs(command, args);
        tcpClientRequest.setMethod(command);
        tcpClientRequest.setArguments(args.indexOf(ARGUMENTS_COMMAND), args);
        return tcpClientRequest;
    }

    private static List<String> getDestArgs(TCPClientRequest tcpClientRequest, List<String> args) {
        String host = "";
        if (args.indexOf(HOST_COMMAND) >= 0) {
            int hostIndex = args.indexOf(HOST_COMMAND);
            args.remove(hostIndex);
            if (args.size() > hostIndex) {
                host = args.get(hostIndex);
                args.remove(hostIndex);
            }
        }
        String port = "";
        if (args.indexOf(PORT_COMMAND) >= 0) {
            int portIndex = args.indexOf(PORT_COMMAND);
            args.remove(portIndex);
            if (args.size() > portIndex) {
                port = args.get(portIndex);
                args.remove(portIndex);
            }
        }
        if (host.length() > 0 && port.length() > 0) {
            tcpClientRequest.setSendToTCPClientKey(host, port);
        }
        if (args.indexOf(REFERENCE_COMMAND) >= 0) {
            int refIndex = args.indexOf(REFERENCE_COMMAND);
            args.remove(refIndex);
            if (args.size() > refIndex) {
                tcpClientRequest.setSendToTCPClientIndex(args.get(refIndex));
                args.remove(refIndex);
            }
        }
        return args;
    }

    /*
     * Args must be in the form: [dest] (-a) [method args]
     * 
     * where [dest] is either empty (no values) or [host, writeport, readport ] or %x (where x is the identifier of an
     * existing connection)
     * 
     * if -a is not present, assumes no method args
     * 
     * depending on command, differing numbers of args are required
     * 
     * @return the index of "-a" or -1 if not included
     */
    private static void validateNumArgs(UICommandHelper command, List<String> args) throws Exception {
        int idxargs = -1;
        int argsRequired = numReqArgs(command);
        if (argsRequired == -1) {
            throw new Exception("Unrecognized Command");
        }
        if (args == null && argsRequired != 0) {
            throw new Exception(String.format(MISSING_ARGS_FORMAT, command.getCommandName(), argsRequired, 0));
        }
        idxargs = args.indexOf(ARGUMENTS_COMMAND);

        // There must be exactly 0 or at least 2 destination arguments after 2nd arg, all will be ignored
        // unless first is an identifier ('%[0-9]*') in which case all but first will be ignored

        // This case - "-a" is not in the arglist
        int templen = idxargs == -1 ? args.size() : idxargs;
        if (templen < 2 && templen != 0 && !args.get(0).startsWith(REFERENCE)) {
            throw new Exception(MISSING_DEST_ARGS);
        }

        // remaining values are method arguments if there are not enough args, complain
        if (idxargs == -1 && argsRequired > 0) {
            throw new Exception(String.format(MISSING_ARGS_FORMAT, command.getCommandName(), argsRequired, 0));
        }
        if (args.size() - idxargs - 1 < argsRequired) {
            throw new Exception(String.format(MISSING_ARGS_FORMAT, command.getCommandName(), argsRequired, args.size() - idxargs - 1));
        }
    }

    private static int numReqArgs(UICommandHelper command) {
        if (command.equals(RETRIEVE_HANDLERS) || command.equals(RETRIEVE_LOGS) || command.equals(RETRIEVE_SUBSCRIPTIONS) || command.equals(RETRIEVE_CONFIG)
                || command.equals(SESSION_BOUNCE)) {
            return 0;
        }
        if (command.equals(RETRIEVE_SNAPSHOT) || command.equals(UNSUBSCRIBE)) {
            return 1;
        }
        if (command.equals(SUBSCRIBE)) {
            return 13;
        }
        return -1;
    }
}